package gov.nasa.worldwind.render;

import com.sun.opengl.util.texture.TextureCoords;
import gov.nasa.worldwind.Movable;
import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.globes.Globe;
import gov.nasa.worldwind.util.Logging;

import javax.media.opengl.GL;
import java.awt.*;
import java.awt.geom.*;
import java.util.Arrays;

/**
 * Renders an icon image over the terrain surface.
 *
 * @author Patrick Murris
 * @version $Id: SurfaceIcon.java 11630 2009-06-14 16:24:15Z patrickmurris $
 */
public class SurfaceIcon extends AbstractSurfaceRenderable implements Movable
{
    private Object imageSource;
    private LatLon location;
    private Vec4 locationOffset;                    // Pixels
    private double scale = 1d;
    private Angle heading;
    private Color color = Color.WHITE;
    private double maxSize = Double.MAX_VALUE;      // Meter
    private double minSize = .1;                    // Meter

    protected WWTexture texture;
    protected int imageWidth = 32;
    protected int imageHeight = 32;

    public SurfaceIcon(Object imageSource, LatLon location)
    {
        this.setImageSource(imageSource);
        this.setLocation(location);
    }

    /**
     * Get the icon reference location on the globe.
     *
     * @return the icon reference location on the globe.
     */
    public LatLon getLocation()
    {
        return this.location;
    }

    /**
     * Set the icon reference location on the globe.
     *
     * @param location the icon reference location on the globe.
     * @throws {@link IllegalArgumentException} if location is <code>null</code>.
     */
    public void setLocation(LatLon location)
    {
        if (location == null)
        {
            String message = Logging.getMessage("nullValue.LatLonIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.location = location;
    }

    /**
     * Get the icon displacement in pixels relative to the reference location.  Can be <code>null</code>.
     * <p>
     * When <code>null</code> the icon will be drawn with it's image center on top of it's refence location -
     * see {@link #setLocation(LatLon)}. Otherwise the icon will be shifted of a distance equivalent to the number
     * of pixels specified as <code>x</code> and <code>y</code> offset values. Positive values will move the icon
     * to the right for <code>x</code> and up for <code>y</code>. Negative values will have the opposite effect.
     *
     * @return the icon displacement in pixels relative to the reference location.
     */
    public Vec4 getLocationOffset()
    {
        return this.locationOffset;
    }

    /**
     * Set the icon displacement in pixels relative to the reference location. Can be <code>null</code>.
     * <p>
     * When <code>null</code> the icon will be drawn with it's image center on top of it's refence location -
     * see {@link #setLocation(LatLon)}. Otherwise the icon will be shifted of a distance equivalent to the number
     * of pixels specified as <code>x</code> and <code>y</code> offset values. Positive values will move the icon
     * to the right for <code>x</code> and up for <code>y</code>. Negative values will have the opposite effect.
     *
     * @param locationOffset the icon displacement in pixels relative to the reference location.
     */
    public void setLocationOffset(Vec4 locationOffset)
    {
        this.locationOffset = locationOffset; // can be null
    }

    /**
     * Get the source for the icon image. Can be a file path to a local image or
     * a {@link java.awt.image.BufferedImage} reference.
     *
     * @return the source for the icon image.
     */
    public Object getImageSource()
    {
        return this.imageSource;
    }

    /**
     * Set the source for the icon image. Can be a file path to a local image or
     * a {@link java.awt.image.BufferedImage} reference.
     *
     * @param imageSource the source for the icon image.
     * @throws {@link IllegalArgumentException} if imageSource is <code>null</code>.
     */
    public void setImageSource(Object imageSource)
    {
        if (imageSource == null)
        {
            String message = Logging.getMessage("nullValue.ImageSourceIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.imageSource = imageSource;
        this.texture = null;
    }

    /**
     * Get the current scaling factor applied to the source image.
     * 
     * @return the current scaling factor applied to the source image.
     */
    public double getScale()
    {
        return this.scale;
    }

    /**
     * Set the scaling factor to apply to the source image. A value of <code>1</code> will produce no change,
     * a value greater then <code>1</code> will enlarge the image and a value smaller then <code>1</code> will
     * reduce it.
     *
     * @param scale the scaling factor to apply to the source image.
     */
    public void setScale(double scale)
    {
        this.scale = scale;
    }

    /**
     * Get the current heading {@link Angle}, clockwise from North or <code>null</code>.
     *
     * @return the current heading {@link Angle}, clockwise from North or <code>null</code>.
     */
    public Angle getHeading()
    {
        return this.heading;
    }

    /**
     * Set the heading {@link Angle}, clockwise from North. Setting this value to <code>null</code> will have
     * the icon follow the view heading so as to always face the eye. The icon will rotate around it's reference
     * location.
     *
     * @param heading the heading {@link Angle}, clockwise from North or <code>null</code>.
     */
    public void setHeading(Angle heading)
    {
        this.heading = heading;  // can be null
    }

    /**
     * Get the minimum size in meter the icon image is allowed to be reduced to once applied to the terrain surface.
     * This limit applies to the source image largest dimension.
     * <p>
     * The icon will try to maintain it's apparent size depending on it's distance from the eye and will extend
     * over a rectangular area which largest dimension is bounded by the values provided with
     * {@link #setMinSize(double)} and {@link #setMaxSize(double)}.
     *
     * @return the minimum size of the icon in meter.
     */
    public double getMinSize()
    {
        return this.minSize;
    }

    /**
     * Set the minimum size in meter the icon image is allowed to be reduced to once applied to the terrain surface.
     * This limit applies to the source image largest dimension.
     * <p>
     * The icon will try to maintain it's apparent size depending on it's distance from the eye and will extend
     * over a rectangular area which largest dimension is bounded by the values provided with
     * {@link #setMinSize(double)} and {@link #setMaxSize(double)}.
     *
     * @param sizeInMeter the minimum size of the icon in meter.
     */
    public void setMinSize(double sizeInMeter)
    {
        this.minSize = sizeInMeter;
    }

    /**
     * Get the maximum size in meter the icon image is allowed to be enlarged to once applied to the terrain surface.
     * This limit applies to the source image largest dimension.
     * <p>
     * The icon will try to maintain it's apparent size depending on it's distance from the eye and will extend
     * over a rectangular area which largest dimension is bounded by the values provided with
     * {@link #setMinSize(double)} and {@link #setMaxSize(double)}.
     *
     * @return the maximum size of the icon in meter.
     */
    public double getMaxSize()
    {
        return this.maxSize;
    }

    /**
     * Get the maximum size in meter the icon image is allowed to be enlarged to once applied to the terrain surface.
     * This limit applies to the source image largest dimension.
     * <p>
     * The icon will try to maintain it's apparent size depending on it's distance from the eye and will extend
     * over a rectangular area which largest dimension is bounded by the values provided with
     * {@link #setMinSize(double)} and {@link #setMaxSize(double)}.
     *
     * @param sizeInMeter the maximum size of the icon in meter.
     */
    public void setMaxSize(double sizeInMeter)
    {
        this.maxSize = sizeInMeter;
    }

    /**
     * Get the {@link Color} the source image is combined with.
     *
     * @return the {@link Color} the source image is combined with.
     */
    public Color getColor()
    {
        return this.color;
    }

    /**
     * Set the {@link Color} the source image will be combined with - default to white.
     * <p>
     * A non white color will mostly affect the white portions from the original image. This is mostly useful
     * to alter the appearance of 'colorless' icons - which mainly contain black, white and shades of gray.
     *
     * @param color the {@link Color} the source image will be combined with.
     * @throws {@link IllegalArgumentException} if color is <code>null</code>.
     */
    public void setColor(Color color)
    {
        if (color == null)
        {
            String message = Logging.getMessage("nullValue.ColorIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }
        this.color = color;
    }

    // *** SurfaceObject interface ***

    public long getLastModifiedTime()
    {
        return System.currentTimeMillis();  // Refresh all the time TODO
    }
    
    public Iterable<? extends Sector> getSectors(DrawContext dc, double texelSizeRadians)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        return Arrays.asList(computeSector(dc));
    }

    public void doRenderToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        this.beginDraw(dc);
        try
        {
            this.applyDrawTransform(dc, sector, x, y, width, height);
            this.applyDrawColor(dc);
            
            //Draw
            this.drawIcon(dc);
        }
        catch (Exception e)
        {
            // TODO: log error
        }
        finally
        {
            // Restore gl state
            this.endDraw(dc);
        }
    }

    protected Sector computeSector(DrawContext dc)
    {
        Globe globe = dc.getGlobe();
        // Compute real world icon extent depending on distance from eye
        Rectangle2D.Double rect = computeDrawDimension(dc); // meter
        // Compute bounding sector and apply location offset to it
        double dLatRadians = rect.height / globe.getRadius();
        double dLonRadians = rect.width / globe.getRadius() / location.getLatitude().cos();
        double offsetLatRadians = locationOffset != null ? locationOffset.y * dLatRadians / this.imageHeight : 0;
        double offsetLonRadians = locationOffset != null ? locationOffset.x * dLonRadians / this.imageWidth : 0;
        Sector sector = new Sector(
            location.getLatitude().subtractRadians(dLatRadians / 2).addRadians(offsetLatRadians),
            location.getLatitude().addRadians(dLatRadians / 2).addRadians(offsetLatRadians),
            location.getLongitude().subtractRadians(dLonRadians / 2).addRadians(offsetLonRadians),
            location.getLongitude().addRadians(dLonRadians / 2).addRadians(offsetLonRadians)
        );
        // Rotate sector around location 
        return computeRotatedSectorBounds(sector, this.location, computeDrawHeading(dc));
    }

    protected Rectangle2D.Double computeDrawDimension(DrawContext dc)
    {
        // Compute icon extent at 1:1
        double pixelSize = computePixelSizeAtLocation(dc, this.location);
        double height = this.imageHeight * this.scale * pixelSize;
        double width = this.imageWidth * this.scale * pixelSize;
        // Clamp to size range
        double size = height > width ? height : width;
        double scale = size > this.maxSize ? this.maxSize / size : size < this.minSize ? this.minSize / size : 1;

        return new Rectangle2D.Double(0, 0, width * scale, height * scale);
    }

    protected Angle computeDrawHeading(DrawContext dc)
    {
        if (this.heading != null)
            return this.heading;

        return getViewHeading(dc);
    }

    protected void beginDraw(DrawContext dc)
    {
        GL gl = dc.getGL();

        int attributeMask = GL.GL_TRANSFORM_BIT // for modelview
                | GL.GL_CURRENT_BIT // for current color
                | GL.GL_COLOR_BUFFER_BIT // for alpha test func and ref, and blend
                | GL.GL_TEXTURE_BIT // for texture env
                | GL.GL_ENABLE_BIT; // for enable/disable changes
        gl.glPushAttrib(attributeMask);

        // Suppress any fully transparent image pixels
        gl.glEnable(GL.GL_ALPHA_TEST);
        gl.glAlphaFunc(GL.GL_GREATER, 0.001f);

        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glPushMatrix();

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPushMatrix();

        if (dc.isPickingMode())
        {
            this.pickSupport.beginPicking(dc);

            // Set up to replace the non-transparent texture colors with the single pick color.
            gl.glEnable(GL.GL_TEXTURE_2D);
            gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_COMBINE);
            gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_SRC0_RGB, GL.GL_PREVIOUS);
            gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_COMBINE_RGB, GL.GL_REPLACE);
        }
        else
        {
            gl.glEnable(GL.GL_TEXTURE_2D);
            gl.glEnable(GL.GL_BLEND);
            gl.glBlendFunc(GL.GL_ONE, GL.GL_ONE_MINUS_SRC_ALPHA);
            gl.glTexEnvf(GL.GL_TEXTURE_ENV, GL.GL_TEXTURE_ENV_MODE, GL.GL_MODULATE);
        }
    }

    protected void endDraw(DrawContext dc)
    {
        if (dc.isPickingMode())
            this.pickSupport.endPicking(dc);

        GL gl = dc.getGL();
        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glPopMatrix();

        gl.glMatrixMode(GL.GL_TEXTURE);
        gl.glPopMatrix();

        gl.glPopAttrib();
    }

    protected void applyDrawTransform(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        // Compute icon viewport point
        Vec4 point = computeDrawPoint(this.location, sector, x, y, width, height);

        // Compute scaling
        double regionPixelSize = computeDrawPixelSize(dc, sector, width, height);
        double drawPixelSize = computeDrawDimension(dc).width / this.imageWidth;
        double drawScale = drawPixelSize / regionPixelSize;

        GL gl = dc.getGL();
        // Translate to location point
        gl.glTranslated(point.x(), point.y(), point.z());
        // Add x scaling transform to maintain icon width and aspect ratio at any latitude
        gl.glScaled(drawScale/ this.location.getLatitude().cos(), drawScale, 1);
        // Add rotation to account for icon heading
        gl.glRotated(computeDrawHeading(dc).degrees, 0, 0, -1);
        // Translate to lower left corner
        gl.glTranslated(-this.imageWidth / 2, -this.imageHeight / 2, 0);
        // Apply location offset if any
        if (this.locationOffset != null)
            gl.glTranslated(this.locationOffset.x, this.locationOffset.y, 0);
    }

    protected void applyDrawColor(DrawContext dc)
    {
        if (!dc.isPickingMode())
            applyPremultipliedAlphaColor(dc.getGL(), this.color, getOpacity());
    }

    protected void drawIcon(DrawContext dc)
    {
        WWTexture texture = getTexture();
        if (texture == null)
            return;

        if (!texture.bind(dc))
            return;

        // Update image width and height
        this.imageWidth = texture.getWidth();
        this.imageHeight = texture.getHeight();
        GL gl = dc.getGL();
        gl.glMatrixMode(GL.GL_TEXTURE);
        texture.applyInternalTransform(dc);

        gl.glMatrixMode(GL.GL_MODELVIEW);
        gl.glScaled(this.imageWidth, this.imageHeight, 1d);
        dc.drawUnitQuad(new TextureCoords(0, 0, 1, 1));
    }

    protected WWTexture getTexture()
    {
        if (this.texture == null)
            this.texture = new BasicWWTexture(this.imageSource);

        return this.texture;
    }

    // *** Movable interface

    public Position getReferencePosition()
    {
        return new Position(this.location, 0);
    }

    public void move(Position delta)
    {
        if (delta == null)
        {
            String msg = Logging.getMessage("nullValue.PositionIsNull");
            Logging.logger().severe(msg);
            throw new IllegalArgumentException(msg);
        }

        this.moveTo(this.getReferencePosition().add(delta));
    }

    public void moveTo(Position position)
    {
        if (position == null)
        {
            String msg = Logging.getMessage("nullValue.PositionIsNull");
            Logging.logger().severe(msg);
            throw new IllegalArgumentException(msg);
        }

        this.setLocation(position.getLatLon());
    }
}
