/* Copyright (C) 2001, 2008 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.awt;

import gov.nasa.worldwind.Configuration;
import gov.nasa.worldwind.util.Logging;

import java.util.*;

/**
 * @author dcollins
 * @version $Id: ViewInputAttributes.java 8975 2009-02-24 18:38:35Z dcollins $
 */
public class ViewInputAttributes
{
    public static class ActionAttributes
    {
        private double minValue;
        private double maxValue;
        private boolean enableSmoothing;
        private double smoothingValue;

        public ActionAttributes(double minValue, double maxValue, boolean enableSmoothing, double smoothingValue)
        {
            this.setValues(minValue, maxValue);
            this.setEnableSmoothing(enableSmoothing);
            this.setSmoothingValue(smoothingValue);
        }

        public ActionAttributes(ActionAttributes attributes)
        {
            if (attributes == null)
            {
                String message = Logging.getMessage("nullValue.AttributesIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }
            
            this.minValue = attributes.minValue;
            this.maxValue = attributes.maxValue;
            this.smoothingValue = attributes.smoothingValue;
        }

        public double[] getValues()
        {
            return new double[] {this.minValue, this.maxValue};
        }

        public void setValues(double minValue, double maxValue)
        {
            if (minValue <= 0)
            {
                String message = Logging.getMessage("generic.ArgumentOutOfRange", "minValue <= 0");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }
            if (maxValue <= 0)
            {
                String message = Logging.getMessage("generic.ArgumentOutOfRange", "maxValue <= 0");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.minValue = minValue;
            this.maxValue = maxValue;
        }

        public void setValue(double value)
        {
            this.setValues(value, value);
        }

        public boolean isEnableSmoothing()
        {
            return this.enableSmoothing;
        }

        public void setEnableSmoothing(boolean enable)
        {
            this.enableSmoothing = enable;
        }

        public double getSmoothingValue()
        {
            return this.smoothingValue;
        }

        public void setSmoothingValue(double smoothingValue)
        {
            if (smoothingValue < 0 || smoothingValue >= 1.0)
            {
                String message = Logging.getMessage("generic.ArgumentOutOfRange", "smoothingValue < 0 || smoothingValue >= 1");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.smoothingValue = smoothingValue;
        }
    }

    public static class DeviceAttributes
    {
        private double sensitivity;

        public DeviceAttributes(double sensitivity)
        {
            this.setSensitivity(sensitivity);
        }

        public DeviceAttributes(DeviceAttributes attributes)
        {
            if (attributes == null)
            {
                String message = Logging.getMessage("nullValue.AttributesIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.sensitivity = attributes.sensitivity;
        }

        public double getSensitivity()
        {
            return this.sensitivity;
        }

        public void setSensitivity(double sensitivity)
        {
            if (sensitivity <= 0)
            {
                String message = Logging.getMessage("generic.ArgumentOutOfRange", "sensitivity <= 0");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.sensitivity = sensitivity;
        }
    }

    public static class ActionAttributesMap
    {
        private Map<Object, ActionAttributes> actionMap = new HashMap<Object, ActionAttributes>();

        public ActionAttributesMap()
        {
        }

        public ActionAttributes getActionAttributes(Object actionKey)
        {
            if (actionKey == null)
            {
                String message = Logging.getMessage("nullValue.ActionKeyIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            return this.actionMap.get(actionKey);
        }

        public void setActionAttributes(Object actionKey, ActionAttributes attributes)
        {
            if (actionKey == null)
            {
                String message = Logging.getMessage("nullValue.ActionKeyIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }
            if (attributes == null)
            {
                String message = Logging.getMessage("nullValue.AttributesIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.actionMap.put(actionKey, attributes);
        }
    }

    public static final String VIEW_FOCUS = "gov.nasa.worldwind.ViewFocus";
    public static final String VIEW_FOCUS_SLOW = "gov.nasa.worldwind.ViewFocusSlow";
    public static final String VIEW_PAN = "gov.nasa.worldwind.ViewPan";
    public static final String VIEW_PAN_SLOW = "gov.nasa.worldwind.ViewPanSlow";
    public static final String VIEW_ROTATE = "gov.nasa.worldwind.ViewRotate";
    public static final String VIEW_ROTATE_SLOW = "gov.nasa.worldwind.ViewRotateSlow";
    public static final String VIEW_ZOOM = "gov.nasa.worldwind.ViewZoom";
    public static final String VIEW_ZOOM_SLOW = "gov.nasa.worldwind.ViewZoomSlow";
    public static final String DEVICE_KEYBOARD = "gov.nasa.worldwind.DeviceKeyboard";
    public static final String DEVICE_MOUSE = "gov.nasa.worldwind.DeviceMouse";
    public static final String DEVICE_MOUSE_WHEEL = "gov.nasa.worldwind.DeviceMouseWheel";

    // Device sensitivity defaults.
    private static final double DEFAULT_KEY_SENSITIVITY = 1.0; // Scalar multiplier
    private static final double DEFAULT_MOUSE_SENSITIVITY = 1.0; // Scalar multiplier
    private static final double DEFAULT_MOUSE_WHEEL_SENSITIVITY = 1.0; // Scalar multiplier
    private static final double DEFAULT_SLOW_VALUE = 0.25; // Scalar multiplier
    // Action smoothing defaults.
    private static final boolean DEFAULT_FOCUS_SMOOTHING_ENABLED = true;
    private static final boolean DEFAULT_PAN_SMOOTHING_ENABLED = true;
    private static final boolean DEFAULT_ROTATE_SMOOTHING_ENABLED = true;
    private static final boolean DEFAULT_ZOOM_SMOOTHING_ENABLED = true;
    private static final double DEFAULT_FOCUS_SMOOTHING_VALUE = 0.0; // [0, 1] smoothing value
    private static final double DEFAULT_PAN_SMOOTHING_VALUE = 0.4; // [0, 1] smoothing value
    private static final double DEFAULT_ROTATE_SMOOTHING_VALUE = 0.7; // [0, 1] smoothing value
    private static final double DEFAULT_ZOOM_SMOOTHING_VALUE = 0.85; // [0, 1] smoothing value
    // Keyboard/Action calibration values.
    private static final double DEFAULT_KEY_PAN_MIN_VALUE = 0.000005; // Speed in degrees per frame
    private static final double DEFAULT_KEY_PAN_MAX_VALUE = 4.0; // Speed in degrees per frame
    private static final double DEFAULT_KEY_ROTATE_MIN_VALUE = 2.0; // Speed in degrees per frame
    private static final double DEFAULT_KEY_ROTATE_MAX_VALUE = 2.2; // Speed in degrees per frame
    private static final double DEFAULT_KEY_ZOOM_VALUE = 0.06; // Speed in log-meters per frame
    // Mouse/Action calibration values.
    private static final double DEFAULT_MOUSE_FOCUS_MIN_VALUE = 0.95; // [0, 1] smoothing value
    private static final double DEFAULT_MOUSE_FOCUS_MAX_VALUE = 0.90; // [0, 1] smoothing value
    private static final double DEFAULT_MOUSE_PAN_MIN_VALUE = 0.00001; // Speed in degrees per mouse movement
    private static final double DEFAULT_MOUSE_PAN_MAX_VALUE = 0.2; // Speed in degrees per mouse movement
    private static final double DEFAULT_MOUSE_ROTATE_MIN_VALUE = 0.14; // Speed in degrees per mouse movement
    private static final double DEFAULT_MOUSE_ROTATE_MAX_VALUE = 0.18; // Speed in degrees per mouse movement
    private static final double DEFAULT_MOUSE_ZOOM_VALUE = 0.003; // Speed in log-meters per mouse movement
    // MouseWheel/Action calibration values.
    private static final double DEFAULT_MOUSE_WHEEL_ZOOM_VALUE = 0.1; // Speed in log-meters per wheel movement
    private static final double DEFAULT_MOUSE_WHEEL_ZOOM_VALUE_OSX = 0.01; // Speed in log-meters per wheel movement

    // Device attributes.
    private Map<Object, DeviceAttributes> deviceMap = new HashMap<Object, DeviceAttributes>();
    // Device/action pairing attributes.
    private Map<Object, ActionAttributesMap> deviceActionMap = new HashMap<Object, ActionAttributesMap>();

    public ViewInputAttributes()
    {
        this.setDefaultDeviceAttributes();
        this.setDefaultActionAttributes();
    }

    public DeviceAttributes getDeviceAttributes(Object deviceKey)
    {
        if (deviceKey == null)
        {
            String message = Logging.getMessage("nullValue.DeviceKeyIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        return this.deviceMap.get(deviceKey);
    }

    public void setDeviceAttributes(Object deviceKey, DeviceAttributes attributes)
    {
        if (deviceKey == null)
        {
            String message = Logging.getMessage("nullValue.DeviceKeyIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }
        if (attributes == null)
        {
            String message = Logging.getMessage("nullValue.AttributesIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.deviceMap.put(deviceKey, attributes);
    }

    public ActionAttributesMap getActionMap(Object deviceKey)
    {
        if (deviceKey == null)
        {
            String message = Logging.getMessage("nullValue.DeviceKeyIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        return this.deviceActionMap.get(deviceKey);
    }

    public void setActionMap(Object deviceKey, ActionAttributesMap map)
    {
        if (deviceKey == null)
        {
            String message = Logging.getMessage("nullValue.DeviceKeyIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }
        if (map == null)
        {
            String message = Logging.getMessage("nullValue.MapIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.deviceActionMap.put(deviceKey, map);
    }

    public ActionAttributes getActionAttributes(Object deviceKey, Object actionKey)
    {
        if (deviceKey == null)
        {
            String message = Logging.getMessage("nullValue.DeviceKeyIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }
        if (actionKey == null)
        {
            String message = Logging.getMessage("nullValue.ActionKeyIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        ActionAttributesMap map = this.getActionMap(deviceKey);
        if (map == null)
            return null;

        return map.getActionAttributes(actionKey);
    }

    //**************************************************************//
    //********************  Default Attributes  ********************//
    //**************************************************************//

    protected void setDefaultDeviceAttributes()
    {
        this.setDeviceAttributes(DEVICE_KEYBOARD, new DeviceAttributes(DEFAULT_KEY_SENSITIVITY));
        this.setDeviceAttributes(DEVICE_MOUSE, new DeviceAttributes(DEFAULT_MOUSE_SENSITIVITY));
        this.setDeviceAttributes(DEVICE_MOUSE_WHEEL, new DeviceAttributes(DEFAULT_MOUSE_WHEEL_SENSITIVITY));
    }

    protected void setDefaultActionAttributes()
    {
        // Default keyboard/action calibration values.
        ActionAttributesMap map = new ActionAttributesMap();
        map.setActionAttributes(VIEW_PAN,
            new ActionAttributes(DEFAULT_KEY_PAN_MIN_VALUE, DEFAULT_KEY_PAN_MAX_VALUE, DEFAULT_PAN_SMOOTHING_ENABLED, DEFAULT_PAN_SMOOTHING_VALUE));
        map.setActionAttributes(VIEW_ROTATE,
            new ActionAttributes(DEFAULT_KEY_ROTATE_MIN_VALUE, DEFAULT_KEY_ROTATE_MAX_VALUE, DEFAULT_ROTATE_SMOOTHING_ENABLED, DEFAULT_ROTATE_SMOOTHING_VALUE));
        map.setActionAttributes(VIEW_ZOOM,
            new ActionAttributes(DEFAULT_KEY_ZOOM_VALUE, DEFAULT_KEY_ZOOM_VALUE, DEFAULT_ZOOM_SMOOTHING_ENABLED, DEFAULT_ZOOM_SMOOTHING_VALUE));
        this.setSlowActionAttributes(map, DEFAULT_SLOW_VALUE);
        this.setActionMap(DEVICE_KEYBOARD, map);

        // Default mouse/action calibration values.
        map = new ActionAttributesMap();
        map.setActionAttributes(VIEW_FOCUS,
            new ActionAttributes(DEFAULT_MOUSE_FOCUS_MIN_VALUE, DEFAULT_MOUSE_FOCUS_MAX_VALUE, DEFAULT_FOCUS_SMOOTHING_ENABLED, DEFAULT_FOCUS_SMOOTHING_VALUE));
        map.setActionAttributes(VIEW_PAN,
            new ActionAttributes(DEFAULT_MOUSE_PAN_MIN_VALUE, DEFAULT_MOUSE_PAN_MAX_VALUE, DEFAULT_PAN_SMOOTHING_ENABLED, DEFAULT_PAN_SMOOTHING_VALUE));
        map.setActionAttributes(VIEW_ROTATE,
            new ActionAttributes(DEFAULT_MOUSE_ROTATE_MIN_VALUE, DEFAULT_MOUSE_ROTATE_MAX_VALUE, DEFAULT_ROTATE_SMOOTHING_ENABLED, DEFAULT_ROTATE_SMOOTHING_VALUE));
        map.setActionAttributes(VIEW_ZOOM,
            new ActionAttributes(DEFAULT_MOUSE_ZOOM_VALUE, DEFAULT_MOUSE_ZOOM_VALUE, DEFAULT_ZOOM_SMOOTHING_ENABLED, DEFAULT_ZOOM_SMOOTHING_VALUE));
        this.setSlowActionAttributes(map, DEFAULT_SLOW_VALUE);
        this.setActionMap(DEVICE_MOUSE, map);

        // Default mousewheel/action calibration values.
        map = new ActionAttributesMap();
        if (Configuration.isMacOS())
        {
            map.setActionAttributes(VIEW_ZOOM,
                new ActionAttributes(DEFAULT_MOUSE_WHEEL_ZOOM_VALUE_OSX, DEFAULT_MOUSE_WHEEL_ZOOM_VALUE_OSX, DEFAULT_ZOOM_SMOOTHING_ENABLED, DEFAULT_ZOOM_SMOOTHING_VALUE));
            this.setSlowActionAttributes(map, DEFAULT_SLOW_VALUE);
        }
        else
        {
            map.setActionAttributes(VIEW_ZOOM,
                new ActionAttributes(DEFAULT_MOUSE_WHEEL_ZOOM_VALUE, DEFAULT_MOUSE_WHEEL_ZOOM_VALUE, DEFAULT_ZOOM_SMOOTHING_ENABLED, DEFAULT_ZOOM_SMOOTHING_VALUE));
            this.setSlowActionAttributes(map, DEFAULT_SLOW_VALUE);
        }
        this.setActionMap(DEVICE_MOUSE_WHEEL, map);
    }

    protected void setSlowActionAttributes(ActionAttributesMap map, double slowCoefficient)
    {
        ActionAttributes attributes = map.getActionAttributes(VIEW_FOCUS);
        if (attributes != null)
            map.setActionAttributes(VIEW_FOCUS_SLOW, this.makeSlowActionAttributes(attributes, slowCoefficient));

        attributes = map.getActionAttributes(VIEW_PAN);
        if (attributes != null)
            map.setActionAttributes(VIEW_PAN_SLOW, this.makeSlowActionAttributes(attributes, slowCoefficient));

        attributes = map.getActionAttributes(VIEW_ROTATE);
        if (attributes != null)
            map.setActionAttributes(VIEW_ROTATE_SLOW, this.makeSlowActionAttributes(attributes, slowCoefficient));

        attributes = map.getActionAttributes(VIEW_ZOOM);
        if (attributes != null)
            map.setActionAttributes(VIEW_ZOOM_SLOW, this.makeSlowActionAttributes(attributes, slowCoefficient));
    }

    protected ActionAttributes makeSlowActionAttributes(ActionAttributes attributes, double slowCoefficient)
    {
        ActionAttributes slowAttributes = new ActionAttributes(attributes);
        double[] values = attributes.getValues();
        slowAttributes.setValues(values[0] * slowCoefficient, values[1] * slowCoefficient);
        return slowAttributes;
    }
}
