/* Copyright (C) 2001, 2009 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.layers;

import gov.nasa.worldwind.*;
import gov.nasa.worldwind.avlist.*;
import gov.nasa.worldwind.exception.*;
import gov.nasa.worldwind.util.*;
import gov.nasa.worldwind.wms.*;
import org.w3c.dom.*;

import javax.xml.xpath.XPathConstants;

/**
 * A factory that creates {@link gov.nasa.worldwind.layers.Layer} instances.
 *
 * @author dcollins
 * @version $Id: BasicLayerFactory.java 11667 2009-06-16 15:50:18Z dcollins $
 */
public class BasicLayerFactory extends BasicFactory
{
    /** Creates an instance of BasicLayerFactory; otherwise does nothing. */
    public BasicLayerFactory()
    {
    }

    /**
     * Creates a layer from a configuration file.
     * <p/>
     * Maps the <code>layerType</code> attribute of the <code>Layer</code> element of the XML configuration file to the
     * appropriate layer type. Layer types recognized are: <ul> <li>"TiledImageLayer" for layers that render tiled
     * surface imagery.</li> </ul>
     * <p/>
     * For tiled image layers, this maps the <code>serviceName</code> attribute of the <code>Layer/Service</code>
     * element of the XML configuration file to the appropriate base tiled image layer type. Service types recognized
     * are: <ul> <li>"WMS" for layers that draw their data from a WMS web service.</li> <li>"WWTileService" for layers
     * that draw their data from a World Wind tile service.</li> <li>"Offline" for layers that draw their data only from
     * the local cache.</li> </ul>
     *
     * @param fileName the path to the configuration file. The file must be either an absolute path or a relative path
     *                 available on the classpath.
     *
     * @return a layer.
     *
     * @throws IllegalArgumentException if the configuration file name is null.
     * @throws WWUnrecognizedException  if the service type is unrecognized.
     * @throws WWRuntimeException       if object creation fails. The exception indicating the source of the failure is
     *                                  included as the {@link Exception#initCause(Throwable)}.
     */
    public Object createFromConfigFile(String fileName)
    {
        if (fileName == null)
        {
            String message = Logging.getMessage("nullValue.FilePathIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Layer layer;

        try
        {
            Document doc = WWXML.openDocumentFile(fileName, this.getClass());

            layer = this.doCreateFromDocument(doc.getDocumentElement(), null);
        }
        catch (IllegalStateException e)
        {
            // Message is passed up the stack as an IllegalStateException, caught here, then thrown as a
            // WWUnrecognizedException.
            throw new WWUnrecognizedException(e.getMessage());
        }
        catch (Exception e)
        {
            String msg = Logging.getMessage("generic.CreationFromConfigurationFileFailed", fileName);
            throw new WWRuntimeException(msg, e);
        }

        if (layer == null)
        {
            String msg = Logging.getMessage("generic.UnrecognizedDocument", fileName);
            throw new WWUnrecognizedException(msg);
        }

        return layer;
    }

    /**
     * Creates a layer from a data configuration.
     * <p/>
     * Maps the <code>Layer/layerType</code> configuration parameter to the appropriate layer type. Layer types
     * recognized are: <ul> <li>"TiledImageLayer" for layers that render tiled surface imagery.</li> </ul>
     * <p/>
     * For tiled image layers, this maps the <code>Layer/Service/serviceName</code> configuration property to the
     * appropriate base tiled image layer type. Service types recognized are: <ul> <li>"WMS" for layers that draw their
     * data from a WMS web service.</li> <li>"WWTileService" for layers that draw their data from a World Wind tile
     * service</li> <li>"Offline" for layers that draw their data only from the local cache.</li> </ul>
     *
     * @param dataConfig the configuration information describing the component.
     * @param params     key-value parameters which override or supplement the information provided in the specified
     *                   configuration. A null reference is permitted.
     *
     * @return a layer.
     *
     * @throws IllegalArgumentException if the data configuration is null.
     * @throws WWUnrecognizedException  if the service type is unrecognized.
     * @throws WWRuntimeException       if object creation fails. The exception indicating the source of the failure is
     *                                  included as the {@link Exception#initCause(Throwable)}.
     */
    public Object createFromDataConfig(DataConfiguration dataConfig, AVList params)
    {
        if (dataConfig == null)
        {
            String message = Logging.getMessage("nullValue.DataConfigurationIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Layer layer;

        try
        {
            layer = this.doCreateFromDataConfig(dataConfig, params);
        }
        catch (IllegalStateException e)
        {
            // Message is passed up the stack as an IllegalStateException, caught here, then thrown as a
            // WWUnrecognizedException.
            throw new WWUnrecognizedException(e.getMessage());
        }
        catch (Exception e)
        {
            String msg = Logging.getMessage("generic.CreationFromDataConfigurationFailed", dataConfig);
            throw new WWRuntimeException(msg, e);
        }

        if (layer == null)
        {
            String msg = Logging.getMessage("generic.UnrecognizedDataConfiguration", dataConfig);
            throw new WWUnrecognizedException(msg);
        }

        return layer;
    }

    protected Layer doCreateFromDataConfig(DataConfiguration dataConfig, AVList params) throws Exception
    {
        String type = dataConfig.getType();

        if (type.equals("Capabilities"))
        {
            Capabilities caps = (Capabilities) dataConfig.getSource();
            return this.doCreateFromCapabilities(caps, params);
        }
        else if (type.equals("Layer"))
        {
            Element domElement = (Element) dataConfig.getSource();
            return this.doCreateFromDocument(domElement, params);
        }

        return null;
    }

    protected Layer doCreateFromCapabilities(Capabilities caps, AVList params)
    {
        String serviceName = caps.getServiceName();
        if (serviceName == null || !(serviceName.equalsIgnoreCase(Capabilities.WMS_SERVICE_NAME)
            || serviceName.equalsIgnoreCase("WMS")))
        {
            String message = Logging.getMessage("WMS.NotWMSService", serviceName != null ? serviceName : "null");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (params == null)
            params = new AVListImpl();

        if (params.getStringValue(AVKey.LAYER_NAMES) == null)
        {
            Element[] namedLayers = caps.getNamedLayers();

            if (namedLayers == null || namedLayers.length == 0)
            {
                String message = Logging.getMessage("WMS.NoLayersFound");
                Logging.logger().severe(message);
                throw new IllegalStateException(message);
            }

            // Use the first named layer if no other guidance given
            params.setValue(AVKey.LAYER_NAMES, caps.getLayerName(namedLayers[0]));
        }

        return new WMSTiledImageLayer(caps, params);
    }

    protected Layer doCreateFromDocument(Element domElement, AVList params) throws Exception
    {
        NodeList emNodes = (NodeList) WWXML.makeXPath().evaluate("//Layer", domElement, XPathConstants.NODESET);

        if (emNodes == null || emNodes.getLength() == 0)
            return null;

        // Ignore all but the first TiledImageLayer elements.
        Element root = (Element) emNodes.item(0);
        return this.createFromLayerDocument(root, params);
    }

    protected Layer createFromLayerDocument(Element domElement, AVList params)
    {
        String layerType = WWXML.getText(domElement, "@layerType");

        if (layerType != null && layerType.equals("TiledImageLayer"))
        {
            return this.createTiledImageLayer(domElement, params);
        }
        else
        {
            String msg = Logging.getMessage("generic.UnrecognizedLayerType", layerType);
            throw new IllegalStateException(msg);
        }
    }

    protected Layer createTiledImageLayer(Element domElement, AVList params)
    {
        Layer layer;

        String serviceName = WWXML.getText(domElement, "Service/@serviceName");

        if (serviceName != null && serviceName.equals("Offline"))
        {
            layer = new BasicTiledImageLayer(domElement, params);
        }
        else if (serviceName != null && serviceName.equals("WWTileService"))
        {
            layer = new BasicTiledImageLayer(domElement, params);
        }
        else if (serviceName != null && serviceName.equals(Capabilities.WMS_SERVICE_NAME))
        {
            layer = new WMSTiledImageLayer(domElement, params);
        }
        else
        {
            String msg = Logging.getMessage("generic.UnrecognizedServiceName", serviceName);
            throw new IllegalStateException(msg);
        }

        return layer;
    }
}
