/*
Copyright (C) 2001, 2006 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.layers.placename;

import gov.nasa.worldwind.*;
import gov.nasa.worldwind.avlist.*;
import gov.nasa.worldwind.cache.*;
import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.globes.Globe;
import gov.nasa.worldwind.layers.*;
import gov.nasa.worldwind.render.*;
import gov.nasa.worldwind.retrieve.*;
import gov.nasa.worldwind.util.*;

import java.net.URL;
import java.nio.*;
import java.nio.channels.ClosedByInterruptException;
import java.util.*;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.logging.Level;

/**
 * @author Paul Collins
 * @version $Id: PlaceNameLayer.java 10925 2009-05-06 22:22:50Z dcollins $
 */
public class PlaceNameLayer extends AbstractLayer implements BulkRetrievable
{
    private final PlaceNameServiceSet placeNameServiceSet;
    private PriorityBlockingQueue<Runnable> requestQ = new PriorityBlockingQueue<Runnable>(64);
    private Vec4 referencePoint;
    private final Object fileLock = new Object();
    private boolean cullNames = false;

    protected static final double LEVEL_A = 0x1 << 25;
    protected static final double LEVEL_B = 0x1 << 24;
    protected static final double LEVEL_C = 0x1 << 23;
    protected static final double LEVEL_D = 0x1 << 22;
    protected static final double LEVEL_E = 0x1 << 21;
    protected static final double LEVEL_F = 0x1 << 20;
    protected static final double LEVEL_G = 0x1 << 19;
    protected static final double LEVEL_H = 0x1 << 18;
    protected static final double LEVEL_I = 0x1 << 17;
    protected static final double LEVEL_J = 0x1 << 16;
    protected static final double LEVEL_K = 0x1 << 15;
    protected static final double LEVEL_L = 0x1 << 14;
    protected static final double LEVEL_M = 0x1 << 13;
    protected static final double LEVEL_N = 0x1 << 12;
    protected static final double LEVEL_O = 0x1 << 11;
    protected static final double LEVEL_P = 0x1 << 10;
    protected static final LatLon GRID_1x1 = new LatLon(Angle.fromDegrees(180d), Angle.fromDegrees(360d));
    protected static final LatLon GRID_4x8 = new LatLon(Angle.fromDegrees(45d), Angle.fromDegrees(45d));
    protected static final LatLon GRID_8x16 = new LatLon(Angle.fromDegrees(22.5d), Angle.fromDegrees(22.5d));
    protected static final LatLon GRID_16x32 = new LatLon(Angle.fromDegrees(11.25d), Angle.fromDegrees(11.25d));
    protected static final LatLon GRID_36x72 = new LatLon(Angle.fromDegrees(5d), Angle.fromDegrees(5d));
    protected static final LatLon GRID_72x144 = new LatLon(Angle.fromDegrees(2.5d), Angle.fromDegrees(2.5d));
    protected static final LatLon GRID_144x288 = new LatLon(Angle.fromDegrees(1.25d), Angle.fromDegrees(1.25d));
    protected static final LatLon GRID_288x576 = new LatLon(Angle.fromDegrees(0.625d), Angle.fromDegrees(0.625d));

    private List<NavigationTile> navTiles = new ArrayList<NavigationTile>();    //top navigation tiles for each service

    /**
     * @param placeNameServiceSet the set of PlaceNameService objects that PlaceNameLayer will render.
     * @throws IllegalArgumentException if  {@link gov.nasa.worldwind.layers.placename.PlaceNameServiceSet} is null
     */
    public PlaceNameLayer(PlaceNameServiceSet placeNameServiceSet)
    {
        if (placeNameServiceSet == null)
        {
            String message = Logging.getMessage("nullValue.PlaceNameServiceSetIsNull");
            Logging.logger().fine(message);
            throw new IllegalArgumentException(message);
        }

        //  
        this.placeNameServiceSet = placeNameServiceSet.deepCopy();
        for (int i = 0; i < this.placeNameServiceSet.getServiceCount(); i++)
        {
            //todo do this for long as well and pick min
            int calc1 = (int) (PlaceNameService.TILING_SECTOR.getDeltaLatDegrees()/this.placeNameServiceSet.getService(i).getTileDelta().getLatitude().getDegrees());
            int numLevels=(int)Math.log(calc1);
            navTiles.add(new NavigationTile(this.placeNameServiceSet.getService(i), PlaceNameService.TILING_SECTOR, numLevels, "top"));
        }

        if (!WorldWind.getMemoryCacheSet().containsCache(Tile.class.getName()))
        {
            long size = Configuration.getLongValue(AVKey.PLACENAME_LAYER_CACHE_SIZE, 2000000L);
            MemoryCache cache = new BasicMemoryCache((long) (0.85 * size), size);
            cache.setName("Placename Tiles");
            WorldWind.getMemoryCacheSet().addCache(Tile.class.getName(), cache);
        }
    }

    public boolean isCullNames()
    {
        return cullNames;
    }

    public void setCullNames(boolean cullNames)
    {
        this.cullNames = cullNames;
        placeNameRenderer.setCullTextEnabled(cullNames);
    }

    public final PlaceNameServiceSet getPlaceNameServiceSet()
    {
        return this.placeNameServiceSet;
    }

    private PriorityBlockingQueue<Runnable> getRequestQ()
    {
        return this.requestQ;
    }

    // ============== Tile Assembly ======================= //
    // ============== Tile Assembly ======================= //
    // ============== Tile Assembly ======================= //

    private class NavigationTile
    {
        String id;
        private PlaceNameService placeNameService;
        public Sector navSector;
        private List<NavigationTile> subNavTiles= new ArrayList<NavigationTile>();
        private List<Integer> tileKeys = new ArrayList<Integer>();
        private int level;

        NavigationTile(PlaceNameService placeNameService, Sector sector, int levels, String id)
        {
            this.placeNameService = placeNameService;
            this.id=id;
            this.navSector = sector;
            level=levels;
        }

        private void buildSubNavTiles()
        {
            if (level > 0)
            {
               //split sector, create a navTile for each quad
                Sector[] subSectors=this.navSector.subdivide();
                for( int j=0; j<subSectors.length; j++)
                {
                     subNavTiles.add(new NavigationTile(placeNameService, subSectors[j], level-1, this.id+"."+j));
                }
            }
        }

        public List<NavigationTile> navTilesVisible(DrawContext dc, double minDistSquared, double maxDistSquared)
        {
            ArrayList<NavigationTile> navList=new ArrayList<NavigationTile>();
            if (this.isNavSectorVisible(dc, minDistSquared, maxDistSquared))
            {
                if ( this.level > 0 && !this.hasSubTiles())
                    this.buildSubNavTiles();

                if (this.hasSubTiles())
                {
                    for(NavigationTile nav: subNavTiles)
                    {
                        navList.addAll(nav.navTilesVisible(dc, minDistSquared, maxDistSquared));
                    }
                }else  //at bottom level navigation tile
                {
                    navList.add(this);
                }
            }

            return navList;
        }

        //Added for bulk download, does not use eye position to determine visibility
        //todo validate sector
        public List<NavigationTile> navTilesVisible(Sector sector)
        {
            ArrayList<NavigationTile> navList=new ArrayList<NavigationTile>();
            if (navSector.intersects(sector))
            {
                if ( this.level > 0 && !this.hasSubTiles())
                    this.buildSubNavTiles();

                if (this.hasSubTiles())
                {
                    for(NavigationTile nav: subNavTiles)
                    {
                        navList.addAll(nav.navTilesVisible(sector));
                    }
                }else  //at bottom level navigation tile
                {
                    navList.add(this);
                }
            }

            return navList;
        }

        private int estimateNumberTilesinSector(Sector searchSector)
        {
            final Angle dLat = placeNameService.getTileDelta().getLatitude();
            final Angle dLon = placeNameService.getTileDelta().getLongitude();

            // Determine the row and column offset from the global tiling origin for the southwest tile corner
            int firstRow = Tile.computeRow(dLat, this.navSector.getMinLatitude());
            int firstCol = Tile.computeColumn(dLon, this.navSector.getMinLongitude());
            int lastRow = Tile.computeRow(dLat, this.navSector.getMaxLatitude().subtract(dLat));
            int lastCol = Tile.computeColumn(dLon, this.navSector.getMaxLongitude().subtract(dLon));

            int tileCount=0;
            Angle p1 = Tile.computeRowLatitude(firstRow, dLat);
            boolean needToCheckDisk = true;
            for (int row = 0; row <= lastRow-firstRow; row++)
            {
                Angle p2;
                p2 = p1.add(dLat);

                Angle t1 = Tile.computeColumnLongitude(firstCol, dLon);
                for (int col = 0; col <= lastCol-firstCol; col++)
                {
                    Angle t2;
                    t2 = t1.add(dLon);
                    Sector tileSector = new Sector(p1, p2, t1, t2);

                    if (tileSector.intersects(searchSector))
                    {
                        if (needToCheckDisk)
                        {
                            //now check if on disk
                            String filePath = this.placeNameService.createFileCachePathFromTile(row+firstRow, col+firstCol);
                            final java.net.URL tileURL = WorldWind.getDataFileStore().findFile(filePath, false);
                            if (tileURL == null)
                                needToCheckDisk=false; //looked and found nothing
                            else
                                return 0;  //found one, assume rest are there
                        }

                        tileCount++;
                    }
                    t1 = t2;
                }
                p1 = p2;
            }

            return tileCount;
        }
        //end added for bulk download


        public List<NavigationTile> getSubNavTiles()
        {
            ArrayList<NavigationTile> navList=new ArrayList<NavigationTile>();
            for(NavigationTile nav: subNavTiles)
            {
                    navList.add(nav);
            }
            return navList;
        }

        public boolean hasSubTiles()
        {
            return !subNavTiles.isEmpty();
        }

        private boolean isNavSectorVisible(DrawContext dc, double minDistanceSquared, double maxDistanceSquared)
        {
            if (!navSector.intersects(dc.getVisibleSector()))
                return false;

            View view = dc.getView();
            Position eyePos = view.getEyePosition();
            if (eyePos == null)
                return false;

            //check for eyePos over globe
            if (Double.isNaN(eyePos.getLatitude().getDegrees()) || Double.isNaN(eyePos.getLongitude().getDegrees()))
                return false;

            Angle lat = clampAngle(eyePos.getLatitude(), navSector.getMinLatitude(),
                navSector.getMaxLatitude());
            Angle lon = clampAngle(eyePos.getLongitude(), navSector.getMinLongitude(),
                navSector.getMaxLongitude());
            Vec4 p = dc.getGlobe().computePointFromPosition(lat, lon, 0d);
            double distSquared = dc.getView().getEyePoint().distanceToSquared3(p);
            //noinspection RedundantIfStatement
            if (minDistanceSquared > distSquared || maxDistanceSquared < distSquared)
                return false;

            return true;
        }

        public List<Tile> getTiles()
        {
           if (tileKeys.isEmpty())
           {
               Tile[] tiles=buildTiles(this.placeNameService, this);
               //load tileKeys
               for (Tile t: tiles)
               {
                   tileKeys.add(t.getHashInt());
                   WorldWind.getMemoryCache(Tile.class.getName()).add(t.getHashInt(), t);
               }
               return Arrays.asList(tiles);
           }
           else
           {
               List<Tile> dataTiles=new ArrayList<Tile>();
               for ( Integer s : tileKeys )
               {
                   Tile t = (Tile) WorldWind.getMemoryCache(Tile.class.getName()).getObject(s);
                   if (t != null)
                   {
                       dataTiles.add(t);
                   }
               }
               return dataTiles;
           }
        }
    }

    private static class Tile implements Cacheable
    {
        final PlaceNameService placeNameService;
        final Sector sector;
        final int row;
        final int column;
        private Integer hashInt=null;
        // Computed data.
        String fileCachePath = null;
        double extentVerticalExaggeration = Double.MIN_VALUE;
        private Vec4 centroid; // Cartesian coordinate of lat/lon center
        private double priority = Double.MAX_VALUE; // Default is minimum priority
        private PlaceNameChunk dataChunk=null;

        Tile(PlaceNameService placeNameService, Sector sector, int row, int column)
        {
            this.placeNameService = placeNameService;
            this.sector = sector;
            this.row = row;
            this.column = column;
            this.fileCachePath = this.placeNameService.createFileCachePathFromTile(this.row, this.column);
            this.hashInt = this.computeHash();
        }

        public void setDataChunk(PlaceNameChunk chunk)
        {
            dataChunk = chunk;
        }

        public PlaceNameChunk getDataChunk()
        {
            return dataChunk;
        }

        public long getSizeInBytes()
        {

            long result = 32; //references

            result += this.getSector().getSizeInBytes();
            if (this.getFileCachePath() != null)
                result += this.getFileCachePath().length();

            if (dataChunk != null)
            {
                result += dataChunk.estimatedMemorySize;
            }

            return result;
        }

        static int computeRow(Angle delta, Angle latitude)
        {
            if (delta == null || latitude == null)
            {
                String msg = Logging.getMessage("nullValue.AngleIsNull");
                Logging.logger().severe(msg);
                throw new IllegalArgumentException(msg);
            }
            return (int) ((latitude.getDegrees() + 90d) / delta.getDegrees());
        }

        static int computeColumn(Angle delta, Angle longitude)
        {
            if (delta == null || longitude == null)
            {
                String msg = Logging.getMessage("nullValue.AngleIsNull");
                Logging.logger().severe(msg);
                throw new IllegalArgumentException(msg);
            }
            return (int) ((longitude.getDegrees() + 180d) / delta.getDegrees());
        }

        static Angle computeRowLatitude(int row, Angle delta)
        {
            if (delta == null)
            {
                String msg = Logging.getMessage("nullValue.AngleIsNull");
                Logging.logger().severe(msg);
                throw new IllegalArgumentException(msg);
            }
            return Angle.fromDegrees(-90d + delta.getDegrees() * row);
        }

        static Angle computeColumnLongitude(int column, Angle delta)
        {
            if (delta == null)
            {
                String msg = Logging.getMessage("nullValue.AngleIsNull");
                Logging.logger().severe(msg);
                throw new IllegalArgumentException(msg);
            }
            return Angle.fromDegrees(-180 + delta.getDegrees() * column);
        }

        public Integer getHashInt()
        {
            return hashInt;
        }

        int computeHash()
        {
            return this.getFileCachePath() != null ? this.getFileCachePath().hashCode() : 0;
        }

        @Override
        public boolean equals(Object o)
        {
            if (this == o)
                return true;
            if (o == null || getClass() != o.getClass())
                return false;

            final Tile tile = (Tile) o;

            return !(this.getFileCachePath() != null ? !this.getFileCachePath().equals(tile.getFileCachePath()) : tile.getFileCachePath() != null);
        }

        String getFileCachePath()
        {
            if (this.fileCachePath == null)
                this.fileCachePath = this.placeNameService.createFileCachePathFromTile(this.row, this.column);

            return this.fileCachePath;
        }


        PlaceNameService getPlaceNameService()
        {
            return placeNameService;
        }

        java.net.URL getRequestURL() throws java.net.MalformedURLException
        {
            return this.placeNameService.createServiceURLFromSector(this.sector);
        }


        Sector getSector()
        {
            return sector;
        }

        public int hashCode()
        {
            return this.hashInt;
        }

        boolean isTileInMemoryWithData()
        {
            //todo switched to string
            Tile t = (Tile) WorldWind.getMemoryCache(Tile.class.getName()).getObject(this.getHashInt());
            if (t ==null || t.getDataChunk() == null)
                return false;
            else
                return true;
        }

        public boolean isTileLocalOrAbsent()
        {
            if (this.getPlaceNameService().isResourceAbsent(
                this.getPlaceNameService().getTileNumber(this.row, this.column)))
                return true;    // tile is absent

            URL url = WorldWind.getDataFileStore().findFile(this.getFileCachePath(), false);
            return url != null; // tile is already in cache
        }

        public Vec4 getCentroidPoint(Globe globe)
        {
            if (globe == null)
            {
                String msg = Logging.getMessage("nullValue.GlobeIsNull");
                Logging.logger().severe(msg);
                throw new IllegalArgumentException(msg);
            }

            if (this.centroid == null)
            {
                LatLon c = this.getSector().getCentroid();
                this.centroid = globe.computePointFromPosition(c.getLatitude(), c.getLongitude(), 0);
            }

            return this.centroid;
        }

        public double getPriority()
        {
            return priority;
        }

        public void setPriority(double priority)
        {
            this.priority = priority;
        }
    }

    private Tile[] buildTiles(PlaceNameService placeNameService, NavigationTile navTile)
    {
        final Angle dLat = placeNameService.getTileDelta().getLatitude();
        final Angle dLon = placeNameService.getTileDelta().getLongitude();

        // Determine the row and column offset from the global tiling origin for the southwest tile corner
        int firstRow = Tile.computeRow(dLat, navTile.navSector.getMinLatitude());
        int firstCol = Tile.computeColumn(dLon, navTile.navSector.getMinLongitude());
        int lastRow = Tile.computeRow(dLat, navTile.navSector.getMaxLatitude().subtract(dLat));
        int lastCol = Tile.computeColumn(dLon, navTile.navSector.getMaxLongitude().subtract(dLon));

        int nLatTiles = lastRow - firstRow + 1;
        int nLonTiles = lastCol - firstCol + 1;

        Tile[] tiles = new Tile[nLatTiles * nLonTiles];

        Angle p1 = Tile.computeRowLatitude(firstRow, dLat);
        for (int row = 0; row <= lastRow-firstRow; row++)
        {
            Angle p2;
            p2 = p1.add(dLat);

            Angle t1 = Tile.computeColumnLongitude(firstCol, dLon);
            for (int col = 0; col <= lastCol-firstCol; col++)
            {
                Angle t2;
                t2 = t1.add(dLon);
                //Need offset row and column to correspond to total ro/col numbering
                tiles[col + row * nLonTiles] = new Tile(placeNameService, new Sector(p1, p2, t1, t2), row+firstRow, col+firstCol);
                t1 = t2;
            }
            p1 = p2;
        }

        return tiles;
    }
    
    // ============== Place Name Data Structures ======================= //
    // ============== Place Name Data Structures ======================= //
    // ============== Place Name Data Structures ======================= //

    private static class PlaceNameChunk implements Cacheable
    {
        final PlaceNameService placeNameService;
        final CharBuffer textArray;
        final int[] textIndexArray;
        final double[] latlonArray;
        final int numEntries;
        final long estimatedMemorySize;

        PlaceNameChunk(PlaceNameService service, CharBuffer text, int[] textIndices,
            double[] positions, int numEntries)
        {
            this.placeNameService = service;
            this.textArray = text;
            this.textIndexArray = textIndices;
            this.latlonArray = positions;
            this.numEntries = numEntries;
            this.estimatedMemorySize = this.computeEstimatedMemorySize();
        }

        long computeEstimatedMemorySize()
        {
            long result = 0;
            if (!textArray.isDirect())
                result += (Character.SIZE / 8) * textArray.capacity();
            result += (Integer.SIZE / 8) * textIndexArray.length;
            result += (Double.SIZE / 8) * latlonArray.length;
            return result;
        }

        Position getPosition(int index)
        {
            int latlonIndex = 2 * index;
            return Position.fromDegrees(latlonArray[latlonIndex], latlonArray[latlonIndex + 1], 0);
        }

        PlaceNameService getPlaceNameService()
        {
            return this.placeNameService;
        }

        CharSequence getText(int index)
        {
            int beginIndex = textIndexArray[index];
            int endIndex = (index + 1 < numEntries) ? textIndexArray[index + 1] : textArray.length();
            return this.textArray.subSequence(beginIndex, endIndex);
        }

        public long getSizeInBytes()
        {
            return this.estimatedMemorySize;
        }

        private Iterable<GeographicText> makeIterable(DrawContext dc)
        {
            //get dispay dist for this service for use in label annealing
            double maxDisplayDistance = this.getPlaceNameService().getMaxDisplayDistance();
            ArrayList<GeographicText> list = new ArrayList<GeographicText>();
            for (int i = 0; i < this.numEntries; i++)
            {
                CharSequence str = getText(i);
                Position pos = getPosition(i);
                GeographicText text = new UserFacingText(str, pos);
                text.setFont(this.placeNameService.getFont());
                text.setColor(this.placeNameService.getColor());
                text.setBackgroundColor(this.placeNameService.getBackgroundColor());
                text.setVisible(isNameVisible(dc, this.placeNameService, pos));
                text.setPriority(maxDisplayDistance);
                list.add(text);
            }
            return list;
        }
    }

    // ============== Rendering ======================= //
    // ============== Rendering ======================= //
    // ============== Rendering ======================= //

    private final GeographicTextRenderer placeNameRenderer = new GeographicTextRenderer();

    @Override
    protected void doRender(DrawContext dc)
    {

        this.referencePoint = this.computeReferencePoint(dc);

        int serviceCount = this.placeNameServiceSet.getServiceCount();
        for (int i = 0; i < serviceCount; i++)
        {
            PlaceNameService placeNameService = this.placeNameServiceSet.getService(i);
            if (!isServiceVisible(dc, placeNameService))
                continue;

            double minDistSquared = placeNameService.getMinDisplayDistance() * placeNameService.getMinDisplayDistance();
            double maxDistSquared = placeNameService.getMaxDisplayDistance() * placeNameService.getMaxDisplayDistance();

            if (isSectorVisible(dc, placeNameService.getMaskingSector(), minDistSquared, maxDistSquared))
            {
                    ArrayList<Tile> baseTiles = new ArrayList<Tile>();
                    NavigationTile navTile = this.navTiles.get(i);
                    //drill down into tiles to find bottom level navTiles visible
                    List<NavigationTile> list=navTile.navTilesVisible(dc, minDistSquared, maxDistSquared);
                    for(NavigationTile nt: list)
                    {
                        baseTiles.addAll(nt.getTiles());
                    }

                    for (Tile tile : baseTiles)
                    {
                        try
                        {
                            drawOrRequestTile(dc, tile, minDistSquared, maxDistSquared);
                        }
                        catch (Exception e)
                        {
                            Logging.logger().log(Level.FINE, Logging.getMessage("layers.PlaceNameLayer.ExceptionRenderingTile"),
                                e);
                        }
                    }
            }
        }

        this.sendRequests();
        this.requestQ.clear();
    }

    private Vec4 computeReferencePoint(DrawContext dc)
    {
        if (dc.getViewportCenterPosition() != null)
            return dc.getGlobe().computePointFromPosition(dc.getViewportCenterPosition());

        java.awt.geom.Rectangle2D viewport = dc.getView().getViewport();
        int x = (int) viewport.getWidth() / 2;
        for (int y = (int) (0.5 * viewport.getHeight()); y >= 0; y--)
        {
            Position pos = dc.getView().computePositionFromScreenPoint(x, y);
            if (pos == null)
                continue;

            return dc.getGlobe().computePointFromPosition(pos.getLatitude(), pos.getLongitude(), 0d);
        }

        return null;
    }

    protected Vec4 getReferencePoint()
    {
        return this.referencePoint;
    }

    private void drawOrRequestTile(DrawContext dc, Tile tile, double minDisplayDistanceSquared,
        double maxDisplayDistanceSquared)
    {
        if (!isTileVisible(dc, tile, minDisplayDistanceSquared, maxDisplayDistanceSquared))
            return;
        
        if (tile.isTileInMemoryWithData())
        {
            PlaceNameChunk placeNameChunk = tile.getDataChunk();
            if (placeNameChunk.numEntries > 0)
            {
                Iterable<GeographicText> renderIter = placeNameChunk.makeIterable(dc);
                this.placeNameRenderer.render(dc, renderIter);
            }
            return;
        }

        // Tile's data isn't available, so request it
        if (!tile.getPlaceNameService().isResourceAbsent(tile.getPlaceNameService().getTileNumber(
                tile.row, tile.column)))
        {
            this.requestTile(dc, tile);
        }
    }

    private static boolean isServiceVisible(DrawContext dc, PlaceNameService placeNameService)
    {
        if (!placeNameService.isEnabled())
            return false;
        //noinspection SimplifiableIfStatement
        if (dc.getVisibleSector() != null && !placeNameService.getMaskingSector().intersects(dc.getVisibleSector()))
            return false;

        return placeNameService.getExtent(dc).intersects(dc.getView().getFrustumInModelCoordinates());
    }


    private static boolean isSectorVisible(DrawContext dc, Sector sector, double minDistanceSquared,
        double maxDistanceSquared)
    {

        View view = dc.getView();
        Position eyePos = view.getEyePosition();
        if (eyePos == null)
            return false;

        Angle lat = clampAngle(eyePos.getLatitude(), sector.getMinLatitude(), sector.getMaxLatitude());
        Angle lon = clampAngle(eyePos.getLongitude(), sector.getMinLongitude(), sector.getMaxLongitude());
        Vec4 p = dc.getGlobe().computePointFromPosition(lat, lon, 0d);
        double distSquared = dc.getView().getEyePoint().distanceToSquared3(p);
        //noinspection RedundantIfStatement
        if (minDistanceSquared > distSquared || maxDistanceSquared < distSquared)
            return false;

        return true;
    }

    private static boolean isTileVisible(DrawContext dc, Tile tile, double minDistanceSquared,
        double maxDistanceSquared)
    {
        if (!tile.getSector().intersects(dc.getVisibleSector()))
            return false;

        View view = dc.getView();
        Position eyePos = view.getEyePosition();
        if (eyePos == null)
            return false;

        Angle lat = clampAngle(eyePos.getLatitude(), tile.getSector().getMinLatitude(),
            tile.getSector().getMaxLatitude());
        Angle lon = clampAngle(eyePos.getLongitude(), tile.getSector().getMinLongitude(),
            tile.getSector().getMaxLongitude());
        Vec4 p = dc.getGlobe().computePointFromPosition(lat, lon, 0d);
        double distSquared = dc.getView().getEyePoint().distanceToSquared3(p);
        //noinspection RedundantIfStatement
        if (minDistanceSquared > distSquared || maxDistanceSquared < distSquared)
            return false;

        return true;
    }

    private static boolean isNameVisible(DrawContext dc, PlaceNameService service, Position namePosition)
    {
        double elevation = dc.getVerticalExaggeration() * namePosition.getElevation();
        Vec4 namePoint = dc.getGlobe().computePointFromPosition(namePosition.getLatitude(),
            namePosition.getLongitude(), elevation);
        Vec4 eyeVec = dc.getView().getEyePoint();

        double dist = eyeVec.distanceTo3(namePoint);
        return dist >= service.getMinDisplayDistance() && dist <= service.getMaxDisplayDistance();
    }

    private static Angle clampAngle(Angle a, Angle min, Angle max)
    {
        double degrees = a.degrees;
        double minDegrees = min.degrees;
        double maxDegrees = max.degrees;
        return Angle.fromDegrees(degrees < minDegrees ? minDegrees : (degrees > maxDegrees ? maxDegrees : degrees));
    }

    // ============== Image Reading and Downloading ======================= //
    // ============== Image Reading and Downloading ======================= //
    // ============== Image Reading and Downloading ======================= //

    private void requestTile(DrawContext dc, Tile tile)
    {
        Vec4 centroid = tile.getCentroidPoint(dc.getGlobe());
        if (this.getReferencePoint() != null)
            tile.setPriority(centroid.distanceTo3(this.getReferencePoint()));

        RequestTask task = new RequestTask(tile, this);
        this.getRequestQ().add(task);
    }

    private void sendRequests()
    {
        Runnable task = this.requestQ.poll();
        while (task != null)
        {
            if (!WorldWind.getTaskService().isFull())
            {
                WorldWind.getTaskService().addTask(task);
            }
            task = this.requestQ.poll();
        }
    }

    private static class RequestTask implements Runnable, Comparable<RequestTask>
    {
        private final PlaceNameLayer layer;
        private final Tile tile;

        RequestTask(Tile tile, PlaceNameLayer layer)
        {
            this.layer = layer;
            this.tile = tile;
        }

        public void run()
        {
            if (this.tile.isTileInMemoryWithData())
                return;

            final java.net.URL tileURL = WorldWind.getDataFileStore().findFile(tile.getFileCachePath(), false);
            if (tileURL != null)
            {
                if (this.layer.loadTile(this.tile, tileURL))
                {
                    tile.getPlaceNameService().unmarkResourceAbsent(tile.getPlaceNameService().getTileNumber(
                        tile.row,
                        tile.column));
                    this.layer.firePropertyChange(AVKey.LAYER, null, this);
                    return;
                }
                else
                {
                    // Assume that something's wrong with the file and delete it.
                    WorldWind.getDataFileStore().removeFile(tileURL);
                    tile.getPlaceNameService().markResourceAbsent(tile.getPlaceNameService().getTileNumber(tile.row,
                        tile.column));
                    String message = Logging.getMessage("generic.DeletedCorruptDataFile", tileURL);
                    Logging.logger().fine(message);
                }
            }

            this.layer.downloadTile(this.tile);
        }

        /**
         * @param that the task to compare
         * @return -1 if <code>this</code> less than <code>that</code>, 1 if greater than, 0 if equal
         * @throws IllegalArgumentException if <code>that</code> is null
         */
        public int compareTo(RequestTask that)
        {
            if (that == null)
            {
                String msg = Logging.getMessage("nullValue.RequestTaskIsNull");
                Logging.logger().severe(msg);
                throw new IllegalArgumentException(msg);
            }
            return this.tile.getPriority() == that.tile.getPriority() ? 0 :
                this.tile.getPriority() < that.tile.getPriority() ? -1 : 1;
        }

        public boolean equals(Object o)
        {
            if (this == o)
                return true;
            if (o == null || getClass() != o.getClass())
                return false;

            final RequestTask that = (RequestTask) o;

            // Don't include layer in comparison so that requests are shared among layers
            return !(tile != null ? !tile.equals(that.tile) : that.tile != null);
        }

        public int hashCode()
        {
            return (tile != null ? tile.hashCode() : 0);
        }

        public String toString()
        {
            return this.tile.toString();
        }
    }

    private boolean loadTile(Tile tile, java.net.URL url)
    {
        if (WWIO.isFileOutOfDate(url, this.placeNameServiceSet.getExpiryTime()))
        {
            // The file has expired. Delete it then request download of newer.
            WorldWind.getDataFileStore().removeFile(url);
            String message = Logging.getMessage("generic.DataFileExpired", url);
            Logging.logger().fine(message);
            return false;
        }

        PlaceNameChunk tileData;
        synchronized (this.fileLock)
        {
            tileData = readTileData(tile, url);
        }

        if (tileData == null)
            return false;

        tile.setDataChunk(tileData);
        WorldWind.getMemoryCache(Tile.class.getName()).add(tile.getHashInt(), tile);
        return true;
    }

    private static PlaceNameChunk readTileData(Tile tile, java.net.URL url)
    {
        java.io.InputStream is = null;

        try
        {
            String path = url.getFile();
            path = path.replaceAll("%20", " "); // TODO: find a better way to get a path usable by FileInputStream

            java.io.FileInputStream fis = new java.io.FileInputStream(path);
            java.io.BufferedInputStream buf = new java.io.BufferedInputStream(fis);
            is = new java.util.zip.GZIPInputStream(buf);

            GMLPlaceNameSAXHandler handler = new GMLPlaceNameSAXHandler();
            javax.xml.parsers.SAXParserFactory.newInstance().newSAXParser().parse(is, handler);
            return handler.createPlaceNameChunk(tile.getPlaceNameService());
        }
        catch (Exception e)
        {
            //todo log actual error 
            Logging.logger().log(Level.FINE,
                    Logging.getMessage("layers.PlaceNameLayer.ExceptionAttemptingToReadFile", url.toString()), e);
        }
        finally
        {
            try
            {
                if (is != null)
                    is.close();
            }
            catch (java.io.IOException e)
            {
                Logging.logger().log(Level.FINE,
                        Logging.getMessage("layers.PlaceNameLayer.ExceptionAttemptingToReadFile", url.toString()), e);
            }
        }

        return null;
    }

    private static CharBuffer newCharBuffer(int numElements)
    {
        ByteBuffer bb = ByteBuffer.allocateDirect((Character.SIZE / 8) * numElements);
        bb.order(ByteOrder.nativeOrder());
        return bb.asCharBuffer();
    }

    private static class GMLPlaceNameSAXHandler extends org.xml.sax.helpers.DefaultHandler
    {
        static final String GML_FEATURE_MEMBER = "gml:featureMember";
        static final String TOPP_FULL_NAME_ND = "topp:full_name_nd";
        static final String TOPP_LATITUDE = "topp:latitude";
        static final String TOPP_LONGITUDE = "topp:longitude";
        final LinkedList<String> internedQNameStack = new LinkedList<String>();
        boolean inBeginEndPair = false;
        StringBuilder latBuffer = new StringBuilder();
        StringBuilder lonBuffer = new StringBuilder();

        StringBuilder textArray = new StringBuilder();
        int[] textIndexArray = new int[16];
        double[] latlonArray = new double[16];
        int numEntries = 0;

        GMLPlaceNameSAXHandler()
        {
        }

        PlaceNameChunk createPlaceNameChunk(PlaceNameService service)
        {
            int numChars = this.textArray.length();
            CharBuffer textBuffer = newCharBuffer(numChars);
            textBuffer.put(this.textArray.toString());
            textBuffer.rewind();
            return new PlaceNameChunk(service, textBuffer, this.textIndexArray, this.latlonArray, this.numEntries);
        }

        void beginEntry()
        {
            int textIndex = this.textArray.length();
            this.textIndexArray = append(this.textIndexArray, this.numEntries, textIndex);
            this.inBeginEndPair = true;
        }

        void endEntry()
        {
            double lat = this.parseDouble(this.latBuffer);
            double lon = this.parseDouble(this.lonBuffer);
            int numLatLon = 2 * this.numEntries;
            this.latlonArray = this.append(this.latlonArray, numLatLon, lat);
            numLatLon++;
            this.latlonArray = this.append(this.latlonArray, numLatLon, lon);

            this.latBuffer.delete(0, this.latBuffer.length());
            this.lonBuffer.delete(0, this.lonBuffer.length());
            this.inBeginEndPair = false;
            this.numEntries++;
        }

        double parseDouble(StringBuilder sb)
        {
            double value = 0;
            try
            {
                value = Double.parseDouble(sb.toString());
            }
            catch (NumberFormatException e)
            {
                Logging.logger().log(Level.FINE,
                        Logging.getMessage("layers.PlaceNameLayer.ExceptionAttemptingToReadFile", ""), e);
            }
            return value;
        }

        int[] append(int[] array, int index, int value)
        {
            if (index >= array.length)
                array = this.resizeArray(array);
            array[index] = value;
            return array;
        }

        int[] resizeArray(int[] oldArray)
        {
            int newSize = 2 * oldArray.length;
            int[] newArray = new int[newSize];
            System.arraycopy(oldArray, 0, newArray, 0, oldArray.length);
            return newArray;
        }

        double[] append(double[] array, int index, double value)
        {
            if (index >= array.length)
                array = this.resizeArray(array);
            array[index] = value;
            return array;
        }

        double[] resizeArray(double[] oldArray)
        {
            int newSize = 2 * oldArray.length;
            double[] newArray = new double[newSize];
            System.arraycopy(oldArray, 0, newArray, 0, oldArray.length);
            return newArray;
        }

        @SuppressWarnings({"StringEquality"})
        public void characters(char ch[], int start, int length)
        {
            if (!this.inBeginEndPair)
                return;

            // Top of QName stack is an interned string,
            // so we can use pointer comparison.
            String internedTopQName = this.internedQNameStack.getFirst();

            StringBuilder sb = null;
            if (TOPP_LATITUDE == internedTopQName)
                sb = this.latBuffer;
            else if (TOPP_LONGITUDE == internedTopQName)
                sb = this.lonBuffer;
            else if (TOPP_FULL_NAME_ND == internedTopQName)
                sb = this.textArray;

            if (sb != null)
                sb.append(ch, start, length);
        }

        public void startElement(String uri, String localName, String qName, org.xml.sax.Attributes attributes)
        {
            // Don't validate uri, localName or attributes because they aren't used.
            // Intern the qName string so we can use pointer comparison.
            String internedQName = qName.intern();
            //noinspection StringEquality
            if (GML_FEATURE_MEMBER == internedQName)
                this.beginEntry();
            this.internedQNameStack.addFirst(internedQName);
        }

        public void endElement(String uri, String localName, String qName)
        {
            // Don't validate uri or localName because they aren't used.
            // Intern the qName string so we can use pointer comparison.
            String internedQName = qName.intern();
            //noinspection StringEquality
            if (GML_FEATURE_MEMBER == internedQName)
                this.endEntry();
            this.internedQNameStack.removeFirst();
        }
    }

    private void downloadTile(final Tile tile)
    {
        downloadTile(tile, null);
    }

    private void downloadTile(final Tile tile, DownloadPostProcessor postProcessor)
    {
        if (!this.isNetworkRetrievalEnabled())
            return;
        
        if (!WorldWind.getRetrievalService().isAvailable())
            return;

        java.net.URL url;
        try
        {
            url = tile.getRequestURL();
            if (WorldWind.getNetworkStatus().isHostUnavailable(url))
                return;
        }
        catch (java.net.MalformedURLException e)
        {
            Logging.logger().log(java.util.logging.Level.SEVERE,
                    Logging.getMessage("layers.PlaceNameLayer.ExceptionCreatingUrl", tile), e);
            return;
        }

        Retriever retriever;

        if ("http".equalsIgnoreCase(url.getProtocol()) || "https".equalsIgnoreCase(url.getProtocol()))
        {
            if (postProcessor == null)
                postProcessor = new DownloadPostProcessor(this, tile);
            retriever = new HTTPRetriever(url, postProcessor);
        }
        else
        {
            Logging.logger().severe(
                    Logging.getMessage("layers.PlaceNameLayer.UnknownRetrievalProtocol", url.toString()));
            return;
        }

        // Apply any overridden timeouts.
        Integer cto = AVListImpl.getIntegerValue(this, AVKey.URL_CONNECT_TIMEOUT);
        if (cto != null && cto > 0)
            retriever.setConnectTimeout(cto);
        Integer cro = AVListImpl.getIntegerValue(this, AVKey.URL_READ_TIMEOUT);
        if (cro != null && cro > 0)
            retriever.setReadTimeout(cro);
        Integer srl = AVListImpl.getIntegerValue(this, AVKey.RETRIEVAL_QUEUE_STALE_REQUEST_LIMIT);
        if (srl != null && srl > 0)
            retriever.setStaleRequestLimit(srl);

        WorldWind.getRetrievalService().runRetriever(retriever, tile.getPriority());
    }

    private void saveBuffer(java.nio.ByteBuffer buffer, java.io.File outFile) throws java.io.IOException
    {
        synchronized (this.fileLock) // sychronized with read of file in RequestTask.run()
        {
            WWIO.saveBuffer(buffer, outFile);
        }
    }

    private static class DownloadPostProcessor implements RetrievalPostProcessor
    {
        final PlaceNameLayer layer;
        final Tile tile;

        private DownloadPostProcessor(PlaceNameLayer layer, Tile tile)
        {
            this.layer = layer;
            this.tile = tile;
        }

        public java.nio.ByteBuffer run(Retriever retriever)
        {
            if (retriever == null)
            {
                String msg = Logging.getMessage("nullValue.RetrieverIsNull");
                Logging.logger().fine(msg);
                throw new IllegalArgumentException(msg);
            }

            try
            {
                if (!retriever.getState().equals(Retriever.RETRIEVER_STATE_SUCCESSFUL))
                    return null;

                URLRetriever r = (URLRetriever) retriever;
                ByteBuffer buffer = r.getBuffer();

                if (retriever instanceof HTTPRetriever)
                {
                    HTTPRetriever htr = (HTTPRetriever) retriever;
                    if (htr.getResponseCode() == java.net.HttpURLConnection.HTTP_NO_CONTENT)
                    {
                        // Mark tile as missing to avoid further attempts
                        tile.getPlaceNameService().markResourceAbsent(tile.getPlaceNameService().getTileNumber(tile.row,
                            tile.column));
                        return null;
                    }
                    else if (htr.getResponseCode() != java.net.HttpURLConnection.HTTP_OK)
                    {
                        // Also mark tile as missing, but for an unknown reason.
                        tile.getPlaceNameService().markResourceAbsent(tile.getPlaceNameService().getTileNumber(tile.row,
                            tile.column));
                        return null;
                    }
                }

                final java.io.File outFile = WorldWind.getDataFileStore().newFile(this.tile.getFileCachePath());
                if (outFile == null)
                    return null;

                if (outFile.exists())
                    return buffer; // info is already here; don't need to do anything

                if (buffer != null)
                {
                    String contentType = retriever.getContentType();
                    if (contentType == null)
                    {
                        // TODO: logger message
                        return null;
                    }

                    this.layer.saveBuffer(buffer, outFile);
                    this.layer.firePropertyChange(AVKey.LAYER, null, this);
                    return buffer;
                }
            }
            catch (ClosedByInterruptException e)
            {
                Logging.logger().log(java.util.logging.Level.FINE,
                    Logging.getMessage("generic.OperationCancelled", "placename retrieval"), e);
            }
            catch (java.io.IOException e)
            {
                tile.getPlaceNameService().markResourceAbsent(tile.getPlaceNameService().getTileNumber(tile.row,
                    tile.column));
                Logging.logger().log(Level.FINE, Logging.getMessage(
                    "layers.PlaceNameLayer.ExceptionSavingRetrievedFile", this.tile.getFileCachePath()), e);
            }
            
            return null;
        }
    }

    // *** Bulk download ***
    // *** Bulk download ***
    // *** Bulk download ***
    private static final long AVG_TILE_SIZE = 8*1024;
    /**
     * Start a new {@link gov.nasa.worldwind.retrieve.BulkRetrievalThread} that will try to download all place name
     * tiles for a given {@link Sector} and resolution. Note that the target resolution is ignored right now.
     *
     * @param sector the {@link Sector} to download tiles for.
     * @param resolution the target resolution - ignored.
     * @return the {@link gov.nasa.worldwind.retrieve.BulkRetrievalThread} that executes the retrieval.
     */
    public BulkRetrievalThread makeLocal(Sector sector, double resolution)
    {
        BulkTileDownloadThread thread = new BulkTileDownloadThread(this, sector, resolution);
        thread.setDaemon(true);
        thread.start();
        return thread;
    }

    /**
     * Get the estimated size in byte of the missing data for the given {@link Sector}
     * and resolution. Note that the target resolution must be provided in radian latitude per
     * data sample - which is the resolution in meter divided by the globe radius.
     *
     * @param sector the {@link Sector} to estimate.
     * @param resolution the target resolution provided in radian latitude per texel.
     * @return the estimated size in byte of the missing imagery.
     */
    public long getEstimatedMissingDataSize(Sector sector, double resolution)
    {
        int tileCount;
        try
        {
            tileCount=this.getMissingTilesCountEstimate(sector, resolution);
        }
        catch (Exception e)
        {
            String message = Logging.getMessage("generic.ExceptionDuringDataSizeEstimate", this.getName());
            Logging.logger().severe(message);
            throw new RuntimeException(message);
        }
        return tileCount * AVG_TILE_SIZE;
    }

    private class BulkTileDownloadThread extends BulkRetrievalThread
    {
        private int MAX_TILE_COUNT_PER_REGION = 200;

        private final PlaceNameLayer layer;
        private ArrayList<Tile> missingTiles;

        public BulkTileDownloadThread(PlaceNameLayer layer, Sector sector, double resolution)
        {
            //resolution is compared to the maxDsiatnce value in each placenameservice
            super(layer, sector, resolution);
            this.layer = layer;
        }

        public void run()
        {
            try
            {
                // Cycle though placenameservices and find missing tiles
                missingTiles = new ArrayList<Tile>();
                ArrayList<Tile> allMissingTiles =layer.getMissingTilesInSector(this.sector, this.resolution);

                this.progress.setTotalCount(allMissingTiles.size());
                // Submit missing tiles requests at 10 sec intervals
                while (allMissingTiles.size() > 0)
                {
                    transferMissingTiles(allMissingTiles, missingTiles, MAX_TILE_COUNT_PER_REGION);

                    while (missingTiles.size() > 0)
                    {
                        submitMissingTilesRequests();
                        if (missingTiles.size() > 0)
                            Thread.sleep(RETRIEVAL_SERVICE_POLL_DELAY);
                    }
                }
            }
            catch (InterruptedException e)
            {
                String message = Logging.getMessage("generic.BulkRetrievalInterrupted", layer.getName());
                Logging.logger().log(java.util.logging.Level.WARNING, message, e);
            }
            catch (Exception e)
            {
                String message = Logging.getMessage("generic.ExceptionDuringBulkRetrieval", layer.getName());
                Logging.logger().severe(message);
                throw new RuntimeException(message);
            }
        }

        private void transferMissingTiles(ArrayList<Tile> source, ArrayList<Tile> destination, int maxCount)
        {
            int i = 0;
            while (i < maxCount && source.size() > 0)
            {
                destination.add(source.remove(0));
                i++;
            }
        }

        private synchronized void submitMissingTilesRequests() throws InterruptedException
        {
            RetrievalService rs = WorldWind.getRetrievalService();
            int i = 0;
            while (this.missingTiles.size() > i && rs.isAvailable())
            {
                Thread.sleep(1); // generates InterruptedException if thread has been interrupted

                Tile tile = this.missingTiles.get(i);
                if (tile.isTileLocalOrAbsent())
                {
                    // No need to request that tile anymore
                    this.missingTiles.remove(i);
                }
                else
                {
                    layer.downloadTile(tile, new BulkDownloadPostProcessor(layer, tile));
                    i++;
                }
            }
        }

        private class BulkDownloadPostProcessor extends DownloadPostProcessor
        {
            public BulkDownloadPostProcessor(PlaceNameLayer layer, Tile tile)
            {
                super(layer, tile);
            }

            public ByteBuffer run(Retriever retriever)
            {
                ByteBuffer buffer = super.run(retriever);
                if (buffer != null)
                    removeRetrievedTile(this.tile);

                return buffer;
            }
        }

        private synchronized void removeRetrievedTile(Tile tile)
        {
            this.missingTiles.remove(tile);
            this.progress.setCurrentCount(this.progress.getCurrentCount() + 1);
            this.progress.setCurrentSize(this.progress.getCurrentSize() + AVG_TILE_SIZE);
            this.progress.setLastUpdateTime(System.currentTimeMillis());
            // Estimate total size
            this.progress.setTotalSize(
            this.progress.getCurrentSize() / this.progress.getCurrentCount() * this.progress.getTotalCount());
        }
    }

    private int getMissingTilesCountEstimate(Sector sector, double resolution)
    {
        int tileCount=0;
        int serviceCount = this.placeNameServiceSet.getServiceCount();
        for (int i = 0; i < serviceCount; i++)
        {
            int serviceTileCount=0;
            PlaceNameService service = this.getPlaceNameServiceSet().getService(i);
             if (service.getMaxDisplayDistance() > resolution)
             {
                 NavigationTile navTile = this.navTiles.get(i);
                 // drill down into tiles to find bottom level navTiles visible
                 List<NavigationTile> list = navTile.navTilesVisible(sector);
                 for(NavigationTile nt: list)
                 {
                     serviceTileCount +=  nt.estimateNumberTilesinSector(sector);
                 }
            }

            tileCount += serviceTileCount;
        }

        return tileCount;
    }
    
    private ArrayList<Tile> getMissingTilesInSector(Sector sector, double resolution) throws InterruptedException
    {
        ArrayList<Tile> allMissingTiles =new ArrayList<Tile>();
        int serviceCount = this.placeNameServiceSet.getServiceCount();
        for (int i = 0; i < serviceCount; i++)
        {
             PlaceNameService service = this.getPlaceNameServiceSet().getService(i);
             if (service.getMaxDisplayDistance() > resolution)
             {
                 // get tiles in sector
                 ArrayList<Tile> baseTiles = new ArrayList<Tile>();

                 NavigationTile navTile = this.navTiles.get(i);
                 // drill down into tiles to find bottom level navTiles visible
                 List<NavigationTile> list = navTile.navTilesVisible(sector);
                 for(NavigationTile nt: list)
                 {
                     baseTiles.addAll(nt.getTiles());
                 }

                for (Tile tile : baseTiles)
                {
                    if ((tile.getSector().intersects(sector)) && (!tile.isTileLocalOrAbsent()))
                        allMissingTiles.add(tile);
                }
            }
        }

        return allMissingTiles;
    }
}
