/* Copyright (C) 2001, 2009 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.render;

import gov.nasa.worldwind.Movable;
import gov.nasa.worldwind.avlist.AVKey;
import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.globes.Globe;
import gov.nasa.worldwind.util.*;
import gov.nasa.worldwind.util.measure.AreaMeasurer;

import javax.media.opengl.GL;

/**
 * @author dcollins
 * @version $Id: AbstractSurfaceShape.java 11457 2009-06-05 07:09:37Z dcollins $
 */
public abstract class AbstractSurfaceShape extends AbstractSurfaceObject implements SurfaceShape, Movable
{
    protected static final String DEFAULT_PATH_TYPE = AVKey.GREAT_CIRCLE;
    protected static final int DEFAULT_TEXELS_PER_EDGE_INTERVAL = 50;
    protected static final int DEFAULT_MIN_EDGE_INTERVALS = 0;
    protected static final int DEFAULT_MAX_EDGE_INTERVALS = 100;
    protected static final int DEFAULT_CACHE_CAPACITY = 16;

    private ShapeAttributes attributes;
    private java.beans.PropertyChangeListener attributeChangeListener;
    protected String pathType;
    protected double texelsPerEdgeInterval;
    protected int minEdgeIntervals;
    protected int maxEdgeIntervals;
    protected boolean showBoundingSectors = false;
    // Shape rendering components.
    private static SurfaceShapeSupport surfaceShapeSupport;
    protected TiledSurfaceObjectRenderer renderer;
    protected java.util.List<Vec4> vertices;
    // Cached computations.
    protected BoundedHashMap<Object, CacheEntry<Iterable<? extends Sector>>> sectorCache;
    protected BoundedHashMap<Object, CacheEntry<java.util.List<LatLon>>> drawLocationCache;
    protected AreaMeasurer areaMeasurer;
    protected long areaMeasurerLastModifiedTime;

    public AbstractSurfaceShape(ShapeAttributes attributes)
    {
        if (attributes == null)
        {
            String message = Logging.getMessage("nullValue.AttributesIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.setAttributes(attributes);
        this.setPathType(DEFAULT_PATH_TYPE);
        this.setTexelsPerEdgeInterval(DEFAULT_TEXELS_PER_EDGE_INTERVAL);
        this.setMinAndMaxEdgeIntervals(DEFAULT_MIN_EDGE_INTERVALS, DEFAULT_MAX_EDGE_INTERVALS);

        this.sectorCache = new BoundedHashMap<Object, CacheEntry<Iterable<? extends Sector>>>(DEFAULT_CACHE_CAPACITY,
            true);
        this.drawLocationCache = new BoundedHashMap<Object, CacheEntry<java.util.List<LatLon>>>(DEFAULT_CACHE_CAPACITY,
            true);
    }

    public AbstractSurfaceShape()
    {
        this(new BasicShapeAttributes());
    }

    public ShapeAttributes getAttributes()
    {
        return this.attributes;
    }

    public void setAttributes(ShapeAttributes attributes)
    {
        if (attributes == null)
        {
            String message = Logging.getMessage("nullValue.AttributesIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (this.attributes != null)
        {
            if (this.attributeChangeListener != null)
            {
                this.attributes.removePropertyChangeListener(this.attributeChangeListener);
            }
        }

        this.attributes = attributes;

        //noinspection ConstantConditions
        if (this.attributes != null)
        {
            this.attributeChangeListener = new java.beans.PropertyChangeListener()
            {
                public void propertyChange(java.beans.PropertyChangeEvent e)
                {
                    onShapeChanged();
                }
            };

            this.attributes.addPropertyChangeListener(this.attributeChangeListener);
        }

        this.onShapeChanged();
    }

    public String getPathType()
    {
        return this.pathType;
    }

    public void setPathType(String pathType)
    {
        if (pathType == null)
        {
            String message = Logging.getMessage("nullValue.PathTypeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.pathType = pathType;
        this.onShapeChanged();
    }

    public double getTexelsPerEdgeInterval()
    {
        return this.texelsPerEdgeInterval;
    }

    public void setTexelsPerEdgeInterval(double texelsPerEdgeInterval)
    {
        if (texelsPerEdgeInterval <= 0)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "texelsPerEdgeInterval <= 0");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.texelsPerEdgeInterval = texelsPerEdgeInterval;
        this.onShapeChanged();
    }

    public int[] getMinAndMaxEdgeIntervals()
    {
        return new int[] {this.minEdgeIntervals, this.maxEdgeIntervals};
    }

    public void setMinAndMaxEdgeIntervals(int minEdgeIntervals, int maxEdgeIntervals)
    {
        if (minEdgeIntervals < 0)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "minEdgeIntervals < 0");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (maxEdgeIntervals < 0)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "maxEdgeIntervals < 0");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.minEdgeIntervals = minEdgeIntervals;
        this.maxEdgeIntervals = maxEdgeIntervals;
        this.onShapeChanged();
    }

    public boolean isShowBoundingSectors()
    {
        return this.showBoundingSectors;
    }

    public void setShowBoundingSectors(boolean showBoundingSectors)
    {
        this.showBoundingSectors = showBoundingSectors;
        this.onShapeChanged();
    }

    public Iterable<? extends LatLon> getLocations(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        java.util.ArrayList<LatLon> locations = new java.util.ArrayList<LatLon>();
        this.computeLocations(globe, locations);
        return locations;
    }

    public Iterable<? extends Sector> getSectors(DrawContext dc, double texelSizeRadians)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        CacheEntry<Iterable<? extends Sector>> entry = this.sectorCache.get(texelSizeRadians);

        if (entry != null && entry.getValue() != null && entry.getLastModifiedTime() >= this.getLastModifiedTime())
            return entry.getValue();

        Iterable<? extends Sector> sectors = this.doGetSectors(dc, texelSizeRadians);

        this.sectorCache.put(texelSizeRadians,
            new CacheEntry<Iterable<? extends Sector>>(sectors, this.getLastModifiedTime()));
        return sectors;
    }

    protected Iterable<? extends Sector> doGetSectors(DrawContext dc, double texelSizeRadians)
    {
        Iterable<? extends LatLon> locations = this.getLocations(dc.getGlobe());
        Iterable<? extends Sector> sectors = getSurfaceShapeSupport().computeBoundingSectors(locations, this.pathType);
        return getSurfaceShapeSupport().adjustSectorsByBorderWidth(sectors, this.getAttributes(), texelSizeRadians);
    }

    public void preRender(DrawContext dc)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (!this.isVisible())
            return;

        this.getRenderer().preRender(dc);
    }

    public void render(DrawContext dc)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (!this.isVisible())
            return;

        this.getRenderer().render(dc);
    }

    public String getRestorableState()
    {
        RestorableSupport rs = RestorableSupport.newRestorableSupport();
        this.doGetRestorableState(rs, null);

        return rs.getStateAsXml();
    }

    public void restoreState(String stateInXml)
    {
        if (stateInXml == null)
        {
            String message = Logging.getMessage("nullValue.StringIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        RestorableSupport rs;
        try
        {
            rs = RestorableSupport.parse(stateInXml);
        }
        catch (Exception e)
        {
            // Parsing the document specified by stateInXml failed.
            String message = Logging.getMessage("generic.ExceptionAttemptingToParseStateXml", stateInXml);
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message, e);
        }

        this.doRestoreState(rs, null);
    }

    public double getArea(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        AreaMeasurer areaMeasurer = this.setupAreaMeasurer(globe);
        return areaMeasurer.getArea(globe);
    }

    public double getArea(Globe globe, boolean terrainConformant)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        AreaMeasurer areaMeasurer = this.setupAreaMeasurer(globe);
        areaMeasurer.setFollowTerrain(terrainConformant);
        return areaMeasurer.getArea(globe);
    }

    public double getPerimeter(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        AreaMeasurer areaMeasurer = this.setupAreaMeasurer(globe);
        return areaMeasurer.getPerimeter(globe);
    }

    public double getWidth(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        AreaMeasurer areaMeasurer = this.setupAreaMeasurer(globe);
        return areaMeasurer.getWidth(globe);
    }

    public double getHeight(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        AreaMeasurer areaMeasurer = this.setupAreaMeasurer(globe);
        return areaMeasurer.getHeight(globe);
    }

    public double getLength(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        AreaMeasurer areaMeasurer = this.setupAreaMeasurer(globe);
        return areaMeasurer.getLength(globe);
    }

    public void move(Position position)
    {
        if (position == null)
        {
            String message = Logging.getMessage("nullValue.PositionIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Position referencePosition = this.getReferencePosition();
        if (referencePosition == null)
            return;

        this.moveTo(referencePosition.add(position));
    }

    public void moveTo(Position position)
    {
        if (position == null)
        {
            String message = Logging.getMessage("nullValue.PositionIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Position oldReferencePosition = this.getReferencePosition();
        if (oldReferencePosition == null)
            return;

        this.doMoveTo(oldReferencePosition, position);
    }

    public abstract Position getReferencePosition();

    //**************************************************************//
    //********************  SurfaceShape Rendering  ****************//
    //**************************************************************//

    protected void doRenderToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (sector == null)
        {
            String message = Logging.getMessage("nullValue.SectorIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.beginRenderToRegion(dc, sector, x, y, width, height);
        try
        {
            this.bindShapeData(dc, sector, x, y, width, height);
            this.renderInteriorToRegion(dc, sector, x, y, width, height);
            this.renderOutlineToRegion(dc, sector, x, y, width, height);
            this.renderBoundingSectors(dc, sector, x, y, width, height);
        }
        finally
        {
            this.endRenderToRegion(dc, sector, x, y, width, height);
        }
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void beginRenderToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        SurfaceShapeSupport shapeSupport = getSurfaceShapeSupport();
        shapeSupport.beginRendering(dc);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void endRenderToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        SurfaceShapeSupport shapeSupport = getSurfaceShapeSupport();
        shapeSupport.endRendering(dc);
    }

    protected void bindShapeData(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        if (this.vertices == null)
        {
            this.vertices = new java.util.ArrayList<Vec4>();
        }

        this.vertices.clear();
        this.assembleShapeVertices(dc, sector, x, y, width, height, this.vertices);

        SurfaceShapeSupport shapeSupport = getSurfaceShapeSupport();
        shapeSupport.bindVertexData(dc, 2, this.vertices);
    }

    protected void assembleShapeVertices(DrawContext dc, Sector sector, int x, int y, int width, int height,
        java.util.List<Vec4> vertices)
    {
        Iterable<? extends LatLon> locations = this.getDrawLocations(dc, sector, x, y, width, height);

        SurfaceShapeSupport shapeSupport = getSurfaceShapeSupport();
        shapeSupport.assembleShapeVertices(sector, x, y, width, height, locations, this.vertices);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void renderInteriorToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        ShapeAttributes attributes = this.getAttributes();

        if (!attributes.isDrawInterior())
            return;

        if (attributes.getInteriorOpacity() <= 0)
            return;

        this.doRenderInteriorToRegion(dc, sector, x, y, width, height);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void renderOutlineToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        ShapeAttributes attributes = this.getAttributes();

        if (!attributes.isDrawOutline())
            return;

        if (attributes.getOutlineOpacity() <= 0)
            return;

        this.doRenderOutlineToRegion(dc, sector, x, y, width, height);
    }

    protected void renderBoundingSectors(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        if (!this.isShowBoundingSectors())
            return;

        this.doRenderBoundingSectors(dc, sector, x, y, width, height);
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected abstract void doRenderInteriorToRegion(DrawContext dc, Sector sector, int x, int y, int width,
        int height);

    @SuppressWarnings({"UnusedDeclaration"})
    protected abstract void doRenderOutlineToRegion(DrawContext dc, Sector sector, int x, int y, int width, int height);

    protected void doRenderBoundingSectors(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (sector == null)
        {
            String message = Logging.getMessage("nullValue.SectorIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();

        SurfaceShapeSupport shapeSupport = getSurfaceShapeSupport();
        java.util.ArrayList<Vec4> vertices = new java.util.ArrayList<Vec4>();
        double texelSizeRadians = sector.getDeltaLatRadians() / height;

        shapeSupport.applyMaterialState(dc, Material.GREEN, 1);
        gl.glLineWidth(1f);

        for (Sector s : this.getSectors(dc, texelSizeRadians))
        {
            vertices.clear();
            shapeSupport.assembleShapeVertices(sector, x, y, width, height, s, vertices);
            shapeSupport.bindVertexData(dc, 2, vertices);
            shapeSupport.drawArrays(dc, GL.GL_LINE_LOOP, 0, vertices.size());
        }
    }

    //**************************************************************//
    //********************  Protected Interface  *******************//
    //**************************************************************//

    protected abstract void doMoveTo(Position oldReferencePosition, Position newReferencePosition);

    protected abstract void computeLocations(Globe globe, java.util.List<LatLon> locations);

    protected abstract void computeDrawLocations(Globe globe, double edgeIntervalsPerDegree,
        java.util.List<LatLon> locations);

    protected void onShapeChanged()
    {
        this.updateModifiedTime();
    }

    protected static SurfaceShapeSupport getSurfaceShapeSupport()
    {
        if (surfaceShapeSupport == null)
        {
            surfaceShapeSupport = new SurfaceShapeSupport();
        }

        return surfaceShapeSupport;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected double computeEdgeIntervalsPerDegree(DrawContext dc, Sector sector, int x, int y, int width, int height)
    {
        double texelsPerDegree = Math.max(width / sector.getDeltaLonDegrees(), height / sector.getDeltaLatDegrees());
        double intervalsPerTexel = 1.0 / this.getTexelsPerEdgeInterval();

        return intervalsPerTexel * texelsPerDegree;
    }

    protected java.util.List<? extends LatLon> getDrawLocations(DrawContext dc, double edgeIntervalsPerDegree)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        CacheEntry<java.util.List<LatLon>> entry = this.drawLocationCache.get(edgeIntervalsPerDegree);

        if (entry != null && entry.getValue() != null && entry.getLastModifiedTime() >= this.getLastModifiedTime())
            return entry.getValue();

        java.util.List<LatLon> locations = null;
        if (entry != null)
            locations = entry.getValue();
        if (locations == null)
            locations = new java.util.ArrayList<LatLon>();

        locations.clear();
        this.computeDrawLocations(dc.getGlobe(), edgeIntervalsPerDegree, locations);

        this.drawLocationCache.put(edgeIntervalsPerDegree,
            new CacheEntry<java.util.List<LatLon>>(locations, this.getLastModifiedTime()));
        return locations;
    }

    protected java.util.List<? extends LatLon> getDrawLocations(DrawContext dc, Sector sector, int x, int y,
        int width, int height)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (sector == null)
        {
            String message = Logging.getMessage("nullValue.SectorIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        double edgeIntervalsPerDegree = this.computeEdgeIntervalsPerDegree(dc, sector, x, y, width, height);
        return this.getDrawLocations(dc, edgeIntervalsPerDegree);
    }

    protected TiledSurfaceObjectRenderer getRenderer()
    {
        if (this.renderer == null)
        {
            this.renderer = new TiledSurfaceObjectRenderer();
            // Don't need the tiled surface object renderer to handle picking. Picking for this shape will be handled
            // like any other Renderable.
            this.renderer.setPickEnabled(false);
            this.renderer.setSurfaceObjects(java.util.Arrays.asList(this));

            // Uncomment this code to view surface shape tile assembly.
            //this.renderer.setShowTileOutlines(true);
            //this.renderer.setTileBackgroundColor(new java.awt.Color(30, 30, 30, 60));
        }

        return this.renderer;
    }

    protected AreaMeasurer setupAreaMeasurer(Globe globe)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (this.areaMeasurer == null)
        {
            this.areaMeasurer = new AreaMeasurer();
        }

        // Try to use the currently cached locations. If the AreaMeasuerer is out of sync with this shape's state,
        // then update the AreaMeasurer's internal location list.
        if (this.areaMeasurerLastModifiedTime < this.getLastModifiedTime())
        {
            java.util.ArrayList<LatLon> locations = new java.util.ArrayList<LatLon>();
            this.computeLocations(globe, locations);
            this.areaMeasurer.setPositions(locations, 0);
            this.areaMeasurerLastModifiedTime = this.getLastModifiedTime();
        }

        // Surface shapes follow the terrain by definition.
        this.areaMeasurer.setFollowTerrain(true);

        return this.areaMeasurer;
    }

    //**************************************************************//
    //******************** Restorable State  ***********************//
    //**************************************************************//

    protected void doGetRestorableState(RestorableSupport rs, RestorableSupport.StateObject context)
    {
        // Note: showBoundingCylinders is a diagnostic flag, therefore it is not saved or restored.

        rs.addStateValueAsBoolean(context, "visible", this.isVisible());
        rs.addStateValueAsString(context, "pathType", this.getPathType());
        rs.addStateValueAsDouble(context, "texelsPerEdgeInterval", this.getTexelsPerEdgeInterval());

        int[] minAndMaxEdgeIntervals = this.getMinAndMaxEdgeIntervals();
        rs.addStateValueAsInteger(context, "minEdgeIntervals", minAndMaxEdgeIntervals[0]);
        rs.addStateValueAsInteger(context, "maxEdgeIntervals", minAndMaxEdgeIntervals[1]);

        this.attributes.getRestorableState(rs, rs.addStateObject(context, "attributes"));
    }

    protected void doRestoreState(RestorableSupport rs, RestorableSupport.StateObject context)
    {
        // Invoke the legacy restore functionality. This will enable the shape to recognize state XML elements
        // from the previous version of SurfaceShape.
        this.legacyRestoreState(rs, context);

        // Note: showBoundingCylinders is a diagnostic flag, therefore it is not saved or restored.

        Boolean b = rs.getStateValueAsBoolean(context, "visible");
        if (b != null)
            this.setVisible(b);

        String s = rs.getStateValueAsString(context, "pathType");
        if (s != null)
        {
            String pathType = this.pathTypeFromString(s);
            if (pathType != null)
                this.setPathType(pathType);
        }

        Double d = rs.getStateValueAsDouble(context, "texelsPerEdgeInterval");
        if (d != null)
            this.setTexelsPerEdgeInterval(d);

        int[] minAndMaxEdgeIntervals = this.getMinAndMaxEdgeIntervals();

        Integer minEdgeIntervals = rs.getStateValueAsInteger(context, "minEdgeIntervals");
        if (minEdgeIntervals != null)
            minAndMaxEdgeIntervals[0] = minEdgeIntervals;

        Integer maxEdgeIntervals = rs.getStateValueAsInteger(context, "maxEdgeIntervals");
        if (maxEdgeIntervals != null)
            minAndMaxEdgeIntervals[1] = maxEdgeIntervals;

        if (minEdgeIntervals != null || maxEdgeIntervals != null)
            this.setMinAndMaxEdgeIntervals(minAndMaxEdgeIntervals[0], minAndMaxEdgeIntervals[1]);

        RestorableSupport.StateObject so = rs.getStateObject(context, "attributes");
        if (so != null)
            this.getAttributes().restoreState(rs, so);
    }

    /**
     * Restores state values from previous versions of the SurfaceShape state XML. These values are stored or named
     * differently than the current implementation. Those values which have not changed are ignored here, and will
     * restored in {# doRestoreState }.
     *
     * @param rs      RestorableSupport object which contains the state value properties.
     * @param context active context in the RestorableSupport to read state from.
     */
    protected void legacyRestoreState(RestorableSupport rs, RestorableSupport.StateObject context)
    {
        // Ignore texture width and height parameters, they're no longer used.
        //Integer width = rs.getStateValueAsInteger(context, "textureWidth");
        //Integer height = rs.getStateValueAsInteger(context, "textureHeight");
        //if (width != null && height != null)
        //    this.setTextureSize(new Dimension(width, height));

        java.awt.Color color = rs.getStateValueAsColor(context, "color");
        if (color != null)
            this.getAttributes().setInteriorMaterial(new Material(color));

        color = rs.getStateValueAsColor(context, "borderColor");
        if (color != null)
            this.getAttributes().setOutlineMaterial(new Material(color));

        Double dub = rs.getStateValueAsDouble(context, "lineWidth");
        if (dub != null)
            this.getAttributes().setOutlineWidth(dub);

        // Ignore numEdgeIntervalsPerDegree, since it's no longer used.
        //Double intervals = rs.getStateValueAsDouble(context, "numEdgeIntervalsPerDegree");
        //if (intervals != null)
        //    this.setEdgeIntervalsPerDegree(intervals.intValue());

        Boolean booleanState = rs.getStateValueAsBoolean(context, "drawBorder");
        if (booleanState != null)
            this.getAttributes().setDrawOutline(booleanState);

        booleanState = rs.getStateValueAsBoolean(context, "drawInterior");
        if (booleanState != null)
            this.getAttributes().setDrawInterior(booleanState);

        booleanState = rs.getStateValueAsBoolean(context, "antialias");
        if (booleanState != null)
            this.getAttributes().setEnableAntialiasing(booleanState);

        // Positions data is a per object property now. This value will be picked up by SurfacePolygon, SurfacePolyline,
        // and SurfaceSector. Other shapes will ignore this property.
        //ArrayList<LatLon> locations = rs.getStateValueAsLatLonList(context, "locations");
        //if (locations != null)
        //    this.positions = locations;
    }

    protected String pathTypeFromString(String s)
    {
        if (s == null)
            return null;

        if (s.equals(AVKey.GREAT_CIRCLE))
        {
            return AVKey.GREAT_CIRCLE;
        }
        else if (s.equals(AVKey.LINEAR))
        {
            return AVKey.LINEAR;
        }
        else if (s.equals(AVKey.LOXODROME))
        {
            return AVKey.LOXODROME;
        }
        else if (s.equals(AVKey.RHUMB_LINE))
        {
            return AVKey.RHUMB_LINE;
        }

        return null;
    }

    //**************************************************************//
    //******************** Cache Utilities  ************************//
    //**************************************************************//

    protected static class CacheEntry<T>
    {
        protected final T value;
        protected final long lastModifiedTime;

        public CacheEntry(T value, long lastModifiedTime)
        {
            this.value = value;
            this.lastModifiedTime = lastModifiedTime;
        }

        public T getValue()
        {
            return this.value;
        }

        public final long getLastModifiedTime()
        {
            return this.lastModifiedTime;
        }
    }
}
