/* Copyright (C) 2001, 2009 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.render;

import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.globes.Globe;
import gov.nasa.worldwind.util.*;
import gov.nasa.worldwind.avlist.AVKey;

/**
 * @author dcollins
 * @version $Id: SurfacePolygon.java 10634 2009-04-29 03:52:29Z dcollins $
 */
public class SurfacePolygon extends SurfaceConcaveShape
{
    protected java.util.List<LatLon> locationList;

    public SurfacePolygon(ShapeAttributes attributes, Iterable<? extends LatLon> locations)
    {
        super(attributes);

        this.locationList = new java.util.ArrayList<LatLon>();
        this.addLocations(locations);
    }

    public SurfacePolygon(ShapeAttributes attributes)
    {
        this(attributes, null);
    }

    public SurfacePolygon(Iterable<? extends LatLon> locations)
    {
        this(new BasicShapeAttributes(), locations);
    }

    public SurfacePolygon()
    {
        this(new BasicShapeAttributes());
    }

    public Iterable<? extends LatLon> getLocations()
    {
        return java.util.Collections.unmodifiableList(this.locationList);
    }

    public void setLocations(Iterable<? extends LatLon> locations)
    {
        this.locationList.clear();
        this.addLocations(locations);
        this.onShapeChanged();
    }

    protected void addLocations(Iterable<? extends LatLon> locations)
    {
        if (locations != null)
        {
            for (LatLon latLon : locations)
            {
                this.locationList.add(latLon);
            }
        }
    }

    public Position getReferencePosition()
    {
        if (this.locationList.isEmpty())
            return null;

        return new Position(this.locationList.get(0), 0);
    }

    protected void computeLocations(Globe globe, java.util.List<LatLon> locations)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (locations == null)
        {
            String message = Logging.getMessage("nullValue.LocationsListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        locations.addAll(this.locationList);
        getSurfaceShapeSupport().makeClosedPath(locations);
    }

    protected void computeDrawLocations(Globe globe, double edgeIntervalsPerDegree, java.util.List<LatLon> locations)
    {
        if (globe == null)
        {
            String message = Logging.getMessage("nullValue.GlobeIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (locations == null)
        {
            String message = Logging.getMessage("nullValue.LocationsListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        getSurfaceShapeSupport().generateIntermediateLocations(this.locationList, this.pathType,
            edgeIntervalsPerDegree, this.minEdgeIntervals, this.maxEdgeIntervals, true, locations);
    }

    protected void doMoveTo(Position oldReferencePosition, Position newReferencePosition)
    {
        int numLocations = this.locationList.size();
        LatLon[] newLocations = new LatLon[numLocations];

        for (int i = 0; i < numLocations; i++)
        {
            LatLon ll = this.locationList.get(i);
            Angle heading = LatLon.greatCircleAzimuth(oldReferencePosition, ll);
            Angle pathLength = LatLon.greatCircleDistance(oldReferencePosition, ll);
            newLocations[i] = LatLon.greatCircleEndPosition(newReferencePosition, heading, pathLength);
        }

        this.setLocations(java.util.Arrays.asList(newLocations));
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void assembleShapeVertices(DrawContext dc, Sector sector, int x, int y, int width, int height,
        java.util.List<Vec4> vertices)
    {
        super.assembleShapeVertices(dc, sector, x, y, width, height, vertices);

        // This is a generic concave polygonal shape, with no guarantees about the caller specified locations. Make
        // sure the winding order is counter clockwise, and strip any redundant endpoints.
        this.makeWindingCounterClockwise(vertices);
        this.stripRedundantEndPoint(vertices);
    }

    protected void makeWindingCounterClockwise(java.util.List<Vec4> vertices)
    {
        // Check the winding order of the polygon locations. If they're not in counter clockwise order, then reverse
        // the locations.
        String windingOrder = GeometryMath.computePolygonWindingOrder2(vertices);
        //noinspection StringEquality
        if (windingOrder != AVKey.COUNTER_CLOCKWISE)
        {
            java.util.Collections.reverse(vertices);
        }
    }

    protected void stripRedundantEndPoint(java.util.List<Vec4> vertices)
    {
        if (vertices.size() < 2)
            return;

        int lastIndex = vertices.size() -1;
        if (vertices.get(0).equals(vertices.get(lastIndex)))
        {
            vertices.remove(lastIndex);
        }
    }

    //**************************************************************//
    //******************** Restorable State  ***********************//
    //**************************************************************//

    protected void doGetRestorableState(RestorableSupport rs, RestorableSupport.StateObject context)
    {
        super.doGetRestorableState(rs, context);

        rs.addStateValueAsLatLonList(context, "locationList", this.getLocations());
    }

    protected void doRestoreState(RestorableSupport rs, RestorableSupport.StateObject context)
    {
        super.doRestoreState(rs, context);

        Iterable<LatLon> iterable = rs.getStateValueAsLatLonList(context, "locationList");
        if (iterable != null)
            this.setLocations(iterable);
    }

    protected void legacyRestoreState(RestorableSupport rs, RestorableSupport.StateObject context)
    {
        super.legacyRestoreState(rs, context);

        java.util.ArrayList<LatLon> locations = rs.getStateValueAsLatLonList(context, "locations");
        if (locations != null)
            this.setLocations(locations);
    }
}
