/* Copyright (C) 2001, 2009 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.render;

import com.sun.opengl.util.BufferUtil;
import gov.nasa.worldwind.avlist.AVKey;
import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.util.*;

import javax.media.opengl.GL;

/**
 * @author dcollins
 * @version $Id: SurfaceShapeSupport.java 11457 2009-06-05 07:09:37Z dcollins $
 */
public class SurfaceShapeSupport
{
    protected int vertexCount;
    protected double[] vertexArray;
    protected java.nio.DoubleBuffer vertexBuffer;
    protected OGLStackHandler stackHandler;
    protected OGLStateSupport stateSupport;

    public SurfaceShapeSupport()
    {
        this.stackHandler = new OGLStackHandler();
        this.stateSupport = new OGLStateSupport();
    }

    //**************************************************************//
    //********************  Shape Rendering  ***********************//
    //**************************************************************//

    public void beginRendering(DrawContext dc)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();

        this.stateSupport.setEnableColor(!dc.isPickingMode());

        this.stackHandler.pushAttrib(gl,
            this.stateSupport.getAttributeBits()
                | GL.GL_LINE_BIT);

        this.stackHandler.pushClientAttrib(gl, GL.GL_CLIENT_VERTEX_ARRAY_BIT);

        gl.glEnableClientState(GL.GL_VERTEX_ARRAY);
    }

    public void endRendering(DrawContext dc)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();

        this.stackHandler.pop(gl);
    }

    public void applyMaterialState(DrawContext dc, Material material, double opacity)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (material == null)
        {
            String message = Logging.getMessage("nullValue.MaterialIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();

        java.awt.Color color = material.getDiffuse();
        this.stateSupport.setColor(color, opacity);
        this.stateSupport.setColorMode(OGLStateSupport.COLOR_NO_PREMULTIPLIED_ALPHA);
        this.stateSupport.apply(gl);
    }

    public void applyInteriorState(DrawContext dc, ShapeAttributes attributes)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (attributes == null)
        {
            String message = Logging.getMessage("nullValue.AttributesIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.applyMaterialState(dc, attributes.getInteriorMaterial(), attributes.getInteriorOpacity());
    }

    public void applyOutlineState(DrawContext dc, ShapeAttributes attributes)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (attributes == null)
        {
            String message = Logging.getMessage("nullValue.AttributesIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();

        double lineWidth = attributes.getOutlineWidth();
        if (dc.isPickingMode() && !attributes.isDrawInterior())
        {
            if (lineWidth != 0)
                lineWidth += 5;
        }
        gl.glLineWidth((float) lineWidth);

        if (!dc.isPickingMode() && attributes.isEnableAntialiasing())
        {
            gl.glEnable(GL.GL_LINE_SMOOTH);
        }
        else
        {
            gl.glDisable(GL.GL_LINE_SMOOTH);
        }

        this.applyMaterialState(dc, attributes.getOutlineMaterial(), attributes.getOutlineOpacity());
    }

    public void bindVertexData(DrawContext dc, int numCoords, java.nio.DoubleBuffer buffer, int count)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (!(numCoords == 2 || numCoords == 3 || numCoords == 4))
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "numCoords must be 2, 3, or 4");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (buffer == null)
        {
            String message = Logging.getMessage("nullValue.VertexBufferNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();
        gl.glVertexPointer(numCoords, GL.GL_DOUBLE, 0, buffer);
        this.vertexCount = count;
    }

    public void bindVertexData(DrawContext dc, int numCoords, java.util.List<? extends Vec4> vertices)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (vertices == null)
        {
            String message = Logging.getMessage("nullValue.VertexListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        int count = vertices.size();
        int numElements = numCoords * count;

        java.nio.DoubleBuffer buffer = this.getVertexBuffer(numElements);
        buffer.rewind();

        this.assembleVertexData(numCoords, vertices, buffer);
        buffer.rewind();

        this.bindVertexData(dc, numCoords, buffer, count);
    }

    public void drawArrays(DrawContext dc, int mode, int first, int count)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (first < 0)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "first < 0");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (count > this.vertexCount)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "count > vertexCount");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        GL gl = dc.getGL();
        gl.glDrawArrays(mode, first, count);
    }

    protected double[] getVertexArray(int capacity)
    {
        if (this.vertexArray == null || this.vertexArray.length < capacity)
        {
            this.vertexArray = new double[capacity];
        }

        return this.vertexArray;
    }

    protected java.nio.DoubleBuffer getVertexBuffer(int capacity)
    {
        if (this.vertexBuffer == null || this.vertexBuffer.capacity() < capacity)
        {
            this.vertexBuffer = BufferUtil.newDoubleBuffer(capacity);
        }

        return this.vertexBuffer;
    }

    //**************************************************************//
    //********************  Shape Location Assembly  ***************//
    //**************************************************************//

    @SuppressWarnings({"StringEquality"})
    public Iterable<? extends Sector> computeBoundingSectors(Iterable<? extends LatLon> locations, String pathType)
    {
        if (locations == null)
        {
            String message = Logging.getMessage("nullValue.IterableIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Iterable<? extends Sector> sectors;

        if (LatLon.positionsCrossDateLine(locations))
        {
            sectors = java.util.Arrays.asList(Sector.splitBoundingSectors(locations));
        }
        else
        {
            sectors = java.util.Arrays.asList(Sector.boundingSector(locations));
        }

        // Great circle paths between two latitudes may result in a latitude which is greater or smaller than either of
        // the two latitudes. All other path types are bounded by the defining locations.
        if (pathType != null && pathType == AVKey.GREAT_CIRCLE)
        {
            java.util.List<Sector> adjustedSectors = new java.util.ArrayList<Sector>();

            for (Sector sector : sectors)
            {
                Angle[] extremeLatitudes = this.greatArcExtremeLatitudes(locations);

                double minLatDegrees = sector.getMinLatitude().degrees;
                double maxLatDegrees = sector.getMaxLatitude().degrees;

                if (minLatDegrees > extremeLatitudes[0].degrees)
                    minLatDegrees = extremeLatitudes[0].degrees;
                if (maxLatDegrees < extremeLatitudes[1].degrees)
                    maxLatDegrees = extremeLatitudes[1].degrees;

                Angle minLat = Angle.fromDegreesLatitude(minLatDegrees);
                Angle maxLat = Angle.fromDegreesLatitude(maxLatDegrees);

                adjustedSectors.add(new Sector(minLat, maxLat, sector.getMinLongitude(), sector.getMaxLongitude()));
            }

            sectors = adjustedSectors;
        }

        return sectors;
    }

    public Iterable<? extends Sector> adjustSectorsByBorderWidth(Iterable<? extends Sector> sectors,
        ShapeAttributes attributes, double texelSizeRadians)
    {
        if (sectors == null)
        {
            String message = Logging.getMessage("nullValue.IterableIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (attributes == null)
        {
            String message = Logging.getMessage("nullValue.AttributesIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        // TODO
        // we are adjusting the bounding sectors to accomodate border width after the dateline split. This means the
        // extra room for the border will not extend across the dateline.

        // Compute the latitude or longitude delta degrees which the border extends beyond the shape fill geometry.
        // Since the border straddles the shape fill geometry, the delta will be 1/2 of the actual border size.
        double borderPaddingRadians = texelSizeRadians * attributes.getOutlineWidth() / 2.0;

        java.util.List<Sector> adjustedSectors = new java.util.ArrayList<Sector>();

        for (Sector sector : sectors)
        {
            double minLatRadians = sector.getMinLatitude().radians - borderPaddingRadians;
            double maxLatRadians = sector.getMaxLatitude().radians + borderPaddingRadians;
            double minLonRadians = sector.getMinLongitude().radians;
            double maxLonRadians = sector.getMaxLongitude().radians;

            if (LatLon.positionsCrossDateLine(sector))
            {
                minLonRadians += borderPaddingRadians;
                maxLonRadians -= borderPaddingRadians;
            }
            else
            {
                minLonRadians -= borderPaddingRadians;
                maxLonRadians += borderPaddingRadians;
            }

            Angle minLat = Angle.fromRadiansLatitude(minLatRadians);
            Angle maxLat = Angle.fromRadiansLatitude(maxLatRadians);
            Angle minLon = Angle.fromRadiansLongitude(minLonRadians);
            Angle maxLon = Angle.fromRadiansLongitude(maxLonRadians);

            adjustedSectors.add(new Sector(minLat, maxLat, minLon, maxLon));
        }

        return adjustedSectors;
    }

    public void makeClosedPath(java.util.List<LatLon> locations)
    {
        if (locations == null)
        {
            String message = Logging.getMessage("nullValue.LocationsListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (locations.size() < 2)
            return;

        int lastIndex = locations.size() - 1;
        if (!locations.get(0).equals(locations.get(lastIndex)))
        {
            locations.add(locations.get(0));
        }
    }

    public void generateIntermediateLocations(Iterable<? extends LatLon> locations,
        String pathType, double edgeIntervalsPerDegree, int minEdgeIntervals, int maxEdgeIntervals,
        boolean makeClosedPath,
        java.util.List<LatLon> newLocations)
    {
        if (locations == null)
        {
            String message = Logging.getMessage("nullValue.IterableIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (newLocations == null)
        {
            String message = Logging.getMessage("nullValue.LocationsListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        LatLon firstLocation = null;
        LatLon lastLocation = null;

        for (LatLon ll : locations)
        {
            if (firstLocation == null)
            {
                firstLocation = ll;
            }

            if (lastLocation != null)
            {
                this.addIntermediateLocations(lastLocation, ll, pathType,
                    edgeIntervalsPerDegree, minEdgeIntervals, maxEdgeIntervals, newLocations);
            }

            newLocations.add(ll);
            lastLocation = ll;
        }

        // If the caller has instructed us to generate locations for a closed path, then check to see if the specified
        // locations define a closed path. If not, then we need to generate intermediate locations between the last
        // and first locations, then close the path by repeating the first location.
        if (makeClosedPath)
        {
            if (firstLocation != null && lastLocation != null && !firstLocation.equals(lastLocation))
            {
                this.addIntermediateLocations(lastLocation, firstLocation, pathType,
                    edgeIntervalsPerDegree, minEdgeIntervals, maxEdgeIntervals, newLocations);
                newLocations.add(firstLocation);
            }
        }
    }

    protected Angle[] greatArcExtremeLatitudes(Iterable<? extends LatLon> locations)
    {
        double minLat = Angle.POS90.degrees;
        double maxLat = Angle.NEG90.degrees;

        LatLon lastLocation = null;

        for (LatLon ll : locations)
        {
            if (lastLocation != null)
            {
                LatLon[] extremes = LatLon.greatCircleArcExtremeLocations(lastLocation, ll);
                if (extremes != null)
                {
                    for (LatLon extreme : extremes)
                    {
                        if (minLat > extreme.getLatitude().degrees)
                            minLat = extreme.getLatitude().degrees;
                        if (maxLat < extreme.getLatitude().degrees)
                            maxLat = extreme.getLatitude().degrees;
                    }
                }
            }

            lastLocation = ll;
        }

        return new Angle[] {Angle.fromDegrees(minLat), Angle.fromDegrees(maxLat)};
    }

    @SuppressWarnings({"StringEquality"})
    protected void addIntermediateLocations(LatLon a, LatLon b, String pathType,
        double edgeIntervalsPerDegree, int minEdgeIntervals, int maxEdgeIntervals, java.util.List<LatLon> newLocations)
    {
        if (pathType != null && pathType == AVKey.GREAT_CIRCLE)
        {
            Angle pathLength = LatLon.greatCircleDistance(a, b);

            double edgeIntervals = WWMath.clamp(edgeIntervalsPerDegree * pathLength.degrees,
                minEdgeIntervals, maxEdgeIntervals);
            int numEdgeIntervals = (int) Math.ceil(edgeIntervals);

            if (numEdgeIntervals > 1)
            {
                double headingRadians = LatLon.greatCircleAzimuth(a, b).radians;
                double stepSizeRadians = pathLength.radians / (numEdgeIntervals + 1);

                for (int i = 1; i <= numEdgeIntervals; i++)
                {
                    LatLon newLocation = LatLon.greatCircleEndPosition(a, headingRadians, i * stepSizeRadians);
                    newLocations.add(newLocation);
                }
            }
        }
        else if (pathType != null && (pathType == AVKey.RHUMB_LINE || pathType == AVKey.LOXODROME))
        {
            Angle pathLength = LatLon.rhumbDistance(a, b);

            double edgeIntervals = WWMath.clamp(edgeIntervalsPerDegree * pathLength.degrees,
                minEdgeIntervals, maxEdgeIntervals);
            int numEdgeIntervals = (int) Math.ceil(edgeIntervals);

            if (numEdgeIntervals > 1)
            {
                double headingRadians = LatLon.rhumbAzimuth(a, b).radians;
                double stepSizeRadians = pathLength.radians / (numEdgeIntervals + 1);

                for (int i = 1; i <= numEdgeIntervals; i++)
                {
                    LatLon newLocation = LatLon.rhumbEndPosition(a, headingRadians, i * stepSizeRadians);
                    newLocations.add(newLocation);
                }
            }
        }
        else // Default to linear interpolation in latitude and longitude.
        {
            // Linear interpolation between 2D coordinates is already performed by GL during shape rasterization.
            // There is no need to duplicate that effort here.
        }
    }

    //**************************************************************//
    //********************  Shape Vertex Assembly  *****************//
    //**************************************************************//

    public void assembleVertexData(int numCoords, double[] vertexArray, int count, java.nio.DoubleBuffer buffer)
    {
        if (vertexArray == null)
        {
            String message = Logging.getMessage("nullValue.VertexArrayIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (buffer == null)
        {
            String message = Logging.getMessage("nullValue.VertexBufferNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        int numElements = numCoords * count;
        buffer.rewind();
        buffer.put(vertexArray, 0, numElements);
        buffer.rewind();
    }

    public void assembleVertexData(int numCoords, java.util.List<? extends Vec4> vertices, java.nio.DoubleBuffer buffer)
    {
        if (!(numCoords == 2 || numCoords == 3 || numCoords == 4))
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "numCoords must be 2, 3, or 4");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (vertices == null)
        {
            String message = Logging.getMessage("nullValue.VertexListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (buffer == null)
        {
            String message = Logging.getMessage("nullValue.VertexBufferNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        int count = vertices.size();
        int numElements = numCoords * count;
        double[] vertexArray = this.getVertexArray(numElements);

        if (numCoords == 2)
        {
            int index = 0;
            for (Vec4 vec : vertices)
            {
                vec.toArray2(vertexArray, numCoords * index);
                index++;
            }
        }
        else if (numCoords == 3)
        {
            int index = 0;
            for (Vec4 vec : vertices)
            {
                vec.toArray3(vertexArray, numCoords * index);
                index++;
            }
        }
        else // (numCoords == 4)
        {
            int index = 0;
            for (Vec4 vec : vertices)
            {
                vec.toArray4(vertexArray, numCoords * index);
                index++;
            }
        }

        this.assembleVertexData(numCoords, vertexArray, count, buffer);
    }

    public void assembleShapeVertices(Sector sector, int x, int y, int width, int height,
        Iterable<? extends LatLon> locations, java.util.List<Vec4> vertices)
    {
        if (sector == null)
        {
            String message = Logging.getMessage("nullValue.SectorIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (locations == null)
        {
            String message = Logging.getMessage("nullValue.IterableIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (vertices == null)
        {
            String message = Logging.getMessage("nullValue.VertexListIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Matrix transform = Matrix.fromGeographicToViewport(sector, x, y, width, height);
        boolean locationsCrossDateLine = LatLon.positionsCrossDateLine(locations);
        boolean inWesternHemisphere = sector.getMaxLongitude().degrees < 0;

        for (LatLon ll : locations)
        {
            double lat = ll.getLatitude().degrees;
            double lon = ll.getLongitude().degrees;

            if (locationsCrossDateLine)
            {
                if (inWesternHemisphere && ll.getLongitude().degrees > 0)
                {
                    lon -= 360;
                }
                else if (!inWesternHemisphere && ll.getLongitude().degrees < 0)
                {
                    lon += 360;
                }
            }

            Vec4 point = new Vec4(lon, lat, 1);
            point = point.transformBy4(transform);
            vertices.add(point);
        }
    }
}
