/* Copyright (C) 2001, 2009 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.render;

import com.sun.opengl.util.texture.*;
import gov.nasa.worldwind.View;
import gov.nasa.worldwind.avlist.*;
import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.globes.Globe;
import gov.nasa.worldwind.layers.*;
import gov.nasa.worldwind.pick.*;
import gov.nasa.worldwind.terrain.SectorGeometryList;
import gov.nasa.worldwind.util.*;

import javax.media.opengl.GL;

/**
 * @author dcollins
 * @version $Id: TiledSurfaceShapeRenderer.java 9992 2009-04-08 04:38:54Z dcollins $
 */
public class TiledSurfaceObjectRenderer
{
    // Default attribute values.
    protected static final LatLon DEFAULT_TILE_DELTA = LatLon.fromDegrees(36, 36);
    protected static final Sector DEFAULT_SECTOR = Sector.FULL_SPHERE;
    protected static final int DEFAULT_NUM_LEVELS = 17;
    protected static final int DEFAULT_NUM_EMPTY_LEVELS = 0;
    protected static final int DEFAULT_TILE_WIDTH = 512;
    protected static final int DEFAULT_TILE_HEIGHT = 512;
    protected static final boolean DEFAULT_PICK_ENABLED = false;
    protected static final boolean DEFAULT_USE_MIPMAPS = false;
    // Default with a ratio of 1.5 surface tile texels to 1 screen pixel. This has been emperically determined to
    // produce good looking results with the standard class of SurfaceShapes such as SurfacePolygon.
    protected static final double DEFAULT_TEXEL_TO_PIXEL_RATIO = 1.5;
    protected static final java.awt.Color DEFAULT_TILE_BACKGROUND_COLOR = new java.awt.Color(0, 0, 0, 0);
    protected static final boolean DEFAULT_SHOW_TILE_OUTLINES = false;
    protected static final String DEFAULT_TILE_COUNT_NAME = "Surface Object Tiles";

    protected static long uniqueId = 1;

    protected Iterable<? extends SurfaceObject> surfaceObjectIterable;
    protected LevelSet levelSet;
    protected boolean pickEnabled;
    protected boolean useMipmaps;
    protected double texelToPixelRatio;
    protected java.awt.Color tileBackgroundColor;
    // State computed in each rendering pass.
    protected java.util.List<SurfaceObjectInfo> currentSurfaceObjects;
    protected java.util.List<TextureTile> currentTiles;
    protected java.util.Map<TileKey, SurfaceObjectState> tileStateMap;
    protected Sector topLevelBoundingSector;
    protected java.awt.Rectangle currentTileViewport;
    protected PickTile pickTile;
    protected boolean havePickTile;
    // Rendering support components.
    protected PickSupport pickSupport = new PickSupport();
    protected OGLOrtho2DSupport renderToTextureSupport = new OGLOrtho2DSupport();
    // Rendering diagnostic attrbiutes
    protected boolean showTileOutlines;
    protected String tileCountName;

    public TiledSurfaceObjectRenderer(LevelSet levelSet)
    {
        if (levelSet == null)
        {
            String message = Logging.getMessage("nullValue.LevelSetIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.levelSet = levelSet;
        this.useMipmaps = DEFAULT_USE_MIPMAPS;
        this.tileBackgroundColor = DEFAULT_TILE_BACKGROUND_COLOR;
        this.texelToPixelRatio = DEFAULT_TEXEL_TO_PIXEL_RATIO;

        this.currentSurfaceObjects = new java.util.ArrayList<SurfaceObjectInfo>();
        this.currentTiles = new java.util.ArrayList<TextureTile>();
        this.tileStateMap = new java.util.HashMap<TileKey, SurfaceObjectState>();
        this.pickTile = new PickTile(Sector.EMPTY_SECTOR);

        this.showTileOutlines = DEFAULT_SHOW_TILE_OUTLINES;
        this.tileCountName = DEFAULT_TILE_COUNT_NAME;
    }

    public TiledSurfaceObjectRenderer(AVList params)
    {
        this(new LevelSet(initParams(params)));
    }

    public TiledSurfaceObjectRenderer()
    {
        this(new AVListImpl());
    }

    protected static AVList initParams(AVList params)
    {
        if (params == null)
        {
            String message = Logging.getMessage("nullValue.ParamsIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Object o;
        Object cacheId = null;

        o = params.getValue(AVKey.LEVEL_ZERO_TILE_DELTA);
        if (o == null || !(o instanceof LatLon))
            params.setValue(AVKey.LEVEL_ZERO_TILE_DELTA, DEFAULT_TILE_DELTA);

        o = params.getValue(AVKey.SECTOR);
        if (o == null || !(o instanceof LatLon))
            params.setValue(AVKey.SECTOR, DEFAULT_SECTOR);

        o = params.getValue(AVKey.NUM_LEVELS);
        if (o == null || !(o instanceof Integer))
            params.setValue(AVKey.NUM_LEVELS, DEFAULT_NUM_LEVELS);

        o = params.getValue(AVKey.NUM_EMPTY_LEVELS);
        if (o == null || !(o instanceof Integer))
            params.setValue(AVKey.NUM_EMPTY_LEVELS, DEFAULT_NUM_EMPTY_LEVELS);

        o = params.getValue(AVKey.TILE_WIDTH);
        if (o == null || !(o instanceof Integer))
            params.setValue(AVKey.TILE_WIDTH, DEFAULT_TILE_WIDTH);

        o = params.getValue(AVKey.TILE_HEIGHT);
        if (o == null || !(o instanceof Integer))
            params.setValue(AVKey.TILE_HEIGHT, DEFAULT_TILE_HEIGHT);

        o = params.getValue(AVKey.DATA_CACHE_NAME);
        if (o == null)
        {
            //noinspection ConstantConditions
            if (cacheId == null)
            {
                cacheId = createUniqueCacheId();
            }

            params.setValue(AVKey.DATA_CACHE_NAME, cacheId);
        }

        o = params.getValue(AVKey.DATASET_NAME);
        if (o == null)
        {
            if (cacheId == null)
            {
                cacheId = createUniqueCacheId();
            }

            params.setValue(AVKey.DATASET_NAME, cacheId);
        }

        // We won't use any tile resource paths, so just supply a dummy format suffix.
        o = params.getValue(AVKey.FORMAT_SUFFIX);
        if (o == null)
            params.setValue(AVKey.FORMAT_SUFFIX, ".xyz");

        return params;
    }

    protected static String createUniqueCacheId()
    {
        StringBuilder sb = new StringBuilder();
        sb.append(TiledSurfaceObjectRenderer.class.getName());
        sb.append("/");
        sb.append(nextUniqueId());

        return sb.toString();
    }

    protected static long nextUniqueId()
    {
        return uniqueId++;
    }

    public boolean isPickEnabled()
    {
        return this.pickEnabled;
    }

    public void setPickEnabled(boolean enabled)
    {
        this.pickEnabled = enabled;
    }

    public boolean isUseMipmaps()
    {
        return this.useMipmaps;
    }

    public void setUseMipmaps(boolean useMipmaps)
    {
        this.useMipmaps = useMipmaps;
    }

    /**
     * Returns the surface object texel to screen pixel ratio. See {#linksetPixelToTexelRatio(double} for a description
     * of this value's meaning.
     *
     * @return surface object texel to screen pixel ratio.
     */
    public double getTexelToPixelRatio()
    {
        return this.texelToPixelRatio;
    }

    /**
     * Sets the surface object texel to screen pixel ratio.
     * <ul>
     * <li>
     * If <code>ratio</code> is 1, the renderer will attempt to create tiles whose texels have approximately the same
     * size as a screen pixel.
     * </li>
     * <li>
     * If <code>ratio</code> is less than 1, the renderer will attempt to create tiles whose texels are smaller than
     * a screen pixel. For example, specifying ratio=0.5 indicates that a surface object texel should be half the size
     * of a screen pixel.
     * </li>
     * <li>
     * If <code>ratio</code> is greater than 1, the renderer will attempt to create tiles whose texels are larger than
     * a screen pixel. For example, specifying ratio=2 indicates that a surface object texel should be twice the size of
     * a screen pixel.
     *
     * @param ratio surface object texel to screen pixel ratio.
     *
     * @throws IllegalArgumentException if <code>ratio</code> is less than or equal to zero.
     */
    public void setTexelToPixelRatio(double ratio)
    {
        if (ratio <= 0)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", "ratio <= 0");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.texelToPixelRatio = ratio;
    }

    public java.awt.Color getTileBackgroundColor()
    {
        return this.tileBackgroundColor;
    }

    public void setTileBackgroundColor(java.awt.Color color)
    {
        if (color == null)
        {
            String message = Logging.getMessage("nullValue.ColorIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.tileBackgroundColor = color;
    }

    public boolean isShowTileOutlines()
    {
        return this.showTileOutlines;
    }

    public void setShowTileOutlines(boolean show)
    {
        this.showTileOutlines = show;
    }

    public Iterable<? extends SurfaceObject> getSurfaceObjects()
    {
        return this.surfaceObjectIterable;
    }

    public void setSurfaceObjects(Iterable<? extends SurfaceObject> objects)
    {
        this.surfaceObjectIterable = objects;
    }

    public void preRender(DrawContext dc)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (this.surfaceObjectIterable == null)
            return;

        java.awt.Rectangle lastTileViewport = this.currentTileViewport;
        this.currentTileViewport = this.computeTileViewport(dc);
        // If the last viewport used to update the tiles is not equal to the current viewport, then invalidate the
        // current tile state, which will force all tiles to update regardless of whether their content has changed.
        if (lastTileViewport != null && !lastTileViewport.equals(this.currentTileViewport))
        {
            this.clearAllTileState();
        }

        this.clearSurfaceObjects(dc);
        this.clearTiles(dc);

        // We've cleared any surface object and tile assembly state from the last rendering pass. Determine whether or
        // not we can assemble and update the tiles. If not, we're done.
        if (!this.canAssembleTiles(dc))
            return;

        this.assembleSurfaceObjects(dc);
        this.assembleTiles(dc);

        this.updateTiles(dc);
    }

    public void render(DrawContext dc)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (this.surfaceObjectIterable == null)
            return;

        this.draw(dc, false);
    }

    public void pick(DrawContext dc, java.awt.Point pickPoint, Layer layer)
    {
        if (dc == null)
        {
            String message = Logging.getMessage("nullValue.DrawContextIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (this.surfaceObjectIterable == null)
            return;

        if (!this.pickEnabled)
            return;

        this.pickSupport.beginPicking(dc);
        try
        {
            this.draw(dc, true);
        }
        finally
        {
            this.pickSupport.endPicking(dc);
        }

        this.resolvePick(dc, pickPoint, layer);
        this.pickSupport.clearPickList();
    }

    protected java.awt.Rectangle computeTileViewport(DrawContext dc)
    {
        // Compute the smallest dimension of level set tile or pick tile.
        int minSize = Integer.MAX_VALUE;

        for (Level level : this.levelSet.getLevels())
        {
            if (minSize > level.getTileWidth())
                minSize = level.getTileWidth();
            if (minSize > level.getTileHeight())
                minSize = level.getTileHeight();
        }

        if (this.pickEnabled && this.havePickTile)
        {
            if (minSize > this.pickTile.getPreferredWidth())
                minSize = this.pickTile.getPreferredWidth();
            if (minSize > this.pickTile.getPreferredHeight())
                minSize = this.pickTile.getPreferredHeight();
        }

        // The viewport may be smaller than the desired dimensions. For that reason, we constrain the desired
        // dimensions by the viewport.
        java.awt.Rectangle viewport = dc.getView().getViewport();
        if (minSize > viewport.width)
            minSize = viewport.width;
        if (minSize > viewport.height)
            minSize = viewport.height;

        // The final dimension used to render all surface tiles will be the smallest square dimension which fits into
        // the viewport.
        int size = WWMath.powerOfTwoFloor(minSize);

        return new java.awt.Rectangle(viewport.x, viewport.y, size, size);
    }

    //**************************************************************//
    //********************  Tile Rendering  ************************//
    //**************************************************************//

    @SuppressWarnings({"UnusedDeclaration"})
    protected boolean canAssembleTiles(DrawContext dc)
    {
        return (this.currentTileViewport.getWidth() > 0) && (this.currentTileViewport.getHeight() > 0);
    }

    protected void draw(DrawContext dc, boolean isPickCall)
    {
        if (this.currentTiles.isEmpty())
            return;

        OGLStackHandler stackHandler = new OGLStackHandler();

        this.beginRendering(dc, stackHandler);
        try
        {
            this.drawTiles(dc, isPickCall);
        }
        finally
        {
            this.endRendering(dc, stackHandler);
        }
    }

    protected void drawTiles(DrawContext dc, boolean isPickCall)
    {
        if (isPickCall)
        {
            this.drawPickTiles(dc);
        }
        else
        {
            this.drawRenderTiles(dc);
        }
    }

    protected void drawRenderTiles(DrawContext dc)
    {
        dc.getGeographicSurfaceTileRenderer().setShowImageTileOutlines(this.showTileOutlines);
        dc.getGeographicSurfaceTileRenderer().renderTiles(dc, this.currentTiles);
    }

    protected void drawPickTiles(DrawContext dc)
    {
        // The surface object pick colors are rendered into the pickTile's texture during the preRender stage. We must
        // therefore render the pickTile's texture as though it was a typical RGBA texture, although the colors are
        // all pick colors or transparent black. Note that blending must be enabled to keep the transparent regions
        // from obscuring other pick colors.

        if (!this.havePickTile)
            return;

        boolean isPickingMode = dc.isPickingMode();
        if (isPickingMode)
        {
            dc.disablePickingMode();
        }

        dc.getGeographicSurfaceTileRenderer().setShowImageTileOutlines(false);
        dc.getGeographicSurfaceTileRenderer().renderTile(dc, this.pickTile);

        if (isPickingMode)
        {
            dc.enablePickingMode();
        }
    }

    protected void beginRendering(DrawContext dc, OGLStackHandler stackHandler)
    {
        GL gl = dc.getGL();

        OGLStateSupport stateSupport = new OGLStateSupport();
        stateSupport.setEnableAlphaTest(true);
        stateSupport.setEnableBlending(!dc.isPickingMode());
        stateSupport.setColorMode(OGLStateSupport.COLOR_PREMULTIPLIED_ALPHA);

        int attribBits =
            stateSupport.getAttributeBits()
            | GL.GL_COLOR_BUFFER_BIT
            | GL.GL_CURRENT_BIT
            | GL.GL_POLYGON_BIT;
        stackHandler.pushAttrib(gl, attribBits);

        stateSupport.apply(gl);

        gl.glPolygonMode(GL.GL_FRONT, GL.GL_FILL);
        gl.glEnable(GL.GL_CULL_FACE);
        gl.glCullFace(GL.GL_BACK);
    }

    protected void endRendering(DrawContext dc, OGLStackHandler stackHandler)
    {
        stackHandler.pop(dc.getGL());
    }

    protected PickedObject bindPickableObject(DrawContext dc, Object userObject, Object objectId)
    {
        java.awt.Color pickColor = dc.getUniquePickColor();
        int colorCode = pickColor.getRGB();
        dc.getGL().glColor3ub((byte) pickColor.getRed(), (byte) pickColor.getGreen(), (byte) pickColor.getBlue());

        PickedObject po = new PickedObject(colorCode, userObject);
        if (objectId != null)
        {
            po.setValue(AVKey.PICKED_OBJECT_ID, objectId);
        }

        this.pickSupport.addPickableObject(po);

        return po;
    }

    protected void resolvePick(DrawContext dc, java.awt.Point pickPoint, Layer layer)
    {
        if (pickPoint == null)
            return;

        PickedObject topObject = this.getTopPickedSurfaceObject(dc, pickPoint);
        if (topObject == null)
            return;

        // Surface objects are by definition on the geometry defined by the terrain. Therefore if no terrain has been
        // picked, we know that no surface object has been picked. But the converse is not true.
        Position terrainPosition = this.getPickedTerrainPosition(dc);
        if (terrainPosition != null)
        {
            topObject.setPosition(terrainPosition);
        }

        if (layer != null)
        {
            topObject.setParentLayer(layer);
        }

        dc.addPickedObject(topObject);
    }

    protected PickedObject getTopPickedSurfaceObject(DrawContext dc, java.awt.Point pickPoint)
    {
        PickedObject topPickedObject = this.pickSupport.getTopObject(dc, pickPoint);
        if (topPickedObject == null)
        {
            return null;
        }

        Object topObject = topPickedObject.getObject();
        if (topObject == null || !(topObject instanceof SurfaceObject))
        {
            return null;
        }

        return topPickedObject;
    }

    protected Position getPickedTerrainPosition(DrawContext dc)
    {
        PickedObject pickedTerrain = dc.getPickedObjects().getTerrainObject();
        if (pickedTerrain != null)
        {
            return pickedTerrain.getPosition();
        }

        View view = dc.getView();
        Globe globe = dc.getGlobe();
        if (view == null || globe == null)
        {
            return null;
        }

        java.awt.Point pickPoint = dc.getPickPoint();
        if (pickPoint == null)
        {
            return null;
        }

        Line pickRay = view.computeRayFromScreenPoint(pickPoint.getX(), pickPoint.getY());

        SectorGeometryList sectorGeometry = dc.getSurfaceGeometry();
        if (sectorGeometry != null)
        {
            Intersection[] intersection = sectorGeometry.intersect(pickRay);
            if (intersection != null && intersection.length != 0)
            {
                return globe.computePositionFromPoint(intersection[0].getIntersectionPoint());
            }
        }

        return globe.getIntersectionPosition(pickRay);
    }

    //**************************************************************//
    //********************  Surface Object Assembly  ***************//
    //**************************************************************//

    protected static class SurfaceObjectInfo
    {
        protected SurfaceObject surfaceObject;
        protected Iterable<? extends Sector> firstLevelSectors;

        public SurfaceObjectInfo(SurfaceObject object, Iterable<? extends Sector> firstLevelSectors)
        {
            if (object == null)
            {
                String message = Logging.getMessage("nullValue.ObjectIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            if (firstLevelSectors == null)
            {
                String message = Logging.getMessage("nullValue.IterableIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.surfaceObject = object;
            this.firstLevelSectors = firstLevelSectors;
        }

        public SurfaceObject getSurfaceObject()
        {
            return this.surfaceObject;
        }

        public Iterable<? extends Sector> getFirstLevelSectors()
        {
            return this.firstLevelSectors;
        }

        public boolean intersects(DrawContext dc, double texelSizeRadians, Sector sector)
        {
            if (dc == null)
            {
                String message = Logging.getMessage("nullValue.DrawContextIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            if (sector == null)
            {
                String message = Logging.getMessage("nullValue.SectorIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            for (Sector s : this.surfaceObject.getSectors(dc, texelSizeRadians))
            {
                if (s.intersects(sector))
                {
                    return true;
                }
            }

            return false;
        }
    }

    protected static class SurfaceObjectState
    {
        protected long lastModifiedTime;
        protected java.util.List<SurfaceObject> surfaceObjectList;

        public SurfaceObjectState(java.util.List<SurfaceObject> currentSurfaceObjects)
        {
            this.surfaceObjectList = currentSurfaceObjects;

            for (SurfaceObject surfaceObject : currentSurfaceObjects)
            {
                if (this.lastModifiedTime < surfaceObject.getLastModifiedTime())
                    this.lastModifiedTime = surfaceObject.getLastModifiedTime();
            }
        }

        public long getLastModifiedTime()
        {
            return this.lastModifiedTime;
        }

        public void setLastModifiedTime(long time)
        {
            this.lastModifiedTime = time;
        }

        public Iterable<? extends SurfaceObject> getSurfaceObjects()
        {
            return this.surfaceObjectList;
        }

        public boolean isStateNewerOrModified(SurfaceObjectState that)
        {
            //noinspection SimplifiableIfStatement
            if (that == null)
            {
                return false;
            }

            return (this.lastModifiedTime < that.lastModifiedTime)
                || !this.surfaceObjectList.equals(that.surfaceObjectList);
        }
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void clearSurfaceObjects(DrawContext dc)
    {
        this.currentSurfaceObjects.clear();
    }

    protected void assembleSurfaceObjects(DrawContext dc)
    {
        for (SurfaceObject surfaceObject : this.surfaceObjectIterable)
        {
            this.addSurfaceObject(dc, surfaceObject);
        }

        this.topLevelBoundingSector = this.computeCurrentBoundingSector();
    }

    protected void addSurfaceObject(DrawContext dc, SurfaceObject surfaceObject)
    {
        double firstLevelTexelSize = this.levelSet.getFirstLevel().getTexelSize();

        Iterable<? extends Sector> sectors = surfaceObject.getSectors(dc, firstLevelTexelSize);
        if (sectors == null)
            return;

        SurfaceObjectInfo objectInfo = new SurfaceObjectInfo(surfaceObject, sectors);
        this.currentSurfaceObjects.add(objectInfo);
    }
    
    protected Sector computeCurrentBoundingSector()
    {
        Sector boundingSector = null;

        for (SurfaceObjectInfo objectInfo : this.currentSurfaceObjects)
        {
            for (Sector sector : objectInfo.getFirstLevelSectors())
            {
                boundingSector = (boundingSector != null) ? boundingSector.union(sector) : sector;
            }
        }

        return boundingSector;
    }

    protected SurfaceObjectState getCurrentStateFor(DrawContext dc, Tile tile)
    {
        java.util.List<SurfaceObject> intersectingObjects = new java.util.ArrayList<SurfaceObject>();

        for (SurfaceObjectInfo objectInfo : this.currentSurfaceObjects)
        {
            if (objectInfo.intersects(dc, tile.getLevel().getTexelSize(), tile.getSector()))
            {
                intersectingObjects.add(objectInfo.getSurfaceObject());
            }
        }

        return new SurfaceObjectState(intersectingObjects);
    }

    //**************************************************************//
    //********************  Tile Assembly  *************************//
    //**************************************************************//

    @SuppressWarnings({"UnusedDeclaration"})
    protected void clearTiles(DrawContext dc)
    {
        this.currentTiles.clear();
    }

    protected void assembleTiles(DrawContext dc)
    {
        if (this.currentSurfaceObjects.isEmpty())
            return;

        if (this.topLevelBoundingSector == null || this.topLevelBoundingSector.equals(Sector.EMPTY_SECTOR))
            return;

        TextureTile[] topLevelTiles = this.assembleTopLevelTiles(this.levelSet, this.topLevelBoundingSector);
        for (TextureTile tile : topLevelTiles)
        {
            this.addTileOrDescendants(dc, tile);
        }

        dc.setPerFrameStatistic(PerformanceStatistic.IMAGE_TILE_COUNT, this.tileCountName,
            this.currentTiles.size());

        this.assemblePickTile(dc);
    }

    protected void assemblePickTile(DrawContext dc)
    {
        this.havePickTile = false;

        if (!this.pickEnabled)
            return;

        Position pickedTerrainPosition = this.getPickedTerrainPosition(dc);
        if (pickedTerrainPosition != null)
        {
            TextureTile pickedTile = this.getIntersectingTile(
                pickedTerrainPosition.getLatitude(), pickedTerrainPosition.getLongitude(), this.currentTiles);
            if (pickedTile != null)
            {
                this.pickTile.setSector(pickedTile.getSector());
                this.pickTile.setPreferredWidth(pickedTile.getLevel().getTileWidth());
                this.pickTile.setPreferredHeight(pickedTile.getLevel().getTileHeight());
                this.havePickTile = true;
            }
        }
    }

    protected void addTileOrDescendants(DrawContext dc, TextureTile tile)
    {
        if (!this.isTileVisible(dc, tile))
            return;

        if (!this.isTileNeeded(dc, tile))
            return;

        if (this.isTileMeetsRenderCriteria(dc, tile))
        {
            this.addTile(dc, tile);
            return;
        }

        Level nextLevel = this.levelSet.getLevel(tile.getLevelNumber() + 1);
        for (TextureTile subTile : tile.createSubTiles(nextLevel))
        {
            this.addTileOrDescendants(dc, subTile);
        }
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected void addTile(DrawContext dc, TextureTile tile)
    {
        this.currentTiles.add(tile);
    }

    protected boolean isTileVisible(DrawContext dc, SurfaceTile tile)
    {
        if (dc.getVisibleSector() != null && !dc.getVisibleSector().intersects(tile.getSector()))
        {
            return false;
        }

        //noinspection RedundantIfStatement
        if (!tile.getExtent(dc).intersects(dc.getView().getFrustumInModelCoordinates()))
        {
            return false;
        }

        return true;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected boolean isTileNeeded(DrawContext dc, Tile tile)
    {
        // If tile does not intersect any surface object's bounding sector, the tile is not needed.
        if (!tile.getSector().intersects(this.topLevelBoundingSector))
        {
            return false;
        }

        // If the tile falls completely outside the bounding sectors of all surface objects, then this tile is not
        // needed.
        boolean haveIntersection = false;
        for (SurfaceObjectInfo objectInfo : this.currentSurfaceObjects)
        {
            if (objectInfo.intersects(dc, tile.getLevel().getTexelSize(), tile.getSector()))
            {
                haveIntersection = true;
                break;
            }
        }

        return haveIntersection;
    }

    protected boolean isTileMeetsRenderCriteria(DrawContext dc, Tile tile)
    {
        Level level = tile.getLevel();

        if (this.levelSet.isFinalLevel(level.getLevelNumber()))
        {
            return true;
        }

        // Test the tile against screen metrics. This will attempt to keep the tile texel to screen pixel ratio
        // close to the user specified value. Substitude the current tile viewport for the level's width and height,
        // sinde the viewport dimensions are used to define the actual texture dimensions for this tile.
        return !this.needToSubdivide(dc, tile.getSector(),
            this.currentTileViewport.width, this.currentTileViewport.height);
    }

    protected TextureTile getIntersectingTile(Angle latitude, Angle longitude, Iterable<? extends TextureTile> tiles)
    {
        for (TextureTile tile : tiles)
        {
            if (tile.getSector().contains(latitude, longitude))
            {
                return tile;
            }
        }

        return null;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected boolean needToSubdivide(DrawContext dc, Sector sector, int width, int height)
    {
        // Compute a point in the sector that is nearest to the eye point. This point will be used to estimate the
        // distance from the eye to the sector.
        Vec4 nearestPoint = this.getNearestPointInSector(dc, sector);

        // Compute the size in meters of a texel on the surface sector given the data dimensions (width, height). If
        // the data dimension is zero, than replace it with 1, which effectively ignores the data dimension.
        double texelSize = nearestPoint.getLength3() * sector.getDeltaLatRadians()
            / (height <= 0 ? 1d : (double) height);

        // Compute the size in meters that a screen pixel would cover on the surface.
        double eyeToSurfaceDistance = dc.getView().getEyePoint().distanceTo3(nearestPoint);
        double pixelSize = dc.getView().computePixelSizeAtDistance(eyeToSurfaceDistance);

        // Compute the the ratio of surface texels to screen pixels.
        // * If ratio is 1, then a surface texel will be about the same size as a screen pixel.
        // * If ratio is less than 1, then a surface texel will be smaller than a screen pixel.
        // * If ratio is greater than 1, then a surface texel will be larger than a screen pixel.
        double ratio = texelSize / pixelSize;

        // Subdivide when the ratio of surface texels to screen pixels to is greater than the desired ratio.
        // Subdividing will create sub tiles with a smaller ratio; the the texel size will decrease while the pixel
        // size stays constant.
        return ratio > this.texelToPixelRatio;
    }

    protected Vec4 getNearestPointInSector(DrawContext dc, Sector sector)
    {
        Position eyePos = dc.getView().getEyePosition();
        if (sector.contains(eyePos))
        {
            Vec4 point = dc.getPointOnGlobe(eyePos.getLatitude(), eyePos.getLongitude());
            if (point == null)
            {
                double elev = dc.getGlobe().getElevation(eyePos.getLatitude(), eyePos.getLongitude())
                    * dc.getVerticalExaggeration();
                return dc.getGlobe().computePointFromPosition(eyePos.getLatitude(), eyePos.getLongitude(),
                    elev);
            }

            return point;
        }

        LatLon nearestCorner = null;
        Angle nearestDistance = Angle.fromDegrees(Double.MAX_VALUE);
        for (LatLon ll : sector)
        {
            Angle d = LatLon.greatCircleDistance(ll, eyePos);
            if (d.compareTo(nearestDistance) < 0)
            {
                nearestCorner = ll;
                nearestDistance = d;
            }
        }

        if (nearestCorner == null)
            return null;

        Vec4 point = dc.getPointOnGlobe(nearestCorner.getLatitude(), nearestCorner.getLongitude());
        if (point != null)
            return point;

        double elev = dc.getGlobe().getElevation(nearestCorner.getLatitude(), nearestCorner.getLongitude())
            * dc.getVerticalExaggeration();
        return dc.getGlobe().computePointFromPosition(nearestCorner.getLatitude(), nearestCorner.getLongitude(),
            elev);
    }

    protected TextureTile[] assembleTopLevelTiles(LevelSet levelSet, Sector sector)
    {
        Level level = levelSet.getFirstLevel();
        Angle dLat = level.getTileDelta().getLatitude();
        Angle dLon = level.getTileDelta().getLongitude();
        Angle latOrigin = levelSet.getTileOrigin().getLatitude();
        Angle lonOrigin = levelSet.getTileOrigin().getLongitude();

        // Determine the row and column offset from the common World Wind global tiling origin.
        int firstRow = Tile.computeRow(dLat, sector.getMinLatitude(), latOrigin);
        int firstCol = Tile.computeColumn(dLon, sector.getMinLongitude(), lonOrigin);
        int lastRow = Tile.computeRow(dLat, sector.getMaxLatitude(), latOrigin);
        int lastCol = Tile.computeColumn(dLon, sector.getMaxLongitude(), lonOrigin);

        int nLatTiles = lastRow - firstRow + 1;
        int nLonTiles = lastCol - firstCol + 1;
        int tileIndex = 0;
        TextureTile[] tiles = new TextureTile[nLatTiles * nLonTiles];

        Angle p1 = Tile.computeRowLatitude(firstRow, dLat, latOrigin);
        for (int row = firstRow; row <= lastRow; row++)
        {
            Angle p2;
            p2 = p1.add(dLat);

            Angle t1 = Tile.computeColumnLongitude(firstCol, dLon, lonOrigin);
            for (int col = firstCol; col <= lastCol; col++)
            {
                Angle t2;
                t2 = t1.add(dLon);

                TileKey tileKey = new TileKey(level.getLevelNumber(), row, col, level.getCacheName());
                TextureTile tile = this.getTileFromMemoryCache(tileKey);
                if (tile == null)
                {
                    tile = new TextureTile(new Sector(p1, p2, t1, t2), level, row, col);
                    TextureTile.getMemoryCache().add(tileKey, tile);
                }

                tiles[tileIndex++] = tile;

                t1 = t2;
            }
            p1 = p2;
        }

        return tiles;
    }

    protected TextureTile getTileFromMemoryCache(TileKey tileKey)
    {
        return (TextureTile) TextureTile.getMemoryCache().getObject(tileKey);
    }
    
    //**************************************************************//
    //********************  Tile Rasterization  ********************//
    //**************************************************************//

    protected void updateTiles(DrawContext dc)
    {
        if (this.currentTiles.isEmpty())
            return;

        OGLStackHandler stackHandler = new OGLStackHandler();
        this.beginUpdateTiles(dc, stackHandler, this.currentTileViewport);
        try
        {
            for (TextureTile tile : this.currentTiles)
            {
                this.updateTile(dc, tile);
            }

            // Must update the pick tile data every frame.
            if (this.pickEnabled && this.havePickTile)
            {
                this.updatePickTile(dc, this.pickTile);
            }
        }
        finally
        {
            this.endUpdateTiles(dc, stackHandler);
        }
    }

    protected void beginUpdateTiles(DrawContext dc, OGLStackHandler stackHandler, java.awt.Rectangle viewport)
    {
        GL gl = dc.getGL();

        this.renderToTextureSupport.beginRendering(gl, viewport.x, viewport.y, viewport.width, viewport.height);

        // Setup GL to use blending with non premultiplied alpha colors. The values actually written to the framebuffer
        // will have their color components multiplied by alpha. However, to use GL line or polygon blending, GL is
        // expecting to work with non premultiplied colors.
        OGLStateSupport stateSupport = new OGLStateSupport();
        stateSupport.setEnableAlphaTest(true);
        stateSupport.setEnableBlending(true);
        stateSupport.setColorMode(OGLStateSupport.COLOR_NO_PREMULTIPLIED_ALPHA);

        stackHandler.pushAttrib(gl,
            stateSupport.getAttributeBits()
            | GL.GL_DEPTH_BUFFER_BIT // For depth test.
            | GL.GL_TEXTURE_BIT); // For texture bind.

        stateSupport.apply(gl);

        gl.glDisable(GL.GL_DEPTH_TEST);
    }

    protected void endUpdateTiles(DrawContext dc, OGLStackHandler stackHandler)
    {
        GL gl = dc.getGL();

        stackHandler.pop(gl);

        this.renderToTextureSupport.endRendering(gl);
    }

    protected void updateTile(DrawContext dc, TextureTile tile)
    {
        SurfaceObjectState currentTileState = this.getCurrentStateFor(dc, tile);
        if (this.isTileCurrent(dc, tile, currentTileState))
        {
            return;
        }

        this.updateTileData(dc, tile, currentTileState);
        
        currentTileState.setLastModifiedTime(System.currentTimeMillis());
        this.setTileState(tile, currentTileState);
    }

    protected void updatePickTile(DrawContext dc, PickTile tile)
    {
        this.updatePickTileData(dc, tile);
    }

    protected boolean isTileCurrent(DrawContext dc, TextureTile tile, SurfaceObjectState state)
    {
        SurfaceObjectState currentState = this.getTileState(tile);
        //noinspection SimplifiableIfStatement
        if (currentState == null || currentState.isStateNewerOrModified(state))
        {
            return false;
        }
        
        return tile.getTexture(dc.getTextureCache()) != null;
    }

    protected SurfaceObjectState getTileState(TextureTile tile)
    {
        return this.tileStateMap.get(tile.getTileKey());
    }

    protected void setTileState(TextureTile tile, SurfaceObjectState state)
    {
        this.tileStateMap.put(tile.getTileKey(), state);
    }

    protected void clearAllTileState()
    {
        this.tileStateMap.clear();
    }

    protected void updateTileData(DrawContext dc, TextureTile tile, SurfaceObjectState state)
    {
        GL gl = dc.getGL();

        java.awt.Rectangle viewport = this.renderToTextureSupport.getViewport();

        // Clear the current framebuffer contents.
        this.renderToTextureSupport.clear(gl, this.tileBackgroundColor);

        // Render each intersecting object into the framebuffer.
        for (SurfaceObject surfaceObject : state.getSurfaceObjects())
        {
            surfaceObject.renderToRegion(dc, tile.getSector(), viewport.x, viewport.y, viewport.width, viewport.height);
        }

        // Copy the framebuffer contents to the tile.
        this.copyFramebufferContentsToTile(dc, tile, viewport.x, viewport.y, viewport.width, viewport.height,
            this.useMipmaps);
    }

    protected void updatePickTileData(DrawContext dc, PickTile tile)
    {
        GL gl = dc.getGL();

        java.awt.Rectangle viewport = this.renderToTextureSupport.getViewport();
        
        // Temporarily force the DrawContext into picking mode while the surface objects are rendered into the pick
        // tile.
        boolean isPicking = dc.isPickingMode();
        if (!isPicking)
        {
            dc.enablePickingMode();
        }

        this.pickSupport.beginPicking(dc);

        // Clear the current framebuffer contents.
        this.renderToTextureSupport.clear(gl, new java.awt.Color(0, 0, 0, 0));

        // Render each intersecting object into the framebuffer with a unique pick color.
        for (SurfaceObjectInfo objectInfo : this.currentSurfaceObjects)
        {
            if (objectInfo.intersects(dc, tile.getTexelSize(), tile.getSector()))
            {
                this.bindPickableObject(dc, objectInfo.getSurfaceObject(), null);
                objectInfo.getSurfaceObject().renderToRegion(dc, tile.getSector(), viewport.x, viewport.y,
                    viewport.width, viewport.height);
            }
        }

        // Copy the framebuffer contents to the tile.
        this.copyFramebufferContentsToTile(dc, tile.getDelegate(), viewport.x, viewport.y,
            viewport.width, viewport.height, false);

        this.pickSupport.endPicking(dc);

        // Restore the DrawContext's previous picking mode state.
        if (!isPicking)
        {
            dc.disablePickingMode();
        }
    }

    protected void copyFramebufferContentsToTile(DrawContext dc, TextureTile tile, int x, int y, int width, int height,
        boolean useMipmaps)
    {
        // Update the tile texture with the current framebuffer contents. If the tile's dimensions do not match the
        // framebuffer viewport's dimensions, or if the tile does not have a texture, then new texture data with the
        // appropriate dimensions.
        Texture texture = tile.getTexture(dc.getTextureCache());
        boolean hasTexture = (texture != null) && (texture.getWidth() == width) && (texture.getHeight() == height);
        if (!hasTexture)
        {
            // TODO: work with texture tile to refresh the texture data
            dc.getTextureCache().remove(tile.getTileKey());
            tile.setTextureData(this.createTileTextureData(width, height, useMipmaps));
        }

        // Bind the tile. If necessary this will transfer any in-memory texture data to OpenGL, or create the texture
        // object. Then this will make this tile's texture the current active texture object.
        if (!tile.bind(dc))
            return;

        texture = tile.getTexture(dc.getTextureCache());
        if (texture == null)
        {
            String message = Logging.getMessage("nullValue.TextureIsNull");
            Logging.logger().warning(message);
            return;
        }

        GL gl = dc.getGL();
        gl.glEnable(GL.GL_TEXTURE_2D);

        if (hasTexture)
        {
            dc.getGL().glCopyTexSubImage2D(
                texture.getTarget(), // target
                0,                   // level
                0, 0,                // xoffset, yoffset
                x, y,                // x, y
                width, height);      // width, height
        }
        else
        {
            dc.getGL().glCopyTexImage2D(
                texture.getTarget(), // target
                0,                   // level
                GL.GL_RGBA,          // internal format
                x, y,                // x, y
                width, height,       // width, height
                0);                  // border
        }

        gl.glDisable(GL.GL_TEXTURE_2D);
    }

    protected TextureData createTileTextureData(int width, int height, boolean useMipmaps)
    {
        return new TextureData(
            GL.GL_RGBA,          // internal format
            width, height,       // width, height
            0,                   // border
            GL.GL_RGBA,          // pixel format
            GL.GL_UNSIGNED_BYTE, // pixel type
            useMipmaps,          // mipmap
            false, false,        // dataIsCompressed, mustFlipVertically
            null, null);       // buffer, flusher
    }

    //**************************************************************//
    //********************  Pick Tile  *****************************//
    //**************************************************************//

    protected static class PickTile implements SurfaceTile
    {
        protected Sector sector;
        protected int preferredWidth;
        protected int preferredHeight;
        protected TextureTile delegate;
        private Extent extent = null; // bounding volume
        private Object globeStateKey;

        public PickTile(Sector sector)
        {
            this.sector = sector;
            this.delegate = new TextureTile(sector);
        }

        public Sector getSector()
        {
            return this.sector;
        }

        public void setSector(Sector sector)
        {
            this.sector = sector;
            this.extent = null;
        }

        public int getPreferredWidth()
        {
            return this.preferredWidth;
        }

        public void setPreferredWidth(int preferredWidth)
        {
            this.preferredWidth = preferredWidth;
        }

        public int getPreferredHeight()
        {
            return this.preferredHeight;
        }

        public void setPreferredHeight(int preferredHeight)
        {
            this.preferredHeight = preferredHeight;
        }

        public TextureTile getDelegate()
        {
            return this.delegate;
        }

        public double getTexelSize()
        {
            return this.sector.getDeltaLatRadians() / this.preferredHeight;
        }

        public boolean bind(DrawContext dc)
        {
            return this.delegate.bind(dc);
        }

        public void applyInternalTransform(DrawContext dc)
        {
            this.delegate.applyInternalTransform(dc);
        }

        public java.util.List<? extends LatLon> getCorners()
        {
            return this.sector.asList();
        }

        public Extent getExtent(DrawContext dc)
        {
            if (dc == null)
            {
                String message = Logging.getMessage("nullValue.DrawContextIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            if (this.extent == null || !this.isExtentValid(dc))
            {
                this.extent = dc.getGlobe().computeBoundingCylinder(dc.getVerticalExaggeration(), this.getSector());
                this.globeStateKey = dc.getGlobe().getStateKey(dc);
            }

            return this.extent;
        }

        protected boolean isExtentValid(DrawContext dc)
        {
            return !(dc.getGlobe() == null || this.globeStateKey == null)
                && this.globeStateKey.equals(dc.getGlobe().getStateKey(dc));
        }
    }
}
