/*
Copyright (C) 2001, 2009 United States Government
as represented by the Administrator of the
National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.terrain;

import gov.nasa.worldwind.*;
import gov.nasa.worldwind.avlist.*;
import gov.nasa.worldwind.exception.*;
import gov.nasa.worldwind.globes.ElevationModel;
import gov.nasa.worldwind.util.*;
import gov.nasa.worldwind.wms.Capabilities;
import org.w3c.dom.*;

import javax.xml.xpath.XPathConstants;

/**
 * A factory that creates {@link gov.nasa.worldwind.globes.ElevationModel} instances.
 *
 * @author tag
 * @version $Id: BasicElevationModelFactory.java 11667 2009-06-16 15:50:18Z dcollins $
 */
public class BasicElevationModelFactory extends BasicFactory
{
    /**
     * Creates an elevation model from a configuration file.
     * <p/>
     * Maps the <code>modelType</code> attribute of the <code>ElevationModel</code> element of the XML configuration
     * file to the appropriate elevation model type. Elevation model types recognized are: <ul> <li>"Compound" for
     * models composed of other models.</li> <li>Any other value defaults to a default elevation model
     * implementation.</li> </ul>
     * <p/>
     * For non-compound models, this maps the <code>serviceName</code> attribute of the
     * <code>ElevationModel/Service</code> element of the XML configuration file to the appropriate base elevation model
     * type. Service types recognized are:" <ul> <li>"WMS" for elevation models that draw their data from a WMS web
     * service.</li> <li>"WWTileService" for elevation models that draw their data from a World Wind tile service.</li>
     * <li>"Offline" for elevation models that draw their data only from the local cache.</li> </ul>
     *
     * @param fileName the path to the configuration file. It must be either an absolute path or a relative path
     *                 available on the classpath.
     *
     * @return an elevation model.
     *
     * @throws IllegalArgumentException if the configuration file name is null.
     * @throws WWUnrecognizedException  if the service type is unrecognized.
     * @throws WWRuntimeException       if object creation fails. The exception indicating the source of the failure is
     *                                  included as the {@link Exception#initCause(Throwable)}.
     */
    public ElevationModel createFromConfigFile(String fileName)
    {
        if (fileName == null)
        {
            String message = Logging.getMessage("nullValue.FilePathIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        ElevationModel model;

        try
        {
            Document doc = WWXML.openDocumentFile(fileName, this.getClass());

            model = this.doCreateFromDocument(doc.getDocumentElement(), null);
        }
        catch (IllegalStateException e)
        {
            // Message is passed up the stack as an IllegalStateException, caught here, then thrown as a
            // WWUnrecognizedException.
            throw new WWUnrecognizedException(e.getMessage());
        }
        catch (Exception e)
        {
            String msg = Logging.getMessage("generic.CreationFromConfigurationFileFailed", fileName);
            throw new WWRuntimeException(msg, e);
        }

        if (model == null)
        {
            String msg = Logging.getMessage("generic.UnrecognizedDocument", fileName);
            throw new WWUnrecognizedException(msg);
        }

        return model;
    }

    /**
     * Creates an elevation model from a data configuration.
     * <p/>
     * Maps the <code>ElevationModel/modelType</code> configuration parameter to the appropriate model type. Elevation
     * model types recognized are: <ul> <li>"Compound" for models composed of other models.</li> <li>Any other value
     * defaults to a default elevation model implementation.</li> </ul>
     * <p/>
     * For non-compound models, this maps the <code>ElevationModel/Service/serviceName</code> configuration property to
     * the appropriate base elevation model type. Service types recognized are: <ul> <li>"WMS" for elevation models that
     * draw their data from a WMS web service.</li> <li>"WWTileService" for elevation models that draw their data from a
     * World Wind tile service.</li> <li>"Offline" for elevation models that draw their data only from the local
     * cache.</li> </ul>
     *
     * @param dataConfig the configuration information describing the component.
     * @param params     key-value parameters which override or supplement the information provided in the specified
     *                   configuration. A null reference is permitted.
     *
     * @return an elevation model.
     *
     * @throws IllegalArgumentException if the data configuration is null.
     * @throws WWUnrecognizedException  if the service type is unrecognized.
     * @throws WWRuntimeException       if object creation fails. The exception indicating the source of the failure is
     *                                  included as the {@link Exception#initCause(Throwable)}.
     */
    public Object createFromDataConfig(DataConfiguration dataConfig, AVList params)
    {
        if (dataConfig == null)
        {
            String message = Logging.getMessage("nullValue.DataConfigurationIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        ElevationModel model;

        try
        {
            model = this.doCreateFromDataConfig(dataConfig, params);
        }
        catch (IllegalStateException e)
        {
            // Message is passed up the stack as an IllegalStateException, caught here, then thrown as a
            // WWUnrecognizedException.
            throw new WWUnrecognizedException(e.getMessage());
        }
        catch (Exception e)
        {
            String msg = Logging.getMessage("generic.CreationFromDataConfigurationFailed", dataConfig);
            throw new WWRuntimeException(msg, e);
        }

        if (model == null)
        {
            String msg = Logging.getMessage("generic.UnrecognizedDataConfiguration", dataConfig);
            throw new WWUnrecognizedException(msg);
        }

        return model;
    }

    protected ElevationModel doCreateFromDataConfig(DataConfiguration dataConfig, AVList params) throws Exception
    {
        String type = dataConfig.getType();

        if (type.equals("Capabilities"))
        {
            Capabilities caps = (Capabilities) dataConfig.getSource();
            return this.doCreateFromCapabilities(caps, params);
        }
        else if (type.equals("ElevationModel"))
        {
            Element domElement = (Element) dataConfig.getSource();
            return this.doCreateFromDocument(domElement, params);
        }

        return null;
    }

    protected ElevationModel doCreateFromCapabilities(Capabilities caps, AVList params)
    {
        String serviceName = caps.getServiceName();
        if (serviceName == null || !(serviceName.equalsIgnoreCase(Capabilities.WMS_SERVICE_NAME)
            || serviceName.equalsIgnoreCase("WMS")))
        {
            String message = Logging.getMessage("WMS.NotWMSService", serviceName != null ? serviceName : "null");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        if (params == null)
            params = new AVListImpl();

        if (params.getStringValue(AVKey.LAYER_NAMES) == null)
        {
            Element[] namedLayers = caps.getNamedLayers();

            if (namedLayers == null || namedLayers.length == 0)
            {
                String message = Logging.getMessage("WMS.NoLayersFound");
                Logging.logger().severe(message);
                throw new IllegalStateException(message);
            }

            // Use the first named layer if no other guidance given
            params.setValue(AVKey.LAYER_NAMES, caps.getLayerName(namedLayers[0]));
        }

        return new WMSBasicElevationModel(caps, params);
    }

    protected ElevationModel doCreateFromDocument(Element domElement, AVList params) throws Exception
    {
        NodeList emNodes = (NodeList) WWXML.makeXPath().evaluate("//ElevationModel", domElement,
            XPathConstants.NODESET);

        if (emNodes == null || emNodes.getLength() == 0)
            throw new IllegalStateException(); // caught and translated in below catch clause

        ElevationModel topModel = null;
        CompoundElevationModel compoundModel = null;

        for (int i = 0; i < emNodes.getLength(); i++)
        {
            Element root = (Element) emNodes.item(i);
            ElevationModel em;

            String modelType = WWXML.getText(root, "@modelType");
            if (modelType != null && modelType.equalsIgnoreCase("Compound"))
            {
                CompoundElevationModel newModel = this.createCompoundModel(root, params);
                if (compoundModel == null)
                {
                    compoundModel = newModel;
                    topModel = compoundModel;
                }
                else
                {
                    // make the new model a child of the current model and make it the new current model
                    compoundModel.addElevationModel(newModel);
                    compoundModel = newModel;
                }
                continue;
            }

            em = this.createDefaultModel(root, params);

            if (compoundModel == null)
                return em; // any compound models must be listed before the first non-compound model
            else
                compoundModel.addElevationModel(em);
        }

        return topModel;
    }

    @SuppressWarnings({"UnusedDeclaration"})
    protected CompoundElevationModel createCompoundModel(Element domElement, AVList params)
    {
        return new CompoundElevationModel();
    }

    protected ElevationModel createDefaultModel(Element domElement, AVList params)
    {
        ElevationModel em;

        String serviceName = WWXML.getText(domElement, "Service/@serviceName");

        if (serviceName.equals("Offline"))
        {
            em = new BasicElevationModel(domElement, params);
        }
        else if (serviceName.equals("WWTileService"))
        {
            em = new BasicElevationModel(domElement, params);
        }
        else if (serviceName.equals(Capabilities.WMS_SERVICE_NAME))
        {
            em = new WMSBasicElevationModel(domElement, params);
        }
        else
        {
            String msg = Logging.getMessage("generic.UnrecognizedServiceName", serviceName);
            throw new IllegalStateException(msg);
        }

        return em;
    }
}
