/*
Copyright (C) 2001, 2008 United States Government
as represented by the Administrator of the
National Aeronautics and Space Administration.
All Rights Reserved.
*/

package gov.nasa.worldwind.util;

import gov.nasa.worldwind.avlist.AVList;
import gov.nasa.worldwind.avlist.AVKey;

import java.nio.*;

/**
 * @author tag
 * @version $Id: BufferWrapper.java 9704 2009-03-27 00:52:15Z tgaskins $
 */
public abstract class BufferWrapper
{
    public abstract int length();

    public abstract int getInt(int index);

    public abstract double getDouble(int index);

    public abstract BufferWrapper getInt(int index, int[] dest, int destPos, int length);

    public abstract BufferWrapper getDouble(int index, double[] dest, int destPos, int length);

    public abstract BufferWrapper putInt(int index, int value);

    public abstract BufferWrapper putDouble(int index, double value);

    public abstract BufferWrapper putInt(int index, int[] src, int srcPos, int length);

    public abstract BufferWrapper putDouble(int index, double[] src, int srcPos, int length);

    public abstract long getSizeInBytes();

    //**************************************************************//
    //********************  Primitive Array Wrappers  **************//
    //**************************************************************//

    public static class IntArrayWrapper extends BufferWrapper
    {
        private int[] source;
        private int length;

        public IntArrayWrapper(int[] source, int length)
        {
            if (source == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.source = source;
            this.length = length;
        }

        public int length()
        {
            return this.length;
        }

        public int getInt(int index)
        {
            return this.source[index];
        }

        public double getDouble(int index)
        {
            return this.source[index];
        }

        public BufferWrapper getInt(int index, int[] dest, int destPos, int length)
        {
            System.arraycopy(this.source, index, dest, destPos, length);
            return this;
        }

        public BufferWrapper getDouble(int index, double[] dest, int destPos, int length)
        {
            for (int i = 0; i < length; i++)
                dest[destPos + i] = this.source[index + i];
            return this;
        }

        public BufferWrapper putInt(int index, int value)
        {
            this.source[index] = value;
            return this;
        }

        public BufferWrapper putDouble(int index, double value)
        {
            this.source[index] = (int) value;
            return this;
        }

        public BufferWrapper putInt(int index, int[] src, int srcPos, int length)
        {
            System.arraycopy(src, srcPos, this.source, index, length);
            return this;
        }

        public BufferWrapper putDouble(int index, double[] src, int srcPos, int length)
        {
            for (int i = 0; i < length; i++)
                this.source[index + i] = (int) src[srcPos + i];
            return this;
        }

        public long getSizeInBytes()
        {
            return (Integer.SIZE / 8) * this.length;
        }
    }

    //**************************************************************//
    //********************  NIO Buffer Wrappers  *******************//
    //**************************************************************//

    public abstract static class NIOBufferWrapper extends BufferWrapper
    {
        private Buffer sourceBuffer;

        public NIOBufferWrapper(Buffer source)
        {
            if (source == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.sourceBuffer = source;
            this.sourceBuffer.rewind();
        }

        public int length()
        {
            return this.sourceBuffer.limit();
        }
    }

    public static class ByteBufferWrapper extends NIOBufferWrapper
    {
        private ByteBuffer byteBuffer;

        public ByteBufferWrapper(ByteBuffer source)
        {
            super(source);
            this.byteBuffer = source;
        }

        public int getInt(int index)
        {
            return this.byteBuffer.get(index);
        }

        public double getDouble(int index)
        {
            return this.byteBuffer.get(index);
        }

        public BufferWrapper getInt(int index, int[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            byte[] buffer = new byte[length];
            this.byteBuffer.position(index);
            this.byteBuffer.get(buffer, 0, length);
            this.byteBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = buffer[i];

            return this;
        }

        public BufferWrapper getDouble(int index, double[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            byte[] buffer = new byte[length];
            this.byteBuffer.position(index);
            this.byteBuffer.get(buffer, 0, length);
            this.byteBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = buffer[i];

            return this;
        }

        public BufferWrapper putInt(int index, int value)
        {
            this.byteBuffer.put(index, (byte) value);
            return this;
        }

        public BufferWrapper putDouble(int index, double value)
        {
            this.byteBuffer.put(index, (byte) value);
            return this;
        }

        public BufferWrapper putInt(int index, int[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            byte[] buffer = new byte[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (byte) src[i + srcPos];

            this.byteBuffer.position(index);
            this.byteBuffer.put(buffer, 0, length);
            this.byteBuffer.rewind();
            return this;
        }

        public BufferWrapper putDouble(int index, double[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            byte[] buffer = new byte[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (byte) src[i + srcPos];

            this.byteBuffer.position(index);
            this.byteBuffer.put(buffer, 0, length);
            this.byteBuffer.rewind();
            return this;
        }

        public long getSizeInBytes()
        {
            return this.byteBuffer.capacity();
        }
    }

    public static class ShortBufferWrapper extends NIOBufferWrapper
    {
        private ShortBuffer shortBuffer;

        public ShortBufferWrapper(ShortBuffer source)
        {
            super(source);
            this.shortBuffer = source;
        }

        public int getInt(int index)
        {
            return this.shortBuffer.get(index);
        }

        public double getDouble(int index)
        {
            return this.shortBuffer.get(index);
        }

        public BufferWrapper getInt(int index, int[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            short[] buffer = new short[length];
            this.shortBuffer.position(index);
            this.shortBuffer.get(buffer, 0, length);
            this.shortBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = buffer[i];

            return this;
        }

        public BufferWrapper getDouble(int index, double[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            short[] buffer = new short[length];
            this.shortBuffer.position(index);
            this.shortBuffer.get(buffer, 0, length);
            this.shortBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = buffer[i];

            return this;
        }

        public BufferWrapper putInt(int index, int value)
        {
            this.shortBuffer.put(index, (short) value);
            return this;
        }

        public BufferWrapper putDouble(int index, double value)
        {
            this.shortBuffer.put(index, (short) value);
            return this;
        }

        public BufferWrapper putInt(int index, int[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            short[] buffer = new short[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (short) src[i + srcPos];

            this.shortBuffer.position(index);
            this.shortBuffer.put(buffer, 0, length);
            this.shortBuffer.rewind();
            return this;
        }

        public BufferWrapper putDouble(int index, double[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            short[] buffer = new short[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (short) src[i + srcPos];

            this.shortBuffer.position(index);
            this.shortBuffer.put(buffer, 0, length);
            this.shortBuffer.rewind();
            return this;
        }

        public long getSizeInBytes()
        {
            return (Short.SIZE / 8) * this.shortBuffer.capacity();
        }
    }

    public static class IntBufferWrapper extends NIOBufferWrapper
    {
        private IntBuffer intBuffer;

        public IntBufferWrapper(IntBuffer source)
        {
            super(source);
            this.intBuffer = source;
        }

        public int getInt(int index)
        {
            return this.intBuffer.get(index);
        }

        public double getDouble(int index)
        {
            return this.intBuffer.get(index);
        }

        public BufferWrapper getInt(int index, int[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.intBuffer.position(index);
            this.intBuffer.get(dest, destPos, length);
            this.intBuffer.rewind();

            return this;
        }

        public BufferWrapper getDouble(int index, double[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            int[] buffer = new int[length];
            this.intBuffer.position(index);
            this.intBuffer.get(buffer, 0, length);
            this.intBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = buffer[i];

            return this;
        }

        public BufferWrapper putInt(int index, int value)
        {
            this.intBuffer.put(index, value);
            return this;
        }

        public BufferWrapper putDouble(int index, double value)
        {
            this.intBuffer.put(index, (int) value);
            return this;
        }

        public BufferWrapper putInt(int index, int[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            this.intBuffer.position(index);
            this.intBuffer.put(src, srcPos, length);
            this.intBuffer.rewind();
            return this;
        }

        public BufferWrapper putDouble(int index, double[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            int[] buffer = new int[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (int) src[i + srcPos];

            this.intBuffer.position(index);
            this.intBuffer.put(buffer, 0, length);
            this.intBuffer.rewind();
            return this;
        }

        public long getSizeInBytes()
        {
            return (Integer.SIZE / 8) * this.intBuffer.capacity();
        }
    }
    
    public static class FloatBufferWrapper extends NIOBufferWrapper
    {
        private FloatBuffer floatBuffer;

        public FloatBufferWrapper(FloatBuffer source)
        {
            super(source);
            this.floatBuffer = source;
        }

        public int getInt(int index)
        {
            return (int) this.floatBuffer.get(index);
        }

        public double getDouble(int index)
        {
            return this.floatBuffer.get(index);
        }

        public BufferWrapper getInt(int index, int[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            float[] buffer = new float[length];
            this.floatBuffer.position(index);
            this.floatBuffer.get(buffer, 0, length);
            this.floatBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = (int) buffer[i];

            return this;
        }

        public BufferWrapper getDouble(int index, double[] dest, int destPos, int length)
        {
            if (dest == null)
            {
                String message = Logging.getMessage("nullValue.DestinationIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }
            
            float[] buffer = new float[length];
            this.floatBuffer.position(index);
            this.floatBuffer.get(buffer, 0, length);
            this.floatBuffer.rewind();

            for (int i = 0; i < length; i++)
                dest[i + destPos] = buffer[i];

            return this;
        }

        public BufferWrapper putInt(int index, int value)
        {
            this.floatBuffer.put(index, (float) value);
            return this;
        }

        public BufferWrapper putDouble(int index, double value)
        {
            this.floatBuffer.put(index, (float) value);
            return this;
        }

        public BufferWrapper putInt(int index, int[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            float[] buffer = new float[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (float) src[i + srcPos];

            this.floatBuffer.position(index);
            this.floatBuffer.put(buffer, 0, length);
            this.floatBuffer.rewind();
            return this;
        }

        public BufferWrapper putDouble(int index, double[] src, int srcPos, int length)
        {
            if (src == null)
            {
                String message = Logging.getMessage("nullValue.SourceIsNull");
                Logging.logger().severe(message);
                throw new IllegalArgumentException(message);
            }

            float[] buffer = new float[length];
            for (int i = 0; i < length; i++)
                buffer[i] = (float) src[i + srcPos];

            this.floatBuffer.position(index);
            this.floatBuffer.put(buffer, 0, length);
            this.floatBuffer.rewind();
            return this;
        }

        public long getSizeInBytes()
        {
            return (Float.SIZE / 8) * this.floatBuffer.capacity();
        }
    }

    public static BufferWrapper wrap(ByteBuffer byteBuffer, AVList parameters)
    {
        if (byteBuffer == null)
        {
            String message = Logging.getMessage("nullValue.ByteBufferIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }
        if (parameters == null)
        {
            String message = Logging.getMessage("nullValue.ParametersIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        String message = validate(parameters);
        if (message != null)
        {
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Object o = parameters.getValue(AVKey.BYTE_ORDER);
        byteBuffer.order(AVKey.LITTLE_ENDIAN.equals(o) ? ByteOrder.LITTLE_ENDIAN : ByteOrder.BIG_ENDIAN);

        o = parameters.getValue(AVKey.DATA_TYPE);
        BufferWrapper wrapper = null;
        if (AVKey.INT8.equals(o))
            wrapper = new ByteBufferWrapper(byteBuffer);
        else if (AVKey.INT16.equals(o))
            wrapper = new ShortBufferWrapper(byteBuffer.asShortBuffer());
        else if (AVKey.INT32.equals(o))
            wrapper = new IntBufferWrapper(byteBuffer.asIntBuffer());
        else if (AVKey.FLOAT32.equals(o))
            wrapper = new FloatBufferWrapper(byteBuffer.asFloatBuffer());

        return wrapper;
    }

    private static String validate(AVList params)
    {
        StringBuilder sb = new StringBuilder();

        Object o = params.getValue(AVKey.BYTE_ORDER);
        if (o == null || !validateByteOrder(o))
            sb.append(sb.length() > 0 ? ", " : "").append(Logging.getMessage("term.byteOrder"));

        o = params.getValue(AVKey.DATA_TYPE);
        if (o == null || !validateDataType(o))
            sb.append(sb.length() > 0 ? ", " : "").append(Logging.getMessage("term.dataType"));

        if (sb.length() == 0)
            return null;

        return Logging.getMessage("BufferWrapper.InvalidBufferParameters", sb.toString());
    }

    private static boolean validateByteOrder(Object o)
    {
        return AVKey.BIG_ENDIAN.equals(o)
            || AVKey.LITTLE_ENDIAN.equals(o);
    }

    private static boolean validateDataType(Object o)
    {
        return AVKey.INT8.equals(o)
            || AVKey.INT16.equals(o)
            || AVKey.INT32.equals(o)
            || AVKey.FLOAT32.equals(o);
    }
}
