/*
Copyright (C) 2001, 2008 United States Government
as represented by the Administrator of the
National Aeronautics and Space Administration.
All Rights Reserved.
*/

package gov.nasa.worldwind.util;

import gov.nasa.worldwind.geom.Sector;
import gov.nasa.worldwind.geom.LatLon;
import gov.nasa.worldwind.render.markers.Marker;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;

/**
 * @author tag
 * @version $Id: SpatialTree.java 6036 2008-08-18 15:58:14Z tgaskins $
 */
public abstract class SpatialTree<S, T>
{
    protected int depth;
    protected int maxDepth;
    protected boolean addToAllLevels;
    protected S coverage;
    protected List<S> subCoverage;
    protected ArrayList<SpatialTree<S, T>> children;
    protected List<T> items = new ArrayList<T>();

    protected abstract boolean intersects(S extent, T item);

    protected abstract boolean intersectsCoverage(S extent1, S extent2);

    protected abstract SpatialTree<S, T> createInstance(S extent, int depth, int maxDepth, boolean addToAllLevels);

    protected abstract boolean contains(S extent, LatLon location);

    protected abstract ArrayList<S> subdivide(S extent, int rows, int cols);

    public SpatialTree(S coverage, int depth, int maxDepth, boolean addToAllLevels) // TODO: arg check
    {
        this(coverage, depth, maxDepth, addToAllLevels, 2, 2);
    }

    public SpatialTree(S coverage, int depth, int maxDepth, boolean addToAllLevels, int rows,
                       int cols) // TODO: arg check
    {
        this.coverage = coverage;
        this.depth = depth;
        this.maxDepth = maxDepth;
        this.addToAllLevels = addToAllLevels;

        if (depth < maxDepth)
        {
            this.subCoverage = this.subdivide(this.coverage, rows, cols);
            children = new ArrayList<SpatialTree<S, T>>(rows * cols);
            for (int i = 0; i < rows * cols; i++)
                children.add(null);
        }
    }

    public boolean add(T item)
    {
        boolean success = false;

        if (depth == maxDepth || addToAllLevels)
        {
            success = maxDepthAdd(item);
        }

        if (depth < maxDepth)
        {
            success = normalAdd(item) || success;
        }

        return success;
    }

    protected boolean normalAdd(T item)
    {
        boolean success = false;

        for (int i = 0; i < subCoverage.size() && !success; i++)
        {
            if (this.intersects(subCoverage.get(i), item))
            {
                if (children.get(i) == null)
                {
                    children.set(i, this.createInstance(subCoverage.get(i), depth + 1, maxDepth, addToAllLevels));
                }
                success = children.get(i).add(item);
            }
        }

        return success;
    }

    protected boolean maxDepthAdd(T item)
    {
        return this.intersects(coverage, item) && items.add(item);
    }

    public boolean remove(Sector itemRegion, T item) // TODO: untested, arg check
    {
        boolean success = false;

        if (depth == maxDepth || addToAllLevels)
        {
            success = maxDepthRemove(item);
        }

        if (depth < maxDepth)
        {
            success = success && normalRemove(itemRegion, item);
        }

        return success;
    }

    protected boolean maxDepthRemove(T item) // TODO: untested
    {
        return items.remove(item);
    }

    protected boolean normalRemove(Sector itemRegion, T item) // TODO: untested
    {
        boolean success = false;
        for (int i = 0; i < subCoverage.size(); i++)
        {
            if (this.intersects(subCoverage.get(i), item))
            {
                if (children.get(i) == null)
                {
                    children.set(i, this.createInstance(subCoverage.get(i), depth + 1, maxDepth, addToAllLevels));
                }
                success = success || children.get(i).remove(itemRegion, item);
            }
        }
        return success;
    }

    public Collection<T> getItems(LatLon location, int depth, Collection<T> itemsOut) // TODO: untested, arg check
    {
        if (!this.contains(this.coverage, location))
        {
            return Collections.emptyList();
        }

        if (this.depth == depth)
        {
            return getAll(itemsOut);
        }
        else if (this.depth < depth)
        {
            for (int i = 0; i < subCoverage.size(); i++)
            {
                if (this.contains(subCoverage.get(i), location))
                {
                    SpatialTree<S, T> child = this.children.get(i);
                    if (child != null)
                    {
                        return child.getItems(location, depth, itemsOut);
                    }
                    return Collections.emptyList();
                }
            }
        }
        return Collections.emptyList();
    }

    public Collection<T> getItems(S extent, Collection<T> itemsOut) // TODO: arg check
    {
        // No interaction with this grid
        if (!this.intersectsCoverage(extent, coverage))
        {
            return Collections.emptyList();
        }

        // Check the case where the input rect equals the coverage area
        if (extent.equals(coverage))
        {
            return getAll(itemsOut);
        }

        if (itemsOut == null)
            itemsOut = new ArrayList<T>();

        // Max Depth Check
        if (this.depth == this.maxDepth)
        {
            for (T item : items)
            {
                if (this.intersects(extent, item))
                    itemsOut.add(item);
            }
        }
        else if (this.depth < this.maxDepth)
        {
            for (int i = 0; i < subCoverage.size(); i++)
            {
                if (intersectsCoverage(extent, subCoverage.get(i)) && children.get(i) != null)
                    children.get(i).getItems(extent, itemsOut);
            }
        }

        return itemsOut;
    }

    public Collection<T> getAll(Collection<T> itemsOut)
    {
        if (addToAllLevels)
        {
            return items;
        }
        else if (this.children == null)
        {
            return items;
        }
        {
            if (itemsOut == null)
                itemsOut = new ArrayList<T>();
            for (SpatialTree<S, T> quad : children)
            {
                if (quad != null)
                {
                    quad.getAll(itemsOut);
                }
            }

            return itemsOut;
        }
    }

    public int getDepth()
    {
        return this.depth;
    }

    public int getMaxDepth()
    {
        return this.maxDepth;
    }

    public S getCoverage()
    {
        return this.coverage;
    }

    private static class ItemInfo<T>
    {
        @SuppressWarnings({"FieldCanBeLocal", "UnusedDeclaration"})
        protected T item;
        protected Sector sector;

        @SuppressWarnings({"UnusedDeclaration"})
        protected ItemInfo(Sector sector, T item)
        {
            this.item = item;
            this.sector = sector;
        }

        protected boolean intersects(Sector s)
        {
            return this.sector.intersects(s);
        }
    }

    public static class SectorTree<S, T> extends SpatialTree<Sector, ItemInfo<T>> // TODO: untested
    {
        public SectorTree(int maxDepth) // TODO: arg check
        {
            this(Sector.FULL_SPHERE, 0, maxDepth, false);
        }

        public SectorTree(Sector coverage, int maxDepth, boolean addToAllLevels) // TODO: arg check
        {
            this(coverage, 0, maxDepth, addToAllLevels);
        }

        protected SectorTree(Sector coverage, int depth, int maxDepth, boolean addToAllLevels)
        {
            super(coverage, depth, maxDepth, addToAllLevels);
        }

        protected SpatialTree<Sector, ItemInfo<T>> createInstance(Sector sector, int depth, int maxDepth,
                                                                  boolean addToAllLevels)
        {
            return new SectorTree<S, T>(sector, depth, maxDepth, addToAllLevels);
        }

        protected boolean intersects(Sector sector, ItemInfo<T> item)
        {
            return sector.intersects(item.sector);
        }

        protected boolean intersectsCoverage(Sector sector1, Sector sector2)
        {
            return sector1.intersects(sector2);
        }

        protected boolean contains(Sector extent, LatLon location)
        {
            return extent.contains(location);
        }

        protected ArrayList<Sector> subdivide(Sector extent, int rows, int cols)
        {
            ArrayList<Sector> children = new ArrayList<Sector>(rows * cols);

            double width = extent.getDeltaLon().degrees / cols;
            double height = extent.getDeltaLat().degrees / rows;

            for (int row = 0; row < rows; row++)
            {
                for (int col = 0; col < cols; col++)
                {
                    children.add(Sector.fromDegrees(
                        extent.getMinLatitude().degrees + row * height,
                        extent.getMinLatitude().degrees + (row + 1) * height,
                        extent.getMinLongitude().degrees + col * width,
                        extent.getMinLongitude().degrees + (col + 1) * width));
                }
            }

            return children;
        }
    }

    public static class LocationTree extends SpatialTree<Sector, LatLon>
    {
        public LocationTree(int maxDepth) // TODO: arg check
        {
            this(Sector.FULL_SPHERE, 0, maxDepth, false);
        }

        public LocationTree(Sector coverage, int maxDepth, boolean addToAllLevels) // TODO: arg check
        {
            super(coverage, 0, maxDepth, addToAllLevels);
        }

        protected LocationTree(Sector coverage, int depth, int maxDepth, boolean addToAllLevels)
        {
            super(coverage, depth, maxDepth, addToAllLevels);
        }

        protected SpatialTree<Sector, LatLon> createInstance(Sector sector, int depth, int maxDepth,
                                                             boolean addToAllLevels)
        {
            return new LocationTree(sector, depth, maxDepth, addToAllLevels);
        }

        protected boolean intersects(Sector sector, LatLon location)
        {
            return sector.contains(location);
        }

        protected boolean intersectsCoverage(Sector sector1, Sector sector2)
        {
            return sector1.intersects(sector2);
        }

        protected boolean contains(Sector extent, LatLon location)
        {
            return extent.contains(location);
        }

        protected ArrayList<Sector> subdivide(Sector extent, int rows, int cols)
        {
            ArrayList<Sector> children = new ArrayList<Sector>(rows * cols);

            double width = extent.getDeltaLon().degrees / cols;
            double height = extent.getDeltaLat().degrees / rows;

            for (int row = 0; row < rows; row++)
            {
                for (int col = 0; col < cols; col++)
                {
                    children.add(Sector.fromDegrees(
                        extent.getMinLatitude().degrees + row * height,
                        extent.getMinLatitude().degrees + (row + 1) * height,
                        extent.getMinLongitude().degrees + col * width,
                        extent.getMinLongitude().degrees + (col + 1) * width));
                }
            }

            return children;
        }
    }

    public static class MarkerTree extends SpatialTree<Sector, Marker>
    {
        public MarkerTree(int maxDepth) // TODO: arg check
        {
            this(Sector.FULL_SPHERE, 0, maxDepth, false);
        }

        public MarkerTree(Sector coverage, int maxDepth, boolean addToAllLevels) // TODO: arg check
        {
            super(coverage, 0, maxDepth, addToAllLevels);
        }

        protected MarkerTree(Sector coverage, int depth, int maxDepth, boolean addToAllLevels)
        {
            super(coverage, depth, maxDepth, addToAllLevels);
        }

        protected SpatialTree<Sector, Marker> createInstance(Sector sector, int depth, int maxDepth,
                                                             boolean addToAllLevels)
        {
            return new MarkerTree(sector, depth, maxDepth, addToAllLevels);
        }

        protected boolean intersects(Sector sector, Marker location)
        {
            return sector.contains(location.getPosition());
        }

        protected boolean intersectsCoverage(Sector sector1, Sector sector2)
        {
            return sector1.intersects(sector2);
        }

        protected boolean contains(Sector extent, LatLon location)
        {
            return extent.contains(location);
        }

        protected ArrayList<Sector> subdivide(Sector extent, int rows, int cols)
        {
            ArrayList<Sector> children = new ArrayList<Sector>(rows * cols);

            double width = extent.getDeltaLon().degrees / cols;
            double height = extent.getDeltaLat().degrees / rows;

            for (int row = 0; row < rows; row++)
            {
                for (int col = 0; col < cols; col++)
                {
                    children.add(Sector.fromDegrees(
                        extent.getMinLatitude().degrees + row * height,
                        extent.getMinLatitude().degrees + (row + 1) * height,
                        extent.getMinLongitude().degrees + col * width,
                        extent.getMinLongitude().degrees + (col + 1) * width));
                }
            }

            return children;
        }
    }
}
//
//    public static void main(String[] args)
//    {
//        int minLat = 23, maxLat = 50, latDelta = 1;
//        int minLon = -130, maxLon = -70, lonDelta = 1;
//        int depth = 3;
//        LocationTree tree = new LocationTree(Sector.fromDegrees(minLat, maxLat, minLon, maxLon), depth, false);
//
//        int count = 0;
//        int iterCount = 1;
//        long start = System.currentTimeMillis();
//        for (int i = 0; i < iterCount; i++)
//        {
////            SpatialTree<Sector, ItemInfo<String>> tree = new MySectorTree<ItemInfo<String>>(5);
////
////            for (int lat = minLat; lat <= maxLat; lat += latDelta)
////            {
////                for (int lon = minLon; lon <= maxLon; lon += lonDelta)
////                {
////                    tree.add(new ItemInfo<String>(Sector.fromDegrees(lon, lat, lon + lonDelta, lat + latDelta), ""));
////                    ++count;
////                }
////            }
//            count = 0;
//            for (int lat = minLat; lat <= maxLat; lat += latDelta)
//            {
//                for (int lon = minLon; lon <= maxLon; lon += lonDelta)
//                {
////                    double latt = lat + 0.9 * latDelta; double lonn = lon + 0.9 * lonDelta;
//                    tree.add(LatLon.fromDegrees(lat, lon));
//                    ++count;
////                    System.out.printf("Count %d, %d, %s\n", count, tree.getAll().size(), LatLon.fromDegrees(latt, lonn));
//                }
//            }
//        }
//
//        System.out.println("Fill: " + (System.currentTimeMillis() - start) + " ms for " + iterCount + " iterations and "
//                           + count + " locations, depth " + depth);
//
//        System.out.println(tree.getAll().size() + " items");
//        Collection<LatLon> cc = tree.getItems(Sector.fromDegrees(minLat, maxLat, minLon, maxLon));
//        System.out.println("Full sector count " + cc.size());
//
//        start = System.currentTimeMillis();
//        for (int i = 0; i < iterCount; i++)
//        {
//            count = 0;
//            double lateps = 0;
//            for (int lat = minLat; lat < maxLat; lat += latDelta)
//            {
//                double loneps = 0;
//                for (int lon = minLon; lon < maxLon; lon += lonDelta)
//                {
//                    Collection<LatLon> c = tree.getItems(
//                        Sector.fromDegrees(lat + lateps, lat + latDelta, lon + loneps, lon + lonDelta));
//                    count += c.size();
//                    loneps = 0.000001;
//                }
//                lateps = 0.000001;
//            }
//        }
//
//        System.out.println("Find: " + (System.currentTimeMillis() - start) + " ms for " + iterCount + " iterations and "
//                           + count + " locations, depth " + depth);
//    }
