/*
Copyright (C) 2001, 2007 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.view;

import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.View;
import gov.nasa.worldwind.util.Logging;

/**
 * @author dcollins
 * @version $Id: EyePositionIterator.java 8914 2009-02-19 19:01:05Z dcollins $
 */
public class EyePositionIterator extends AbstractViewStateIterator
{
    private final Position begin;
    private final Position end;

    public EyePositionIterator(long lengthMillis, Position beginEyePosition, Position endEyePosition)
    {
        super(lengthMillis);
        if (beginEyePosition == null || endEyePosition == null)
        {
            String message = Logging.getMessage("nullValue.PositionIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.begin = beginEyePosition;
        this.end = endEyePosition;
    }

    public EyePositionIterator(long startTimeMillis, long lengthMillis, Position beginEyePosition, Position endEyePosition)
    {
        super(startTimeMillis, lengthMillis);
        if (beginEyePosition == null || endEyePosition == null)
        {
            String message = Logging.getMessage("nullValue.PositionIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        this.begin = beginEyePosition;
        this.end = endEyePosition;
    }

    public final Position getBeginEyePosition()
    {
        return this.begin;
    }

    public final Position getEndEyePosition()
    {
        return this.end;
    }

    protected void doNextState(double interpolant, View view)
    {
        if (interpolant < 0 || interpolant > 1)
        {
            String message = Logging.getMessage("generic.ArgumentOutOfRange", interpolant);
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);    
        }
        if (view  == null)
        {
            String message = Logging.getMessage("nullValue.ViewIsNull");
            Logging.logger().severe(message);
            throw new IllegalArgumentException(message);
        }

        Position nextPosition = this.computeNextPosition(interpolant);
        view.setEyePosition(nextPosition);
    }

    protected Position computeNextPosition(double interpolant)
    {
        Angle azimuth = LatLon.greatCircleAzimuth(this.begin, this.end);
        Angle distance = LatLon.greatCircleDistance(this.begin, this.end);
        Angle pathLength = Angle.fromDegrees(distance.degrees * interpolant);

        LatLon nextLatLon = LatLon.greatCircleEndPosition(this.begin, azimuth, pathLength);
        double nextElevation = (1 - interpolant) * this.begin.getElevation() +  interpolant * this.end.getElevation();

        return new Position(nextLatLon, nextElevation);
    }
}
