/* Copyright (C) 2001, 2009 United States Government as represented by
the Administrator of the National Aeronautics and Space Administration.
All Rights Reserved.
*/
package gov.nasa.worldwind.view;

import gov.nasa.worldwind.geom.*;
import gov.nasa.worldwind.util.RestorableSupport;

/**
 * <code>OrbitViewLimits</code> defines a restriction on the standard viewing parameters of an <code>OrbitView</code>.
 *
 * @author dcollins
 * @version $Id: OrbitViewLimits.java 10301 2009-04-17 20:21:46Z dcollins $
 */
public interface OrbitViewLimits
{
    /**
     * Returns the <code>Sector</code> which limits the <code>OrbitView</code> center latitude and longitude.
     *
     * @return <code>Sector</code> which limits the center latitude and longitude.
     */
    Sector getCenterLocationLimits();

    /**
     * Sets the <code>Sector</code> which will limit the <code>OrbitView</code> center latitude and longitude.
     *
     * @param sector <code>Sector</code> which will limit the center latitude and longitude.
     * @throws IllegalArgumentException if <code>sector</code> is null.
     */
    void setCenterLocationLimits(Sector sector);

    /**
     * Returns the minimum and maximum values for the <code>OrbitView</code> center elevation.
     *
     * @return minimum and maximum allowable values for center elevation.
     */
    double[] getCenterElevationLimits();

    /**
     * Sets the minimum and maximum values which will limit the <code>OrbitView</code> center elevation.
     *
     * @param minValue the minimum allowable value for center elevation.
     * @param maxValue the maximum allowable value for center elevation.
     */
    void setCenterElevationLimits(double minValue, double maxValue);

    /**
     * Returns the minimum and maximum <code>Angles</code> for the <code>OrbitView</code> heading property.
     *
     * @return minimum and maximum allowable <code>Angles</code> for heading.
     */
    Angle[] getHeadingLimits();

    /**
     * Sets the minimum and maximum <code>Angles</code> which will limit the <code>OrbitView</code> heading property.
     *
     * @param minAngle the minimum allowable angle for heading.
     * @param maxAngle the maximum allowable angle for heading.
     * @throws IllegalArgumentException if eiter <code>minAngle</code> or <code>maxAngle</code> is null.
     */
    void setHeadingLimits(Angle minAngle, Angle maxAngle);

    /**
     * Returns the minimum and maximum <code>Angles</code> for the <code>OrbitView</code> pitch property.
     *
     * @return minimum and maximum allowable <code>Angles</code> for pitch.
     */
    Angle[] getPitchLimits();

    /**
     * Sets the minimum and maximum <code>Angles</code> which will limit the <code>OrbitView</code> pitch property.
     *
     * @param minAngle the minimum allowable angle for pitch.
     * @param maxAngle the maximum allowable angle for pitch.
     * @throws IllegalArgumentException if eiter <code>minAngle</code> or <code>maxAngle</code> is null.
     */
    void setPitchLimits(Angle minAngle, Angle maxAngle);

    /**
     * Returns the minimum and maximum values for the <code>OrbitView</code> zoom property.
     *
     * @return minimum and maximum allowable values for zoom.
     */
    double[] getZoomLimits();

    /**
     * Sets the minimum and maximum values which will limit the <code>OrbitView</code> zoom property.
     *
     * @param minValue the mimimum allowable value for zoom.
     * @param maxValue the maximum allowable value for zoom.
     */
    void setZoomLimits(double minValue, double maxValue);

    void getRestorableState(RestorableSupport rs, RestorableSupport.StateObject context);

    void restoreState(RestorableSupport rs, RestorableSupport.StateObject context);
}
